/*
	The Malete project - the Z39.2/Z39.50 database framework of OpenIsis.
	Version 0.9.x (patchlevel see file Version)
	Copyright (C) 2001-2004 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	see README for more information
EOH */

/*
	$Id: cds.c,v 1.1 2005/01/24 17:58:35 kripke Exp $
	malete CDS/ISIS conversion
*/

#include "tool.h"

/* structs with 4-byte aligned ("UNIX") or packed (Intel "DOS"+WinIsis) ints */
#ifdef CPU_BIG_ENDIAN
#	define BUILD_CDSUNIX /* no packed version on SPARCs, M68K or PPCs */
#endif
#ifdef BUILD_CDSUNIX
#	define PACKED
#else
#	ifdef __GNUC__
#		define PACKED __attribute__((packed))
#	else
#		error "please define how to pack"
#	endif
#endif

enum {
	CDS_MST, CDS_XRF, /* master file, cross ref */
	CDS_CNT, CDS_IFP, /* index control and record pointers */
	/* do not change this order: see f2 */
	CDS_N01, CDS_L01, /* 10 byte nodes and leaves */
	CDS_N02, CDS_L02, /* 30 byte nodes and leaves */
	CDS_FILES
};
static const char *cds_ext[] = {
	"mst", "xrf", "cnt", "ifp", "n01", "l01", "n02", "l02"
};
typedef file cds_files[CDS_FILES];


typedef struct { /* master file control header */
	int   ctlm; /* should be 0 */
	int   nmfn; /* next avail MFN */
	int   nmfb; /* next free block, 1st = 1 */
	short nmfp; /* next free pos (in block) */
	short type; /* 0 (1 is for "messages" mst) */
	int   rcnt; /* record count; nmfn-1 - #deleted */
	int   mfx1; /* unused */
	int   mfx2; /* unused; "LAN lock" */
	int   mfx3; /* unused; "LAN lock" */
} Mfc;


typedef struct { /* field dictionary in master file record */
	unsigned short tag;
	unsigned short pos;
	unsigned short len;
} Dct;

typedef struct { /* master file record */
	int   mfn;
	short mfrl;
	int   bwb PACKED;
	short bwp;
	short base;
	short nvf;
	short stat;
	Dct   dict[1];
} Mfr;
/* it is common belief, that the first 4 fields
(12 bytes packed, 14 aligned) are to be in one 512-byte block
*/
enum {
	MFR_BASEOFF = ((char*)&((Mfr*)0)->base) - (char*)0,
	MFR_DICTOFF = ((char*)&((Mfr*)0)->dict) - (char*)0
};

typedef struct { /* cross ref block */
	int xpos;
	int xrec[127];
}	Xrf;

typedef struct { /* index control */
	short type;	/* 1 or 2 of N0x */
	short ordn; /* 5 */
	short ordf; /* 5 */
	short n;    /* 15 */
	short k;    /* 5 */
	short lev;  /* depth: N0x levels under root (root only: 0) */
	int   posr; /* root pos */
	int   nmax; /* N0x blocks */
	int   fmax; /* L0x blocks */
	short abno; /* 0 (!) if "abnormal" (root only) */
}	Cnt;
#ifdef BUILD_CDSUNIX /* while Cnt layout is independent of alignment ... */
#	define CNT_SIZE 28 /* ... sizeof(Cnt) is not */
#else
#	define CNT_SIZE 26
#endif

typedef struct { /* "posting" pointer to record */
	char b[8]; /* 3+2+1+2 byte rid+tag+occ+pos big endian */
}	Pst;

typedef struct { /* inverted file postings header */
	int nxtb; /* continuation */
	int nxtp;
	int totp; /* total postings */
	int segp; /* actually in this segment <= totp */
	int segc; /* segment capacity >= segc */
	/* Pst pst[...]; possibly spanning several Ifb blocks */
} Ifp;

typedef struct { /* inverted file block */
	int ifpblk;
	int ifprec[127]; /* ... */
} Ifb;

#define NSTRUCT(n) struct { \
	int   pos; \
	short ock; \
	short it; \
	struct { \
		char key[n]; \
		int  ref PACKED; \
	} idx[10]; \
}
typedef NSTRUCT(10) N01;
typedef NSTRUCT(30) N02;

#define LSTRUCT(n) struct { \
	int   pos; \
	short ock; \
	short it; \
	int   ps; \
	struct { \
		char key[n]; \
		int  infb PACKED; \
		int  infp PACKED; \
	} idx[10]; \
}
typedef LSTRUCT(10) L01;
typedef LSTRUCT(30) L02;

static const unsigned char cp850[] = { /* Latin1 values of Cp850 upper half */
	199,252,233,226,228,224,229,231,234,235,232,239,238,236,196,197,
	201,230,198,244,246,242,251,249,255,214,220,248,163,216,215,'?',
	225,237,243,250,241,209,170,186,191,174,172,189,188,161,171,187,
	'?','?','?','?','?',193,194,192,169,'?','?','?','?',162,165,'?',
	'?','?','?','?','?','?',227,195,'?','?','?','?','?','?','?',164,
	240,208,202,203,200,'?',205,206,207,'?','?','?','?',166,204,'?',
	211,223,212,210,245,213,181,254,222,218,219,217,253,221,175,180,
	173,177,'?',190,182,167,247,184,176,168,183,185,179,178,'?',160
};

typedef struct {
	Fld base;  /* points into env->opt */
	int notab; /* do not convert ^ <-> tabs */
	int novt;  /* do not convert newline <-> vtabs */
	int no850; /* do not convert Cp850 <-> ISO Latin1 */
	int write; /* write cds */
} CdsOpt;


static int cdsOpt (CdsOpt *opt, Fld *args, Db *db)
{
	Fld o;
	memset(opt, 0, sizeof(*opt));
	for ( o.val = 0; vGet(&o, args, "BN"); )
		switch (o.tag) {
		case 'B':
			opt->base = o;
			continue;
		case 'N':
			if (VEQZ(&o, "otab"))
				opt->notab = 1;
			else if (VEQZ(&o, "ovt"))
				opt->novt = 1;
			else if (VEQZ(&o, "o850"))
				opt->no850 = 1;
			else
				return eRr(ERR_INVAL, "bad cds opt '-N%.*s'", o.len, o.val);
		}
	if ( !opt->base.len )
		opt->base.len = strlen(opt->base.val = db->pat);
	return 0;
}	/* cdsOpt */


int cdsuse ()
{
	eRr(ERR_INVAL, "[cdsimp|cdsexp]:\n"
		"-Bbase: path/basename of cds db (default = malete db)\n"
		"-Notab: do not convert ^ <-> tab\n"
		"-Novt: do not convert newline <-> vtab\n"
		"-No850: do not convert Cp850 <-> ISO Latin1"
	);
	return ERR_INVAL;
}	/* cdsuse */


static int cdsOpen (cds_files cds, CdsOpt *opt)
{
	int filmode = opt->write ? FIL_RDWR|FIL_CREAT|FIL_TRUNC : FIL_RD;
	int l = opt->base.len, i;
	char *path = mAlloc(l+5);
	memcpy(path, opt->base.val, l);
	path[l++] = '.';
	for (i=0; i<CDS_FILES; i++)
		cds[i] = FIL_NONE;
	for (i=0; i<CDS_FILES; i++) {
		memcpy(path+l, cds_ext[i], 4);
		if ( fOpen( cds+i, path, filmode ) )
			return eRr(LOG_ERROR, "no access to CDS file '%s'", path);
	}
	return 0;
}	/* cdsOpen */


static void cdsClose (cds_files cds)
{
	int i;
	for (i=0; i<CDS_FILES; i++)
		fClose(cds+i);
}	/* cdsClose */


int cdsimp (Db *db, Fld *args)
{
	CdsOpt opt;
	cds_files cds;
	int ret = 0, m, i;
	Mfc mfc;
	List rec;
	char recode[256];
	L01 l01;
	L02 l02;
	Ifb ifbs[2];
	int n1, n2;
	Key key;

	if ( cdsOpt(&opt, args, db) )
		return cdsuse();
	if ( cdsOpen(cds, &opt) )
		return eRr(ERR_INVAL, "bad CDS base");
#define CLEANUP( args ) do { ret = eRr args; goto cleanup; } while (0)
	if ( sizeof(mfc) != fPread(cds+CDS_MST, &mfc, sizeof(mfc), 0) )
		CLEANUP((ERR_TRASH, "no mfc"));
	eRr(LOG_TRACE,
		"MFC ctlm %d nmfn %d nmfb %d nmfp %d type %d"	
		" rcnt %d x1 %d x2 %d x3 %d",
		mfc.ctlm, mfc.nmfn, mfc.nmfb, mfc.nmfp, mfc.type,
		mfc.rcnt, mfc.mfx1, mfc.mfx2, mfc.mfx3);
	eRr(LOG_INFO, "importing %d recs", mfc.nmfn-1);
	lInit(&rec, "");
	/* prepare recode table */
	for (i=0; i<128; i++) recode[i] = (char)i;
	if ( !opt.notab ) recode['^'] = TAB;
	if ( !opt.novt ) recode[LF] = VT;
	if ( opt.no850 )
		for (i=128; i<256; i++) recode[i] = (char)i;
	else
		for (i=128; i<256; i++) recode[i] = (char)cp850[i-128];
	/* loop m := ceil(numrecs/127) xrf blocks */
	m = (mfc.nmfn+125)/127;
	for (i=1; i<=m; i++) {
		Xrf xrf;
		int f = 1+127*(i-1), t = f+126, mfn;
		int got = fRead(cds+CDS_XRF, &xrf, sizeof(xrf));
		if ( (int)sizeof(xrf) != got )
			CLEANUP((ERR_TRASH, "reading xrf block %d", i));
		if ( xrf.xpos != (i==m ? -i : i) )
			CLEANUP((ERR_TRASH, "bad pos %d in xrf block %d", xrf.xpos, i));
		if ( t >= mfc.nmfn )
			t = mfc.nmfn-1;
		for (mfn=f; mfn<=t; mfn++) {
			int pos = xrf.xrec[mfn-f];
			int mfb = (pos >> 11);
			union {
				Mfr  mfr;
				char buf[0x7fff];
			} da;
			Dct *d = da.mfr.dict;
			char *b, *e;

			if ( 0 > mfb ) {
				eRr(LOG_TRACE, "rec %d was deleted", mfn);
				continue;
			}
			/* ignore bits 9, 10 signaling pending index update */
			pos &= 0x1ff;
			pos |= (mfb-1)<<9;
			got = fPread(cds+CDS_MST, da.buf, sizeof(Mfr), pos);
			if ( got < MFR_DICTOFF )
				CLEANUP((ERR_TRASH, "rec %d too short %d", mfn, got));
			if ( da.mfr.mfrl < 0 ) {
				eRr(LOG_WARN, "rec %d pos %d locked", mfn, pos);
				da.mfr.mfrl = -da.mfr.mfrl;
			}
			if ( da.mfr.mfrl < (int)(MFR_DICTOFF+da.mfr.nvf*sizeof(Dct)) )
				CLEANUP((ERR_TRASH, "rec %d pos %d bad mfrl %d nvf %d",
					mfn, pos, da.mfr.mfrl, da.mfr.nvf));
			got = fPread(cds+CDS_MST, da.buf, da.mfr.mfrl, pos);
			if ( da.mfr.mfrl != got ) {
				if ( got == -da.mfr.mfrl )
					da.mfr.mfrl = -da.mfr.mfrl;
				else
				CLEANUP((ERR_TRASH, "rec %d pos %d got %d bytes want %d",
					mfn, pos, got, da.mfr.mfrl));
			}
			b = da.buf+da.mfr.base;
			e = da.buf+da.mfr.mfrl;
			lReset(&rec);
			for ( ; da.mfr.nvf--; d++ ) {
				char *p = b+d->pos, *q = p+d->len;
				if ( p < b || q > e )
					CLEANUP((ERR_TRASH, "bad pos %d len %d field %d",
						d->pos, d->len, d-da.mfr.dict));
				for ( ; p<q; p++ ) *p = recode[(unsigned char)*p];
				LADD(&rec, d->tag, b+d->pos, d->len);
			}
			dWrite(db, rec.fld, mfn);
		}
	}
	/* mst done */
	/* we assume first leaf has pos 1; maybe we should use N0x to search */
	l01.ps = l02.ps = 1;
	l01.ock = l02.ock = n1 = n2 = 0;
	ifbs[0].ifpblk = ifbs[1].ifpblk = 0;
	for (;;) { /* loop leaves */
		Fld use;
		int infb, infp, got, more = 0;
		Ifp ifp;
		Ifb *ifb;

		if ( n1 == l01.ock && (l01.pos = l01.ps) ) {
			if ((int)sizeof(l01) != (got 
				= fPread(cds+CDS_L01,&l01,sizeof(l01),(l01.ps-1)*sizeof(l01)))
			)
				CLEANUP((ERR_TRASH, "l01 %d got %d", l01.ps, got));
			eRr(LOG_TRACE, "l01 pos %d ock %d nxt %d", l01.pos, l01.ock, l01.ps);
			n1 = 0;
			if ( !l01.ock )
				continue;
		}
		if ( n2 == l02.ock && (l02.pos = l02.ps) ) {
			if ((int)sizeof(l02) != (got 
				= fPread(cds+CDS_L02,&l02,sizeof(l02),(l02.ps-1)*sizeof(l02)))
			)
				CLEANUP((ERR_TRASH, "l02 %d got %d", l02.ps, got));
			eRr(LOG_TRACE, "l02 pos %d ock %d nxt %d", l02.pos, l02.ock, l02.ps);
			n2 = 0;
			if ( !l02.ock )
				continue;
		}
		if ( !l01.pos && !l02.pos )
			break;
		if ( n1 == l01.ock )
			use.tag = 1;
		else if ( n2 == l02.ock )
			use.tag = 0;
		else if ( 0 > memcmp(l02.idx[n2].key, l01.idx[n1].key, 10) )
			use.tag = 1;
		else
			use.tag = 0;
		if ( use.tag ) {
			infb = l02.idx[n2].infb;
			infp = l02.idx[n2].infp;
			use.val = l02.idx[n2++].key;
			use.len = 30;
		} else {
			infb = l01.idx[n1].infb;
			infp = l01.idx[n1].infp;
			use.val = l01.idx[n1++].key;
			use.len = 10;
		}
		ifb = ifbs+use.tag;
		/* kill trailing blanks */
		while ( use.len && ' ' == use.val[use.len-1] ) use.len--;
		eRr(LOG_TRACE, "'%.*s'", use.len, use.val);
		memcpy(key.byt, use.val, key.len = use.len);
		key.val.len = db->qdx.vsz;
		while (infb) {
			Pst *pst;
			int n;
			if ( ifb->ifpblk != infb
				&& (512 != fPread(cds+CDS_IFP,ifb,sizeof(Ifb),(infb-1)*512)
					|| ifb->ifpblk != infb)
			)
				CLEANUP((ERR_TRASH, "ifp %d",infb));
			if ( more ) {
				pst = (Pst*)ifb->ifprec;
				if ( (n = 127/2) > more )
					n = more;
				more -= n;
				eRr(LOG_TRACE, "+%d @%d", n, infb);
			} else {
				if ( infp > 127-5 )
					CLEANUP((ERR_TRASH, "bad infp %d for infb %d", infp, infb));
				ifp = *(Ifp*)(ifb->ifprec + infp);
				pst = (Pst*)(ifb->ifprec + infp + 5);
				if ( (n = (127-5-infp)/2) > ifp.segp )
					n = ifp.segp;
				more = ifp.segp - n;
				eRr(LOG_TRACE, "ifp @%d,%d: %d %d %d %d %d", infb, infp,
					ifp.nxtb, ifp.nxtp, ifp.totp, ifp.segp, ifp.segc);
			}
			if ( more )
				infb++;
			else {
				infb = ifp.nxtb;
				infp = ifp.nxtp;
			}
			for ( ; n--; pst++ ) {
				if ( QDX_ISIS == db->qdx.ptr )
					memcpy(key.val.byt, pst->b, 8);
				else {
					Ptr ptr;
					ptr.ext = 0;
					ptr.rid = pst->b[0]<<16 | pst->b[1]<<8 | pst->b[2];
					ptr.tag = pst->b[3]<<8 | pst->b[4];
					ptr.pos = pst->b[5]<<16 | pst->b[6]<<8 | pst->b[7];
					qMkVal(&key.val, &ptr, db->qdx.ptr);
				}
				qLoad(&db->qdx, &key);
			}
		} /* while infb */
	}
	key.val.len = 0;
	qLoad(&db->qdx, &key);
cleanup:
#undef CLEANUP
	cdsClose(cds);
	return ret;
}	/* cdsimp */


/* export index */
typedef struct {
	QLoop qlp;
	file *cds;
	int   totp; /* in current segment */
	int   infb; /* of current ifp */
	int   infp; /* of current ifp */
	Ifb   ifb; /* current block; flushed on demand */
	int   rec; /* next in ifb.ifprec 0..126 */
	L01   l01; /* new keys are immediatly inserted here ... */
	L02   l02; /* ... structs flushed on demand */
	N01   n01;
	N02   n02;
} CdsLoop;


static void closeifp (CdsLoop *self) /* write Ifp header */
{
	Ifp buf, *ifp;
	if ( self->infb != self->ifb.ifpblk ) /* Ifp not in current blk */
		ifp = &buf;
	else
		ifp = (Ifp*)(self->ifb.ifprec+self->infp);
	ifp->nxtb = ifp->nxtp = 0; /* used only for partial updates */
	ifp->totp = ifp->segp = ifp->segc = self->totp; /* diff only after split */
	if ( self->infb != self->ifb.ifpblk )
		fPwrite(self->cds+CDS_IFP, ifp, sizeof(Ifp),
			((self->infb-1)<<9) + 4*(self->infp+1));
	self->totp = 0;
/*
	 hmmmm ... From the manual:
   Cuando se carga secuencialmente el archivo           
   invertido (por ejemplo despues de una generacion completa del archivo        
   invertido con ISISINV), cada lista esta formada por uno o mas segmentos      
   adyacentes. Si IFPTOT <= 32768, entonces: IFPNXTB/IFPNXTP = 0/0 y IFPTOT =   
   IFPSEGP = IFPSEGC.
	 This seems to imply that numbers really should be used as signed shorts!?
*/
}	/* closeifp */

static void closeifb (CdsLoop *self)
{
	eRr(LOG_TRACE, "w ifb %d", self->ifb.ifpblk);
	fPwrite(self->cds+CDS_IFP,&self->ifb,sizeof(Ifb),(self->ifb.ifpblk-1)<<9);
	self->ifb.ifpblk++;
	self->rec = 0;
}	/* closeifb */


static int cdsCb (CdsLoop *self)
{
	Pst *pst;
	unsigned i, vsz = self->qlp.qdx->vsz;
	const unsigned char *v = self->qlp.vals;

	if ( !(QSAME & self->qlp.flg) ) { /* add new key -- new segment */
		L01 *l0x; /* might really be a L02 */
		int l, n, oref; /* key length, offset of ref int(s) in idx */
		int f2, sl, *ref;
		char *key;

		if ( self->totp )
			closeifp(self);
		if ( self->rec > 127-5 )
			closeifb(self);
		if ( (l = self->qlp.cur.len) <= 10 ) { /* l01 entry */
			l0x = &self->l01;
			f2 = 0;
			n = 10;
			sl = sizeof(L01);
		} else {
			l0x = (L01*)&self->l02;
			f2 = 2;
			if ( l > (n = 30) )
				l = 30;
			sl = sizeof(L02);
		}
#ifdef BUILD_CDSUNIX
		oref = n+2;
#else
		oref = n;
#endif
		if ( 10 == l0x->ock ) { /* need new block */
			l0x->ps = l0x->pos+1;
			eRr(LOG_TRACE, "w l0 n %d pos %d", n, l0x->pos);
			fPwrite(self->cds+CDS_L01+f2, l0x, sl, (l0x->pos-1)*sl);
			l0x->pos++;
			l0x->ock = 0;
		}
		key = l0x->idx[0].key + l0x->ock*(oref+8); /* sizeof idx */
		memset(key, ' ', n);
		memcpy(key, self->qlp.cur.byt, l);
		ref = (int*)(key+oref);
		ref[0] = self->infb = self->ifb.ifpblk;
		ref[1] = self->infp = self->rec;
		self->rec += 5; /* reserve space for Ifp header */
		if ( !l0x->ock++ ) { /* also write node entry */
			N01 *n0x = 10==n ? &self->n01 : (N01*)&self->n02;
			int sn = 10==n ? sizeof(N01) : sizeof(N02);
			char *nkey;
			if ( 10 == n0x->ock ) {
				eRr(LOG_TRACE, "w n0 n %d pos %d", n, n0x->pos);
				fPwrite(self->cds+CDS_N01+f2, n0x, sn, (n0x->pos-1)*sn);
				n0x->pos++;
				n0x->ock = 0;
			}
			nkey = n0x->idx[0].key + n0x->ock++*(oref+4);
			if ( 2 == n0x->pos && 1 == n0x->ock )
				memset(nkey, ' ', n); /* 1st key - start of level */
			else
				memcpy(nkey, key, n);
			*(int*)(nkey+oref) = -l0x->pos;
		}
	}
	pst = (Pst*)(self->ifb.ifprec + self->rec);
	eRr(LOG_TRACE, "+ %d ptr @ %d.%d", self->qlp.nvals,
		self->ifb.ifpblk, self->rec);
	for ( i=self->qlp.nvals; i--; self->rec+=2, pst++, v += vsz) {
		if ( self->rec > 125 ) {
			closeifb(self);
			pst = (Pst*)self->ifb.ifprec;
		}
		if ( QDX_ISIS == self->qlp.qdx->ptr )
			memcpy(pst->b, v, 8);
		else {
			Ptr ptr;
			qRdVal(&ptr, v, self->qlp.qdx->ptr);
			pst->b[0] = ptr.rid >> 16;
			pst->b[1] = ptr.rid >> 8;
			pst->b[2] = ptr.rid;
			pst->b[3] = ptr.tag >> 8;
			pst->b[4] = ptr.tag;
			pst->b[5] = ptr.pos >> 16;
			pst->b[6] = ptr.pos >> 8;
			pst->b[7] = ptr.pos;
		}
	}
	self->totp += self->qlp.nvals;
	return 0;
}	/* cdsCb */


int cdsexp (Db *db, Fld *args)
{
	CdsOpt opt;
	cds_files cds;
	int ret = 0, pos, i, j;
	Mfc mfc;
	List rec;
	char recode[256];
	Xrf xrf;
	CdsLoop clp;
	Cnt cnt[2];

	if ( cdsOpt(&opt, args, db) )
		return cdsuse();
	opt.write = 1;
	if ( cdsOpen(cds, &opt) )
		return eRr(ERR_INVAL, "error creating CDS base");
	eRr(LOG_INFO, "exporting %d recs", db->rdx.mid);
	lInit(&rec, "");
	/* prepare recode table */
	for (i=0; i<128; i++) recode[i] = (char)i;
	if ( !opt.notab ) recode[TAB] = '^';
	if ( !opt.novt ) recode[VT] = LF;
	if ( opt.no850 )
		for (i=128; i<256; i++) recode[i] = (char)i;
	else {
		for (i=128; i<256; i++) recode[i] = '?';
		for (i=160; i<256; i++) /* reverse cp850 */
			for (j=0; j<128; j++) if ( i == cp850[j] ) { recode[i] = j+128; break; }
	}
	xrf.xpos = 1;
	mfc.rcnt = 0; /* record count; nmfn-1 - #deleted */
	pos = 64; /* as of folklore, 1st rec should start here */
	for (i=1; i<=db->rdx.mid; i++) { /* loop records */
		int xref;
		if ( 0 >= dRead(lClr(&rec), db, i) ) { /* no rec */
			eRr(LOG_TRACE, "rec %d was deleted", i);
			xref = -1 << 11; 
		} else {
			union {
				Mfr  mfr;
				char buf[0x7fff];
			} da;
			char *b, *p, *e = da.buf+sizeof(da.buf);
			Dct *d = da.mfr.dict, *de;
			Fld *f, *fe = LEND(&rec);
			unsigned avl = 0x7fff - MFR_DICTOFF;

			xref = pos & 511; /* pos in block */
			/* the manual explicitly demands 0-498 and forbids 500 */
			if ( xref > 498 /*512 - MFR_BASEOFF = 500 packed */ ) {
				pos += 512; /* advance to next block */
				pos &= ~511;
				xref = 0;
			}
			xref |= ((pos >> 9)+1) << 11; 
			da.mfr.mfn  = i;
			da.mfr.bwb  = 0;
			da.mfr.bwp  = 0;
			da.mfr.base = 0;
			da.mfr.nvf  = 0;
			da.mfr.stat = 0;
			for ( f=rec.fld; ++f<fe; ) {
				if ( avl < (int)sizeof(Dct) )
					break;
				da.mfr.nvf++;
				avl -= sizeof(Dct);
				if ( avl >= f->len )
					avl -= f->len;
				else
					avl = 0;
			}
			p = b = da.buf + (da.mfr.base = MFR_DICTOFF + da.mfr.nvf*sizeof(Dct));
			de = d + da.mfr.nvf;
			for ( f=rec.fld; d<de; d++ ) {
				unsigned char *v = (unsigned char*)(++f)->val, *ve;
				d->tag = f->tag;
				d->pos = p-b;
				d->len = e-p < (int)f->len ? e-p : (int)f->len;
				for ( ve = v+d->len; v < ve; ) *p++ = recode[*v++];
			}
			da.mfr.mfrl = p - da.buf;
			eRr(LOG_VERBOSE, "write rec %d len %d pos %d", i, da.mfr.mfrl, pos);
			fPwrite(cds+CDS_MST, da.buf, da.mfr.mfrl, pos);
			if ( 1 & (pos += da.mfr.mfrl) ) /* round to even */
				pos++;
			mfc.rcnt++;
		}
		xrf.xrec[j = (i-1)%127] = xref;
		if ( i == db->rdx.mid ) { /* last */
			while ( ++j < 127 ) xrf.xrec[j] = 0;
			xrf.xpos = -xrf.xpos;
		} else if ( 126 != j )
			continue;
		eRr(LOG_VERBOSE, "%d: write Xrf %d", i, xrf.xpos);
		fWrite(cds+CDS_XRF, &xrf, sizeof(Xrf));
		xrf.xpos++;
	}
	mfc.ctlm = 0; /* should be 0 */
	mfc.nmfn = i; /* next avail MFN */
	mfc.nmfb = pos >> 9; /* next free block, 1st = 1 */
	mfc.nmfp = pos & 511; /* next free pos (in block) */
	mfc.type = 0; /* 0 (1 is for "messages" mst) */
	mfc.mfx1 = 0; /* unused */
	mfc.mfx2 = 0; /* unused; "LAN lock" */
	mfc.mfx3 = 0; /* unused; "LAN lock" */
	fPwrite(cds+CDS_MST, &mfc, sizeof(mfc), 0);
	/* mst done */
	memset(&clp, 0, sizeof(clp));
	clp.qlp.qcb = (QCb*)cdsCb;
	clp.cds = cds;
	clp.ifb.ifpblk = 1;
	clp.l01.pos = clp.l02.pos = 1;
	clp.n01.pos = clp.n02.pos = 2; /* start at 2; 1 will hold the root */
	clp.l01.it = clp.n01.it = 1;
	clp.l02.it = clp.n02.it = 2;
	clp.qlp.qdx = &db->qdx;
	qLoop(&clp.qlp);
	eRr(LOG_VERBOSE, "loop got pos ifp %d l01 %d l02 %d n01 %d n02 %d",
		clp.ifb.ifpblk, clp.l01.pos, clp.l02.pos, clp.n01.pos, clp.n02.pos);
	if ( clp.totp )
		closeifp(&clp);
	if ( clp.rec )
		closeifb(&clp);
	clp.l01.ps = clp.l02.ps = 0; /* no next */
	fPwrite(cds+CDS_L01, &clp.l01, sizeof(L01), (clp.l01.pos-1)*sizeof(L01));
	fPwrite(cds+CDS_L02, &clp.l02, sizeof(L02), (clp.l02.pos-1)*sizeof(L02));
	for ( i=2; i--; ) {
		cnt[i].type = i+1;
		cnt[i].ordn = 5;
		cnt[i].ordf = 5;
		cnt[i].n    = 15;
		cnt[i].k    = 5;
		cnt[i].lev  = 0;
		cnt[i].posr = 1;
		cnt[i].nmax = 1;
		cnt[i].abno = 1; 
	}
	cnt[0].fmax = clp.l01.pos;
	cnt[1].fmax = clp.l02.pos;
	if ( 2 == clp.n01.pos ) { /* the poor one and only node */
		/* since we did not open a 2nd node, this one was not written */
		clp.n01.pos = 1; /* make it the root */
		cnt[0].abno = 0; /* really the field means "normal", not "abnormal" */
	}
	fPwrite(cds+CDS_N01, &clp.n01, sizeof(N01), (clp.n01.pos-1)*sizeof(N01));
	if ( 2 == clp.n02.pos ) {
		clp.n02.pos = 1;
		cnt[1].abno = 0;
	}
	fPwrite(cds+CDS_N02, &clp.n02, sizeof(N02), (clp.n02.pos-1)*sizeof(N02));
	/* build the trees */
	if ( cnt[0].abno ) {
		int f = 2, t = clp.n01.pos; /* from - to on level */
		for (;;) {
			N01 c, p; /* child, parent */
			int start = t+1;
			cnt[0].lev++;
			p.pos = 10 >= (t-f) ? 1 : start;
			p.ock = 1; /* blank key ref to f */
			p.it = 1;
			memset(p.idx[0].key, ' ', sizeof(p.idx[0].key));
			p.idx[0].ref = f;
			for ( i=f+1; i<=t; i++ ) {
				fPread(cds+CDS_N01, &c, sizeof(c), (i-1)*sizeof(c));
				eRr(LOG_TRACE, "lev %d blk %d '%.*s'->%d - %d:'%.*s'->%d",
					cnt[0].lev-1, c.pos, sizeof(c.idx[0].key), c.idx[0].key, c.idx[0].ref,
					c.ock, sizeof(c.idx[0].key), c.idx[c.ock-1].key, c.idx[c.ock-1].ref);
				if ( 10 == p.ock ) {
					eRr(LOG_TRACE, "lev %d blk %d '%.*s'->%d - '%.*s'->%d",
						cnt[0].lev, p.pos, sizeof(p.idx[0].key), p.idx[0].key, p.idx[0].ref,
						sizeof(p.idx[0].key), p.idx[9].key, p.idx[9].ref);
					fPwrite(cds+CDS_N01, &p, sizeof(p), (p.pos-1)*sizeof(p));
					p.pos++;
					p.ock = 0;
				}
				memcpy(p.idx[p.ock].key, c.idx[0].key, sizeof(c.idx[0].key));
				p.idx[p.ock++].ref = i;
			}
			eRr(LOG_TRACE, "lev %d blk %d '%.*s'->%d - %d:'%.*s'->%d",
				cnt[0].lev, p.pos, sizeof(p.idx[0].key), p.idx[0].key, p.idx[0].ref,
				p.ock, sizeof(p.idx[0].key), p.idx[p.ock-1].key, p.idx[p.ock-1].ref);
			fPwrite(cds+CDS_N01, &p, sizeof(p), (p.pos-1)*sizeof(p));
			if ( 1 == p.pos )
				break;
			f = start;
			cnt[0].nmax = t = p.pos;
		}
	}
	if ( cnt[1].abno ) {
		int f = 2, t = clp.n02.pos; /* from - to on level */
		for (;;) {
			N02 c, p; /* child, parent */
			int start = t+1;
			cnt[1].lev++;
			p.pos = 10 >= (t-f) ? 1 : start;
			p.ock = 1; /* blank key ref to f */
			p.it = 2;
			memset(p.idx[0].key, ' ', sizeof(p.idx[0].key));
			p.idx[0].ref = f;
			for ( i=f+1; i<=t; i++ ) {
				fPread(cds+CDS_N02, &c, sizeof(c), (i-1)*sizeof(c));
				eRr(LOG_TRACE, "lev %d blk %d '%.*s'->%d - %d:'%.*s'->%d",
					cnt[1].lev-1, c.pos, sizeof(c.idx[0].key), c.idx[0].key, c.idx[0].ref,
					c.ock, sizeof(c.idx[0].key), c.idx[c.ock-1].key, c.idx[c.ock-1].ref);
				if ( 10 == p.ock ) {
					eRr(LOG_TRACE, "lev %d blk %d '%.*s'->%d - '%.*s'->%d",
						cnt[1].lev, p.pos, sizeof(p.idx[0].key), p.idx[0].key, p.idx[0].ref,
						sizeof(p.idx[0].key), p.idx[9].key, p.idx[9].ref);
					fPwrite(cds+CDS_N02, &p, sizeof(p), (p.pos-1)*sizeof(p));
					p.pos++;
					p.ock = 0;
				}
				memcpy(p.idx[p.ock].key, c.idx[0].key, sizeof(c.idx[0].key));
				p.idx[p.ock++].ref = i;
			}
			eRr(LOG_TRACE, "lev %d blk %d '%.*s'->%d - %d:'%.*s'->%d",
				cnt[0].lev, p.pos, sizeof(p.idx[0].key), p.idx[0].key, p.idx[0].ref,
				p.ock, sizeof(p.idx[0].key), p.idx[p.ock-1].key, p.idx[p.ock-1].ref);
			fPwrite(cds+CDS_N02, &p, sizeof(p), (p.pos-1)*sizeof(p));
			if ( 1 == p.pos )
				break;
			f = start;
			cnt[1].nmax = t = p.pos;
		}
	}
	fWrite(cds+CDS_CNT,cnt,CNT_SIZE);
	fWrite(cds+CDS_CNT,cnt+1,CNT_SIZE);
	cdsClose(cds);
	return ret;
}	/* cdsexp */
