/*
	The Malete project - the Z39.2/Z39.50 database framework of OpenIsis.
	Version 0.9.x (patchlevel see file Version)
	Copyright (C) 2001-2003 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	see README for more information
EOH */

/*
	$Id: cdx.c,v 1.1 2005/01/24 17:58:49 kripke Exp $
	charset collation
*/

#include "core.h"

/*
We use
-	p codes 0,1,..p-1 for primaries (word and nonword)
-	v codes p..p+v-1 for secondary and tertiary level variants
-	m codes p+v..p+v+m-1 for maps (targets)

We index code numbers to an array[p+v+m] of unsigned ints as code values.
The highest bit is the nonword indicator.
The lower nibble of the highest byte is the length of the code's byte sequence
(for maps in #code values, else in # cleartext bytes).

If the sequence does not fit into the lower 3 bytes,
the value&0xffffff is the offset of the bytes.

For primaries and variants, the bytes are cleartext.
For maps, the bytes are code text (length given as #codes),
where every code uses one byte, if p+v<=256, else two.

With variants, we also have an array[p+v] of variant info.
For a primary code 0..p-1,
this is the number of secondary and tertiary variants,
and the code of the first of the (sec+1)*(ter+1) variants - 1.

For a variant code p..p+v-1, this is it's secondary and tertiary weights
and the code of the associated primary.
*/

typedef union {
	unsigned u;
	struct {
#ifdef CPU_BIG_ENDIAN
		unsigned char hi;
		unsigned char c[3];
#else
		unsigned char c[3];
		unsigned char hi;
#endif
	} b;
} Cv;

typedef struct Var {	/* variant info */
	unsigned char  sec;	/* secondary variant  */
	unsigned char  ter;	/* tertiary variant */
	unsigned short rel;	/* code of related variant */
} Var;

/*
For the encoding, we map every byte to:
-	it's code number
-	the "word" bit
-	the index of a table of possible following bytes

To cover the BMP with UTF-8, we need 1072 tables:
- 32 bytes 110* initiating a 2 byte sequence,
	using one table each
- 16 bytes 1110* initiating a 3 byte sequence,
	each having a table with 64 second bytes 10*,
	using 16+16*64 = 1040 tables
	(- the 32 tables used for the UTF-16 surrogates D800-DFFF)
	CJK 3400-9FFF alone uses 432 tables (some 27.000 ideographs)
*/
typedef struct Bin { /* byte info */
	unsigned short cod; /* byte's collation code */
	unsigned short tab; /* table */
} Bin;

typedef Bin Bins [256]; /* full table */

typedef struct Tab { /* table info */
	unsigned       off; /* bytes offset of Bin for min */
	unsigned char  min;
	unsigned char  max;
	unsigned short unu;
} Tab;


enum {
	BIT_VARIANTS = 0x20,
	BIT_FRENCH   = 0x10,
	TYP_PLAIN    = 0,
	TYP_VARIANTS = BIT_VARIANTS, /* TODO */
	TYP_FRENCH   = BIT_VARIANTS|BIT_FRENCH, /* TODO */
	CVLAT1 = 104 /* # primary codes for builtin lat1cdx */
};

/* header of a dumped/mapped cx */
struct Cdx {
	unsigned char  mag[3]; /* magic MCX or mcx for mapped cdx */
	unsigned char  typ; /* base type | bits per (primary) code-1 */
	/* currently only 8(7) and 16(15) bits supported */
	unsigned short pri;	/* p # primary codes incl. 0,1 */
	unsigned short var;	/* p + v # variants */
	unsigned short map;	/* p+v + m # maps */
	unsigned short tab;	/* # non-root tables 1..tab */
	/* redundant offsets */
	unsigned       ovi; /* offset variant_infos */
	unsigned       otp; /* offset table_pointers - 4 */
	unsigned       siz; /* total size */
	Bins           bt0; /* root table, always full */
	unsigned       cv[CVLAT1];	/* actually Cv code_values[map=p+v+m] */
	/* Var      variant_infos[var=p+v], if BIT_VARIANTS */
	/* Tab      table_pointers[tab] */
	/* Bin      byte_tables ... */
	/* unsigned char *more_bytes */
};


static const char MAGIC[3] =
#ifdef CPU_BIG_ENDIAN
	{'M','C','X'};
#	define W(b) (0x01000000|(b)<<16)
#	define N(b) (0x81000000|(b)<<16)
#else
	{'m','c','x'};
#	define W(b) (0x01000000|(b))
#	define N(b) (0x81000000|(b))
#endif

#if 0
const Cdx lat1cdx = {
	{'s','t','a'}, /* mark as static */
	TYP_PLAIN|7, CVLAT1, CVLAT1, CVLAT1, 0,0,0,0,
	{ /* byte infos */
#define B(b) {b,0}
		/* 32 C0 controls */
		B(0),B(0),B(0),B(0),B(0),B(0),B(0),B(0),
		B(0),B(0),B(0),B(0),B(0),B(0),B(0),B(0),
		B(0),B(0),B(0),B(0),B(0),B(0),B(0),B(0),
		B(0),B(0),B(0),B(0),B(0),B(0),B(0),B(0),
		/*       !     "     #     $     %     &     ' */
		B( 2),B( 3),B( 4),B( 5),B( 6),B( 7),B( 8),B( 9),
		/* (     )     *     +     ,     -     .     / */
		B(10),B(11),B(12),B(13),B(14),B(15),B(16),B(17),
		/* 0     1     2     3     4     5     6     7 */
		B(18),B(19),B(20),B(21),B(22),B(23),B(24),B(25),
		/* 8     9     :     ;     <     =     >     ? */
		B(26),B(27),B(28),B(29),B(30),B(31),B(32),B(33),
		/* @     A     B     C     D     E     F     G */
		B(34),B(35),B(36),B(37),B(38),B(39),B(40),B(41),
		B(42),B(43),B(44),B(45),B(46),B(47),B(48),B(49),	/* H-O */
		B(50),B(51),B(52),B(53),B(54),B(55),B(56),B(57),	/* P-W */
		/* X     Y     Z     [     \     ]     ^     _ */
		B(58),B(59),B(60),B(61),B(62),B(63),B(64),B(65),
		/* `     a     b     c     d     e     f     g */
		B(66),B(35),B(36),B(37),B(38),B(39),B(40),B(41),
		B(42),B(43),B(44),B(45),B(46),B(47),B(48),B(49),	/* h-o */
		B(50),B(51),B(52),B(53),B(54),B(55),B(56),B(57),	/* p-w */
		/* x     y     z     {     |     }     ~  DEL */
		B(58),B(59),B(60),B(67),B(68),B(69),B(70),B(0),
		/* 32 C1 controls */
		B(0),B(0),B(0),B(0),B(0),B(0),B(0),B(0),
		B(0),B(0),B(0),B(0),B(0),B(0),B(0),B(0),
		B(0),B(0),B(0),B(0),B(0),B(0),B(0),B(0),
		B(0),B(0),B(0),B(0),B(0),B(0),B(0),B(0),
		/* 32 mostly symbols */
		B( 2),B(71),B(72),B(73),B(74),B(75),B(76),B(77),
		B(78),B(79),B(80),B(81),B(82),B(83),B(84),B(85),
		B(86),B(87),B(88),B(89),B(90),B(91),B(92),B(93),
		B(94),B(95),B(96),B(97),B(98),B(99),B(100),B(101),
		/* 64 Latin alphas including 2 symbols */
		B(35),B(35),B(35),B(35),B(35),B(35),B(35),B(37),	/* 7A,1C */
		B(39),B(39),B(39),B(39),B(43),B(43),B(43),B(43),	/* 4E,4I */
		B(54),B(49),B(49),B(49),B(49),B(49),B(49),B(102),	/* TN5O* */
		B(49),B(55),B(55),B(55),B(55),B(59),B(54),B(53),	/* O4UYTS */
		B(35),B(35),B(35),B(35),B(35),B(35),B(35),B(37),	/* 7a,1c */
		B(39),B(39),B(39),B(39),B(43),B(43),B(43),B(43),	/* 4e,4i */
		B(54),B(49),B(49),B(49),B(49),B(49),B(49),B(103),	/* tn5o% */
		B(49),B(55),B(55),B(55),B(55),B(59),B(54),B(59) 	/* o4uyty */
#undef B
	},
	{	/* code values, all using immediate bytes */
		0, N(9),
		N(32),N(33),N(34),N(35),N(36),N(37),N(38),N(39), /*  !"#$%&' */
		N(40),N(41),N(42),N(43),N(44),N(45),N(46),N(47), /* ()*+,-./ */
		W(48),W(49),W(50),W(51),W(52),W(53),W(54),W(55), /* 01234567 */
		W(56),W(57),N(58),N(59),N(60),N(61),N(62),N(63), /* 89:;<=>? */
		N(64),W(65),W(66),W(67),W(68),W(69),W(70),W(71), /* @ABCDEFG */
		W(72),W(73),W(74),W(75),W(76),W(77),W(78),W(79), /* HIJKLMNO */
		W(80),W(81),W(82),W(83),W(84),W(85),W(86),W(87), /* PQRSTUWW */
		W(88),W(89),W(90),N(91),N(92),N(93),N(94),N(95), /* XYZ[\]^_ */
		N(96), /*a-z*/ N(123),N(124),N(125),N(126),      /* `  {|}~  */
		       N(161),N(162),N(163),N(164),N(165),N(166),N(167),
		N(168),N(169),N(170),N(171),N(172),N(173),N(174),N(175),
		N(176),N(177),N(178),N(179),N(180),N(181),N(182),N(183),
		N(184),N(185),N(186),N(187),N(188),N(189),N(190),N(191),
		N(215), N(247)
	}
}; /* lat1cdx */
#endif


/*
	encode l byte in b to key.
	key->len is the max len on input, resulting len on return.
	stop if max key len is hit or on word boundary, if words.
	return #used bytes in b.
*/
int cEnc ( const Cdx *cdx, Key *key, unsigned char *b, int l, int words )
{
	const unsigned char * const base = (unsigned char *)cdx;
	const Bin * const bt0 = cdx->bt0;
	const Tab * const tp = (Tab*)(base + cdx->otp);
	const int bits = 1+(0xf&cdx->typ), mapcodeshift = 256<cdx->var ? 1 : 0;
	const unsigned char *e = b+l;
	int bitsleft = 8*key->len, pfxlen = 0;
	int unassigned = 0;
	Bin seq[CDX_MAXSEQ], *top; /* current sequence */
	unsigned u, len;
	unsigned short code;
	Cv cv;

	key->len = 0;
	for ( ; b < e && bits <= bitsleft; b++ ) {
		eRr(LOG_TRACE, "byte %c of %d bits %d/%d", *b, e-b, bitsleft, bits);
		/* traverse byte info tables to find longest matching sequence */
		for (*(top = seq) = bt0[*b]; top->tab && b < e; ) {
			const Tab * const t = tp + top->tab;
			unsigned char n = b[1];
			LOG_DBG(LOG_DEBUG, "table %d check %d %d-%d off %d %d",
				top->tab, n, t->min, t->max, t->off, *(int*)(base+t->off));
			if ( n < t->min || n > t->max	/* out of bounds */
				|| !(u = ((unsigned*)(base+t->off))[n - t->min])	/* unassigned */
			)
				break;
			*++top = *(Bin*)&u;
			b++;
		}
		for ( ; top > seq && !top->cod; top-- ) /* no complete sequence */
			b--;
		code = top->cod;
		eRr(LOG_TRACE, "code %d", code);
		if (words) {
			if ( 1<words /* leave 1st words-1 bytes alone */
				&& b+l-e >= words /* ok, passed it for the 1st time */
			) {
				pfxlen = key->len;
				words = 1;
			} else if (0x80000000&cdx->cv[code]) {	/* hit nonword */
				if (pfxlen == key->len) /* skip to word */
					continue;
				break; /* had something after prefix */
			}
		}
		if (cdx->pri > code) {
			if ( code )
				unassigned = 0;
			else {	/* unassigned */
				if ( unassigned )
					continue;
				unassigned = 1;
				code = 1;
			}
			switch (bits) {
			case 16: key->byt[key->len++] = code >> 8; /* always bigend */
			case  8: key->byt[key->len++] = (char)code;
			}
			bitsleft -= bits;
			continue;
		}
		/* if (cdx->var > top->cod) variant */
		/* else map: */
		cv.u = cdx->cv[code];
		if ((len = (0xf & cv.b.hi)<<mapcodeshift)) { /* len = #codes */
			/* mapped variants TODO */
			if ( 0 > (bitsleft -= (len << 3)) )
				break;
			memcpy(key->byt+key->len, ~3&len ? base+(0xffffff&cv.u) : cv.b.c, len);
			key->len += len;
		}
	}
	if (/*words &&*/ key->len == pfxlen) /* if pfxlen: found no words */
		key->len = 0;

	return b+l-e;
}	/* cEnc */


int cDec ( const Cdx *cdx, unsigned char *b, int l, Key *key )
{
	const unsigned char * const base = (unsigned char *)cdx;
	const unsigned char *k = key->byt, *e = b+l-CDX_MAXSEQ;
	const int bits = 1+(0xf&cdx->typ);
	int bitsleft = 8*key->len;
	Cv cv;

	for ( ; b < e && bits <= bitsleft; bitsleft -= bits ) {
		unsigned short v = *k++;
		if ( 16 == bits )
			v = v<<8 | *k++;
		cv.u = cdx->cv[v];
		switch ( cv.b.hi &= 0xf ) { /* discard word bit */
		case 1: *b++ = cv.b.c[0]; continue;
		case 2: *b++ = cv.b.c[0]; *b++ = cv.b.c[1]; continue;
		case 3: *b++ = cv.b.c[0]; *b++ = cv.b.c[1]; *b++ = cv.b.c[2]; continue;
		}
		memcpy(b, base+(cv.u&0xffffff), cv.b.hi);
		b += cv.b.hi;
	}
	return b+l-CDX_MAXSEQ-e;
}	/* cDec */



typedef struct {
	Cdx      cdx;
	unsigned cvb[0x10000-CVLAT1];
	Var      *vi, vib[0x10000];
	Tab      tp[1100];
	Bins     *bt, btb[1100];
	unsigned char *p, b[200000];
} CdxMake;


/* make byte table entries */
static int mapSeq (CdxMake *mk, unsigned char *p, int len,
	unsigned short code, int save)
{
	Bins  *bt = &mk->cdx.bt0;
	LOG_DBG(LOG_DEBUG, "map '%.*s' -> %d", len, p, code);
	if (save) {
		Cv *cv = (Cv*)mk->cdx.cv+code;
		if (15 < len) {
			eRr(ERR_INVAL, "sequence '%.*s' too long %d", len, p, len);
			return 1;
		}
		cv->b.hi = (0x80&save) | len;
		if (!(~3 & len))
			memcpy(cv->b.c, p, len);
		else {
			cv->u |= mk->p - mk->b;
			memcpy(mk->p, p, len);
			mk->p += len;
		}
	}
	for (;; p++) {
		Bin *b = *bt+*p;
		if (!--len) { /* last byte -- assign code */
			if (b->cod) {
				eRr(ERR_INVAL, "attempt to reassign code %d for %d", code, b->cod);
				return 1;
			}
			b->cod = code;
			return 0;
		}
		if (b->tab)
			bt = mk->btb + (b->tab - 1);
		else {
			bt = mk->bt++;
			b->tab = mk->bt - mk->btb;
			LOG_DBG(LOG_DEBUG, "new table %d", b->tab);
		}
	}
	return 0;
}

/** compile collation src */
static Cdx *cMake ( const Fld *src )
{
	const Fld *eof = REND(src), *f;
	CdxMake mk;
	int skipalias = 1, mapcodeshift;
	unsigned short firstcode = 1; /* of last row */
	unsigned bins=0, off, u;
	Cdx *cdx;

	LOG_DBG(LOG_DEBUG, "cMake %d", sizeof mk);
	memset(&mk, 0, sizeof mk);
	mk.vi = mk.vib;
	mk.bt = mk.btb;
	mk.p  = mk.b;
	mk.cdx.cv[0] = 0x80000000; /* 0: unassigned */
	mk.cdx.cv[1] = 0x81000009; /* 1: tab */
	mk.cdx.pri = 2;

	/* first pass: gather primaries and aliases */
	for (f = src; ++f<eof; ) if (MET_COL == f->tag) {
		unsigned char *p = (unsigned char*)f->val, *e, *t, save;
		unsigned short code;

		if (3 > f->len || TAB != p[1]) continue;
		e = p + f->len;
		switch (*p) {
		case 'A': if (skipalias) continue; save = 0; goto alias;
		case 'W': save = 1; break;
		case 'N': save = 0x80; break;
		case 'S':
		case 'T':
			skipalias = 1;
		default: continue;
		}
		firstcode = mk.cdx.pri;
		skipalias = 0;
	alias:
		code = firstcode;
		for (p+=2; e > p; p = t+1) {
			if (!(t = memchr(p, TAB, e-p))) t = e;
			if (t > p && mapSeq(&mk, p, t-p, code, save)) return 0;
			if (save || code < mk.cdx.pri-1) code++;
		}
		if (save) mk.cdx.pri = code;
	}
	mk.cdx.typ = TYP_PLAIN|(256 < mk.cdx.pri ? 15 : 7);

	mk.cdx.var = mk.cdx.pri; /* second pass: gather variants */
	for (f = src; ++f<eof; ) if (MET_COL == f->tag) {
		unsigned char *p = (unsigned char*)f->val;

		if (3 > f->len || TAB != p[1]) continue;
		switch (*p) {
		case 'S':
		case 'T':
			eRr(ERR_INVAL, "this version does not support multilevel");
			return 0;
		}
	}
	if (mk.cdx.var > mk.cdx.pri)
		mk.cdx.typ |= BIT_VARIANTS;

	mk.cdx.map = mk.cdx.var; /* third pass: resolve maps */
	mapcodeshift = 256<mk.cdx.var ? 1 : 0;
	for (f = src; ++f<eof; ) if (MET_COL == f->tag) {
		unsigned short code, codes[CDX_MAXSEQ];
		unsigned char ccodes[CDX_MAXSEQ], *pcodes;
		unsigned char *p = (unsigned char*)f->val, *e, *t;
		unsigned short n = 0;
		Bin seq[CDX_MAXSEQ], *bin; /* current sequence */
		Cv *cv;

		if (3 > f->len || 'M' != p[0] || TAB != p[1]) continue;
		e = p + f->len;
		p += 2;
		if (!(t = memchr(p, TAB, e-p))) continue;

		for ( ; p<t && n<CDX_MAXSEQ; p++ ) { /* get n codes for map */
			for (*(bin = seq) = mk.cdx.bt0[*p]; bin->tab && p < t;) {
				/* slightly simpler than in cEnc, since we have full tables */
				int *ent = (int*)mk.btb[bin->tab-1];
				if (!ent[p[1]]) break; /* unassigned */
				*++bin = *(Bin*)(ent + *++p);
				if (seq+CDX_MAXSEQ == bin) break;
			}
			for (; bin>seq && (!bin->cod || mk.cdx.var<=bin->cod); bin--)
				p--;
			if (mk.cdx.var>bin->cod) /* ignore map codes */
				codes[n++] = bin->cod ? bin->cod : 1;
		}
		p = t+1;

		/* got n codes ... */
		if (1 == n) /* single code ~ alias */
			code = codes[0];
		else { /* make entry cv[mk.cdx.map++] (even for n=0) */
			if (mapcodeshift) {
				pcodes = (unsigned char*)codes; /* native */
#ifndef CPU_BIG_ENDIAN
				{ unsigned short *us = codes+n;
					while (us-- > codes) *us = (*us>>8) | (*us<<8); /* mk be */
				}
#endif
			} else {
				for (u=n; u--;) ccodes[u] = codes[u];
				pcodes = ccodes;
			}
			cv = (Cv*)mk.cdx.cv + mk.cdx.map;
			cv->b.hi = n;
			if (n && 0x80000000&mk.cdx.cv[codes[0]]) /* inherit from 1st code */
				cv->b.hi |= 0x80;
			n <<= mapcodeshift;
			if (!(~3 & n))
				memcpy(cv->b.c, pcodes, n);
			else {
				cv->u |= mk.p - mk.b;
				memcpy(mk.p, pcodes, n);
				mk.p += n;
			}
			code = mk.cdx.map++;
		}
		for (; e > p; p = t+1) { /* map 'em all to mk.cdx.map */
			if (!(t = memchr(p, TAB, e-p))) t = e;
			if (t > p && mapSeq(&mk, p, t-p, code, 0)) return 0;
		}
	}
	/* finish: compact */
	mk.cdx.tab = mk.bt - mk.btb;
	mk.cdx.otp =
	mk.cdx.ovi = (((char *)&((Cdx*)0)->cv) - (char*)0) /* offsetoff(cv) */
		+ mk.cdx.map*sizeof (unsigned);
	if (mk.cdx.var > mk.cdx.pri)
		mk.cdx.otp += mk.cdx.var * sizeof (Var);
	off = mk.cdx.otp + mk.cdx.tab * sizeof (Tab);
	for (u=0; u<mk.cdx.tab; u++) /* compact tables */ {
		int *base = (int *)(mk.btb + u), *pi = base;
		mk.tp[u].off = off;
		while (!*pi) pi++;
		LOG_DBG(LOG_DEBUG, "found code %x", *pi);
		mk.tp[u].min = pi - base;
		for (pi = base+256; !*--pi; ) ;
		mk.tp[u].max = pi - base;
		bins += 1+mk.tp[u].max-mk.tp[u].min;
		off += (1+mk.tp[u].max-mk.tp[u].min)*sizeof (Bin);
		LOG_DBG(LOG_DEBUG, "table %d %d-%d", u+1, mk.tp[u].min, mk.tp[u].max);
	}
	mk.cdx.siz = off + (mk.p - mk.b);
	eRr(LOG_INFO,
		"%d primaries %d variants %d maps %d tables %d bins, off %d size %d",
		mk.cdx.pri, mk.cdx.var-mk.cdx.pri, mk.cdx.map-mk.cdx.var, mk.cdx.tab,
		bins, off, mk.cdx.siz);
	/* adjust indirect CVs by off */ {
		unsigned i = 0, mask = 0x0c000000;
		for (; i < mk.cdx.var; i++)
			if (0x0c000000 & mk.cdx.cv[i])
				mk.cdx.cv[i] += off;
		if (mapcodeshift) mask |= 0x02000000; /* only 1 code fits */
		for (; i < mk.cdx.map; i++)
			if (mask & mk.cdx.cv[i])
				mk.cdx.cv[i] += off;
	}
	cdx = (Cdx*)mAlloc(mk.cdx.siz);
	memcpy(cdx, &mk.cdx, mk.cdx.ovi);
	if (cdx->var > cdx->pri)
		memcpy((char*)cdx + cdx->ovi, mk.vib, mk.cdx.var*sizeof (Var));
	if (cdx->tab) {
		memcpy((char*)cdx + cdx->otp, mk.tp, mk.cdx.tab*sizeof (Tab));
		for (u=0; u<mk.cdx.tab; u++)
			memcpy((char*)cdx + mk.tp[u].off,
				mk.btb[u] + mk.tp[u].min,
				(1+mk.tp[u].max-mk.tp[u].min)*sizeof (Bin));
	}
	memcpy((char*)cdx + off, mk.b, mk.p - mk.b);
	cdx->otp -= sizeof (Tab); /* adjust 0 based */

	return cdx;
}	/* cMake */



/*
	list of open shared cdx
*/
typedef struct Foo Foo;

struct Foo {
	char nln; /* namelen */
	char nam[31];
	Foo *foo;
	int  ref;
	Cdx *cdx;
	FMap fm;
};
#if 0
static Foo lat1foo = { 7, "Latin-1", 0, 1, (Cdx*)&lat1cdx };
static Foo *fools = &lat1foo;
#endif
static Foo *fools = 0;



/** open or compile collation src */
const Cdx *cOpen (const Fld *src)
{
	char *nam = 0, *p;
	int nln = 0, coldef = 0;
	Foo *foo = fools;
	Cdx *cdx = 0;
	const Fld *eof, *f;
	char fname[20];
	FMap fm;

	if (src)
		for (f = src, eof = REND(src); ++f < eof; )
			if (MET_COL == f->tag) {
				coldef = 1;
				if (2 < f->len && 'C' == f->val[0]
					&& TAB == f->val[1] && TAB != f->val[2]
				) { /* named */
					nam = f->val+2;
					nln = f->len-2;
					if ((p = memchr(nam, TAB, nln))) nln = p-nam;
					if (nln > 15) nln = 15;
					LOG_DBG(LOG_DEBUG, "collation name '%.*s'", nln, nam);
					for (; foo; foo = foo->foo)
						if (nln == foo->nln && !memcmp(nam, foo->nam, nln)) { /* got it */
							foo->ref++; /* ref it */
							return foo->cdx; /* ret it */
						}
					/* TODO: try to map -- check type */
					memset(fname, 0, sizeof fname);
					memcpy(fname, nam, nln);
					memcpy(fname+nln, ".mcx", 5);
					memset(&fm, 0, sizeof fm);
					fm.fil = FIL_NONE;
					if (!fMOpen(&fm, fname, FIL_RD)) {
						int size = fSize(fm.fil);
						fm.lim = (size + env.psz-1)>>env.psh;
						if ( (int)fm.lim != fMap(&fm, fm.lim)
							|| memcmp(MAGIC, fm.map, 3)
							|| size != (int)((Cdx*)fm.map)->siz
						) {
							eRr(ERR_TRASH, "bad coll file '%s'", fname);
							fMClose(&fm);
						} else {
							cdx = (Cdx*)fm.map;
							eRr(LOG_VERBOSE, "mapped coll '%s' %d bytes", fname, size);
#ifndef WIN32
							fClose(&fm.fil); /* don't need the file open */
#endif
						}
					}
				}
			}
	if (!cdx && (!coldef || !(cdx = cMake(src))))
		return 0;
	if (nam) { /* foo an mmaped or named made */
		Foo *newfoo = mAlloc(sizeof *newfoo);
		memcpy(newfoo->nam, nam, newfoo->nln = nln);
		newfoo->foo = fools;
		newfoo->ref = 1;
		newfoo->cdx = cdx;
		if (fm.map) /* was mapped */
			newfoo->fm = fm;
		else { /* try to write */
			file fil = FIL_NONE;
			if (!fOpen(&fil, fname, FIL_WR|FIL_CREAT|FIL_TRUNC)) {
				memcpy(cdx->mag, MAGIC, 3);
				fWrite(&fil, cdx, cdx->siz);
				fClose(&fil);
				eRr(LOG_INFO, "saved coll '%s' %d bytes", fname, cdx->siz);
			}
			memcpy(cdx->mag, "nam", 3); /* yet we keep using our selfmade copy */
		}
		fools = newfoo;
	}
	return cdx;
}	/* cOpen */


void cClose (Cdx *cdx)
{
	Foo *foo, *f = 0;
	switch (cdx->mag[0]) {
	case 0: /* internal anonymous */
		mFree(cdx);
	case 's': /* internal static */
		return;
#ifdef CPU_BIG_ENDIAN
	case 'M':
#else
	case 'm': /* mapped - always named */
#endif
	case 'n': /* internal named */
		for (foo = fools; foo; foo = (f = foo)->foo)
			if (foo->cdx == cdx) {
				if (!--foo->ref) {
					if ('n' == cdx->mag[0])
						mFree(cdx);
					else
						fMClose(&foo->fm);
					if (f)
						f->foo = foo->foo;
					else
						fools = foo->foo;
					mFree(foo);
				}
				return;
			}
		/* panic time ? */
	}
}	/* cClose */
