/*
	The Malete project - the Z39.2/Z39.50 database framework of OpenIsis.
	Version 0.9.x (patchlevel see file Version)
	Copyright (C) 2001-2003 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	see README for more information
EOH */

/*
	$Id: db.c,v 1.1 2005/01/24 17:59:47 kripke Exp $
	implementation of general db access functions.
*/


#include "core.h"


/* ************************************************************
	private types
*/

/** file extensions not listed on filext.com */

/** B-Link-Tree malete query data/index */
static const char EXT_MQD[] = ".mqd";
static const char EXT_MQX[] = ".mqx";
/** plaintext malete record data/index */
static const char EXT_MRD[] = ".mrd";
static const char EXT_MRX[] = ".mrx";
/** record 0 options data */
static const char EXT_M0D[] = ".m0d";



/* ************************************************************
	data
*/

/* linked list of open dbs. */
static Db *dbs;


enum {
	/* additional flg outside the short range */
	OPEN_TRY = 0x10000, /* try writable, open readonly else */
	OPEN_UC  = 0x20000, /* use uppercase ext */
	/* commonly used combinations */
	/* 1) open as is, do not complain about any failure, do not create */
	OPEN_ASIS  = FIL_RDWR|OPEN_TRY,
	/* 2) open readonly, do not complain about any failure, do not create */
	OPEN_RDIF  = FIL_RD|FIL_TRY,
	/* 3) open readonly, complain about any failure */
	OPEN_RD    = FIL_RD,
	/* 4) open or create writable, complain on failure */
	OPEN_NEW   = FIL_RDWR|FIL_CREAT,
	OPEN_BLANK = FIL_RDWR|FIL_CREAT|FIL_TRUNC
};


/* set extension. fname MUST already end with .xxx.
*/
static char *setext ( char *fname, const char *ext )
{
	int l = strlen( fname ) - 4;
	memcpy( fname+l, ext, 4 );
	return fname;
}


Db *dOpen (const char *dbname)
{
	const Fld *cfg;
	Fld opt;
	file optf;
	/* options during open */
	int withrdx = !0;
	int withqdx = !0;
	int filmode = FIL_RD, async = DX_ASYNC; /* changed by -w */
	int lck = FIL_TLOCK; /* set to BLOCK by -b */
	/* misc vars */
	Db *db;
	int plen, ret = 0;

#ifdef BUILD_SHMODE
	if (ENV_SHARED==env.wri) async = 0; /* default to sync */
#endif
	for ( db = dbs; db; db = db->nxt )
		if ( !strcmp(db->nam, dbname) ) {
			if ( DX_OPEN & db->flg ) {
				db->mnt++;
				return db;
			}
			return 0; /* hmmm ... should not be here */
		}

	cfg = rKey(env.opt->fld, 1, dbname);
	db = mAlloc(sizeof(Db));
	db->rdx.mrd = db->rdx.mrx.fil = FIL_NONE;
	db->rdx.mrx.lim = env.rml;
	db->qdx.mqd = db->qdx.mqx.fil = FIL_NONE;
	db->qdx.mqx.lim = env.qml;
	plen = strlen(dbname);
	memcpy(db->nam = mAlloc(plen+1), dbname, plen+1);

	/* pre-options defaults */
	db->qdx.typ = QDX_LEAF1K;
	db->qdx.ptr = QDX_ISIS;
	db->qdx.ksz = 255;
	if ( env.wri )
		db->flg |= DX_WRITE;

	if ( cfg ) for ( opt.val = 0; vGet(&opt, cfg, "bfqrw"); )
		switch (opt.tag) {
		case 'b': /* blocking */
			lck = FIL_BLOCK;
			continue;
		case 'f': /* file/path */
			if ( !opt.len ) {
				eRr(LOG_WARN, "bad empty file");
				continue;
			}
			db->pat = mAlloc(opt.len + plen + 5);
			memcpy(db->pat, opt.val, opt.len);
			switch ( opt.val[opt.len - 1] ) {
			case '/':
#ifdef WIN32
			case '\\':
#endif
				memcpy(db->pat+opt.len, dbname, plen);
				plen += opt.len;
				break;
			default:
				plen = opt.len;
			}
			eRr(LOG_INFO, "using path %.*s", plen, db->pat);
			continue;
		case 'q': /* qdx type */
			if ( !opt.len )
				withqdx = 0;
			else {
				unsigned char val = (unsigned char) a2i(opt.val+1, opt.len-1);
				switch (*opt.val) {
				case 'k': db->qdx.ksz = val; break;
				case 'l': db->qdx.let = val; break;
				case 'p': db->qdx.ptr = val; break;
				case 't': db->qdx.typ = val; break;
				}
			}
			continue;
		case 'r': /* mrx type */
			if ( !opt.len )
				withrdx = 0;
			else
				db->rdx.typ = a2i(opt.val, opt.len);
			continue;
		case 'w': /* writable */
			if ( opt.len )
				switch ( *opt.val ) {
				case '0': db->flg &= ~DX_WRITE; break;
				case 'a': async = DX_ASYNC; break;
				case 's': async = 0; break;
				}
			continue;
		}

	/* post-options defaults */
	if ( DX_WRITE & db->flg ) {
		db->rdx.flg |= DX_WRITE;
		db->qdx.flg |= DX_WRITE;
		filmode |= FIL_WR|FIL_CREAT;
		if ( !(DX_ASYNC & (db->flg |= async)) )
			filmode |= FIL_SYNC;
	}
	if ( !db->pat )
		memcpy(db->pat = mAlloc(plen+5), dbname, plen);
	memcpy(db->pat+plen, ".???", 5);

#if 1
	/* check options file */
	if ( ! fOpen(&optf, setext(db->pat,EXT_M0D), FIL_RD|FIL_TRY) ) {
		char buf[0x10000];
		int sz;
		if ( 0 < (sz = fSize(optf)) ) {
			char *p = sz < (int)sizeof(buf) ? buf : mAlloc(sz);
			eRr(LOG_INFO, "reading %d bytes options from '%s'", sz, db->pat);
			if ( sz == fRead(&optf, p, sz) ) {
				List dbopt;
				lParse(lInit(&dbopt,0), p, sz);
				db->opt = LCAN(&dbopt);
			}
			if ( buf != p )
				mFree(p);
		}
		fClose(&optf);
	}
#endif

#ifdef BUILD_SHMODE
	if (ENV_SHARED==env.wri) lck = 0; /* we lock on demand */
#endif
#define CLEANUP( args ) do { ret = eRr args; goto cleanup; } while (0)
	if ( withrdx ) {
		if ( fOpen( &db->rdx.mrd, setext(db->pat,EXT_MRD), filmode|lck ) )
			CLEANUP((LOG_ERROR, "no master file '%s'", db->pat ));
		if ( fMOpen( &db->rdx.mrx, setext(db->pat,EXT_MRX), filmode ) )
			CLEANUP((LOG_ERROR, "no pointer file '%s'", db->pat ));
		if ( rInit( &db->rdx ) )
			CLEANUP((LOG_ERROR, "could not rInit '%s'", db->pat ));
	}

	if ( withqdx ) {
		if (db->opt)
			db->qdx.cdx = cOpen(db->opt);
		if ( fOpen( &db->qdx.mqd, setext(db->pat,EXT_MQD), filmode|lck ) )
			CLEANUP((LOG_ERROR, "no index leaves file '%s'", db->pat ));
		if ( fMOpen( &db->qdx.mqx, setext(db->pat,EXT_MQX), filmode ) )
			CLEANUP((LOG_ERROR, "no index tree file '%s'", db->pat ));
		if ( qInit( &db->qdx ) )
			CLEANUP((LOG_ERROR, "could not qInit '%s'", db->pat ));
	}

	db->pat[plen] = 0; /* cut .ext */
	db->flg |= DX_OPEN;
	db->nxt = dbs;
	dbs = db;
	return db;

cleanup:
	/* cleanup ... */
	db->flg = 0;
	fClose( &db->qdx.mqd );
	fMClose( &db->qdx.mqx );
	fClose( &db->rdx.mrd );
	fMClose( &db->rdx.mrx );
	return 0;
}	/* dOpen */


void dClose ( Db *db )
{
	Db **d;

	for ( d = &dbs; *d; d = &(*d)->nxt ) /* unlink */
		if ( db == *d ) {
			if ( --db->mnt )
				return;
			*d = db->nxt;
			break;
		}
	if ( FIL_NONE != db->rdx.mrd ) {
		rFini( &db->rdx );
		fClose( &db->rdx.mrd );
		fMClose( &db->rdx.mrx );
	}
	if ( FIL_NONE != db->qdx.mqd ) {
		qFini( &db->qdx );
		fClose( &db->qdx.mqd );
		fMClose( &db->qdx.mqx );
	}
	db->flg = 0L;
	mFree( (void*)db->pat );
	mFree( (void*)db->opt );
	mFree( (void*)db->nam );
	memset( db, 0, sizeof(db) );
	mFree( db );
}	/* dClose */


void dCloseAll ()
{
	Db *d, *nxt;
	for ( d = dbs; d; d = nxt ) {
		nxt = d->nxt;
		dClose(d);
	}
}	/* dCloseAll */
