/*
	The Malete project - the Z39.2/Z39.50 database framework of OpenIsis.
	Version 0.9.x (patchlevel see file Version)
	Copyright (C) 2001-2003 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	see README for more information
EOH */

/*
	$Id: hash.c,v 1.1 2005/01/24 18:00:20 kripke Exp $
	the hashtable
*/

#include "pw.h"

/* ************************************************************
		private functions
*/

/*
	add a bucket
*/
#define HLNEWF( l, t ) do { \
	if ( (l)->fav || lExtend( l, 0, 1 ) ) { \
		Fld *__f = (l)->fld - (l)->fld->tag--; \
		__f->tag = t; \
		__f->val = 0; \
		__f->len = 0; \
		(l)->fav --; \
	} } while(0)

/*
	the cdb hash function ( without h/256 )
*/
#define CDB_CALC(i,s) i ^= ( i += i << 5, (int) *s++ )

static OPT_INLINE int hs ( const char *s, register int len )
{
	register int i = 5381, n = (len + 7 ) / 8;
	if ( ! len ) return i;
	switch ( len % 8 ) {
	case 0: do { CDB_CALC(i,s);
	case 7:      CDB_CALC(i,s);
	case 6:      CDB_CALC(i,s);
	case 5:      CDB_CALC(i,s);
	case 4:      CDB_CALC(i,s);
	case 3:      CDB_CALC(i,s);
	case 2:      CDB_CALC(i,s);
	case 1:      CDB_CALC(i,s);
	        } while ( --n );
	}
	return i;
} /* hs */

/*
	lookup key in a bucket, if found the corresponding entry is assigned
	to res
*/
static OPT_INLINE
HEntry * gethe ( Fld *fld, int hv, const char *key, int len )
{
	int i;
	register const char *p, *q;
	HEntry *he;
	if ( fld->len ) {
		for ( he = (HEntry *) fld->val; he; he=he->nxt ) {
			if ( hv != he->hv || ( i = len ) != he->len ) continue;
			for ( p = key, q = he->key, --i; *p++ == *q++; --i ) {
				if ( !i ) {
					return he;
				}
			}
		}
	}
	return 0;
} /* gethe */

/*
	rebuild the table. this is done if:
		1. the load factor is > HLOADMAX, or
		2. the fragmentation is > HMAXFRAG
	mode signals if the number of buckets has to be extended
	or to be reduced, this is done if less than HSHRINKLOAD of the buckets are
	used
*/
static void OPT_STDCALL hRebuild ( HTable *table, int mode )
{
	int i, nsiz, osiz = HSIZ(table);
	HEntry *ohe, *nhe;
	Fld *fld;
	List *nl = (List * )mAlloc(sizeof(List));
	HSTAT(table)
#ifndef NDEBUG
	table->coll = 0;
#endif
	if ( mode ) {
		nsiz = osiz << HSHIFT;
		table->mask |= ( table->mask <<= HSHIFT, 1 );
	} else if ( HSHRINKLOAD > HCURLOAD(table) && HINITIAL < HSIZ(table) ) {
		nsiz = (unsigned) osiz >> HSHIFT;
		table->mask = (unsigned) table->mask >> HSHIFT;
	} else
		nsiz = osiz;
	lInit( nl, "%.*s", table->list->fld->len, table->list->fld->val );
	lExtend( nl, 0, nsiz );
	for ( i = 0; nsiz > i; ++i )
		HLNEWF( nl, i+1 );
	for ( i = 1; osiz >= i; ++i ) {
		if ( ( fld = table->list->fld+i)->len ) {
			for ( ohe = (HEntry *) fld->val; ohe; ohe = ohe->nxt ) {
				int j = ohe->hv & table->mask;
				unsigned n = HESIZ(ohe);
				fld = nl->fld+j+1;
				if ( LAVL( nl ) >= n || lExtend( nl, n, -1 ) ) {
					nhe = (HEntry *) nl->buf;
					memcpy( nhe, ohe, n );
					nhe->nxt = 0;
#ifndef NDEBUG
					nhe->dpt = 1;
#endif
					if ( fld->len ) {
						nhe->nxt = (HEntry *) fld->val;
#ifndef NDEBUG
						nhe->dpt = nhe->nxt->dpt + 1;
						++table->coll;
#endif
					}
					fld->val = (char *) nhe;
					fld->len += n;
					nl->buf += n;
				}
			}
		}
	}
	table->dsize = 0;
	lFini( table->list );
	mFree( table->list );
	table->list = nl;
	HSTAT(table)
} /* hRebuild */

/* ************************************************************
		public functions
*/

/*
	get
*/
HEntry * OPT_STDCALL hGet ( HTable *table, const char *key, int len )
{
	int hv = hs( key, len);
	return gethe( (table)->list->fld+1+(hv & table->mask), hv, key, len );
} /* hGet */

/*
	add
*/
HEntry * OPT_STDCALL hAdd ( HTable *table, const char *key, int len )
{
	int hv = hs( key, len);
	Fld *fld = (table)->list->fld+1+(hv & table->mask);
	HEntry *he = gethe( fld, hv, key, len );
	if ( he ) {
#ifndef NDEBUG
		eRr( LOG_DEBUG, "already there %.*s <--> %.*s", he->len, he->key,
				len, key );
#endif
		return he;
	}
	/*
		create new entry
	*/
	{
		unsigned need = sizeof(HEntry) + len - HEDEFKEYLEN;
		if ( LAVL( table->list ) >= need
				|| lExtend( table->list, need, -1 ) ) {
			he = (HEntry *) table->list->buf;
			he->nxt = 0;
			he->obj = 0;
#ifndef NDEBUG
			he->dpt = 1;
#endif
			he->hv = hv;
			he->len = len;
			memcpy( he->key, key, len );
			if ( fld->len ) {
				he->nxt = (HEntry *) fld->val;
#ifndef NDEBUG
				++table->coll;
				he->dpt = he->nxt->dpt + 1;
#endif
			}
			fld->val = (char *) he;
			fld->len += need;
			table->list->buf += need;
			++table->nume;
			if ( HMAXLOAD < HCURLOAD(table) ) {
				hRebuild( table, 1 );
				return gethe ( (table)->list->fld+1+(hv & table->mask),
						hv, key, len );
			}
			return he;
		}
	}
	return 0;
} /* hAdd */


HEntry* OPT_STDCALL hSet ( HTable *t, const char *key, int len, Obj *o )
{
	HEntry *he = hAdd(t, key, len);
	if ( he )
		he->obj = o;
	return he;
}	/* hSet */


/*
	delete entry and check fragmentation
*/
int OPT_STDCALL hDel (  HTable *table, const char *key, int len )
{
	int hv = hs( key, len);
	Fld *fld = (table)->list->fld+1+(hv & table->mask);
	HEntry *he = gethe( fld, hv, key, len );
	if ( he ) {
		int siz = HESIZ(he);
		HEntry *prev = (HEntry *) fld->val;
		if ( prev == he ) {
			fld->val = (char *)he->nxt;
		} else {
			for ( ; prev->nxt != he; prev = prev->nxt );
			prev->nxt = he->nxt;
		}
#ifndef NDEBUG
		if ( fld->val ) {
			((HEntry *) fld->val)->dpt--;
			--table->coll;
		}
#endif
		fld->len -= siz;
		--table->nume;
		table->dsize += siz;
		if ( HMAXFRAG < HCURFRAG(table) )
			hRebuild( table, 0 );
		return 1;
	}
	return 0;
} /* hDel */

/*
	table initialisation
*/
HTable * OPT_STDCALL hInit ( HTable *table, const char* header )
{
	int i;
	table->list = (List * ) mAlloc( sizeof( *table->list ) );
	lInit( table->list, header );
	for ( i = 0; HINITIAL > i; ++i )
		HLNEWF( table->list, i+1 );
	table->mask = HINITIAL-1;
	table->nume = table->dsize = 0;
#ifndef NDEBUG
	table->coll = 0;
#endif
	return table;
} /* hInit */

/*
	free the list
	the table itself is not freed
*/
void OPT_STDCALL hFini ( HTable *table )
{
	lFini( table->list );
	mFree( table->list );
	table->mask = table->nume = table->dsize = 0;
#ifndef NDEBUG
	table->coll = 0;
#endif
} /* hFini */

#ifndef NDEBUG
/*
	hashtable statistics
*/
void OPT_STDCALL hstat ( HTable *table )
{
	int j, i, n =  HSIZ(table);
	HEntry *he;
	for ( i = 1, j = 0; n >= i; ++i ) {
		he = (HEntry *) (table->list->fld+i)->val;
		if ( he )
			j = he->dpt > j ?  he->dpt : j;
	}
	eRr( LOG_DEBUG, "hash stats:\n\tnum buckets: %u\n\tnum entries: %u\n\tload: %u\n\tlist buf size: %u\n\tdel size: %u\n\tfrag: %u\n\tcollisions: %u\n\tcollision rate: %u\n\tmaxdepth: %u", HSIZ(table), table->nume, HCURLOAD(table), LSIZ(table->list), table->dsize, HCURFRAG(table), table->coll, (table->coll*100)/( table->nume ? table->nume : 1 ), j );

} /* hstat */

#endif
