/*
	The Malete project - the Z39.2/Z39.50 database framework of OpenIsis.
	Version 0.9.x (patchlevel see file Version)
	Copyright (C) 2001-2004 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	see README for more information
EOH */

/*
	$Id: iif.c,v 1.1 2005/01/24 18:00:40 kripke Exp $
	malete IIF(ISO2709/Z39.2)/MARC conversion
	less general than IIF, but less restrictive than MARC
*/

#include "tool.h"

/*
implemented structure of IIF files:

RT, FT and delimiter as defined by Z39.2-1979
other chars as used by WinIsis (esp. delimiter ^) are NOT supported

the bad habit of WinIsis to insert \r\n every 80 chars is NOT supported

Z39.2 specifies that
tag 1 is for a control number, every record must have exactly one
tag 2 is for subrecord purposes
tags 3-9,a-z are other control fields
control field tags 001-00z must be first and sorted in dict and data,
	so base of data contains always the "control number" (not nec. a number)
control fields contain neither indicators nor subfields/delimiters, but FT

MARC specifies that
tag 5 is date and time of latest transaction
there are 2 indicators in [ a-z0-9]
unused indicators should be blank; printed as # in specs
other unused but fixed coded data should use a '|'
*/
enum {
	RT = 29, /* ^] record terminator;  printed \ in MARC specs */
	FT = 30, /* ^^ field terminator;   printed ^ or @ in MARC specs */
	DL = 31  /* ^_ subfield delimiter; printed $ in MARC specs */
};

typedef struct { /* 24 byte Leader; mostly ASCII digits */
	char len[5]; /* record length incl. RT */
	char sts; /* status (new, updated); "ASCII graphic" */
	char typ; /* type of record; "ASCII graphic" */
	char id0[3]; /* impl.def., used by some MARCs */
	char ind; /* indicator count; 2 in MARC */
	char idl; /* identifier length incl delimiter; 2 in MARC */
	char off[5]; /* base address of data = 24 + dict size */
	char id1[3]; /* impl.def., used by some MARCs */
	/* entry map; lof+scp > 0 */
	char lof; /* # bytes for length of field; 4 in MARC */
	char scp; /* # bytes for starting character pos; 5 in MARC */
	char idp; /* # bytes for impl.def. portion; 0 in MARC */
	char res; /* reserved; 0 in MARC */
} Leader;

typedef struct { /* dictionary entries after leader, MARC layout */
	char tag[3]; /* IIF allows alpha; MARC demands digits */
	char lof[4]; /* [lof] length of field incl indicators and FT */
	/* if lof is 0, next entry has addtl. lof for same field */
	char scp[5]; /* [scp]; starting char pos rel to off */
	/* char idp[idp]; impl.def. portion */
} Entry;
/* entries followed by an FT */

typedef struct {
	Fld file;  /* points into env->opt */
	int rid; /* use control number as MFN */
	int nomarc;  /* do not assume MARC */
	int write; /* write iif */
	char ind[2];
	char sub;
} IifOpt;


static int iifOpt (IifOpt *opt, Fld *args)
{
	Fld o;
	memset(opt, 0, sizeof(*opt));
	for ( o.val = 0; vGet(&o, args, "FNPR"); )
		switch (o.tag) {
		case 'F':
			opt->file = o;
			continue;
		case 'N':
			opt->nomarc = !0;
			continue;
		case 'P':
			opt->ind[0] = opt->ind[1] = ' ';
			opt->sub = '0';
			switch (o.len) {
			default: opt->sub = o.val[2];
			case 2: opt->ind[1] = o.val[1];
			case 1: opt->ind[0] = o.val[0];
			case 0: ;
			}
			continue;
		case 'R':
			opt->rid = !0;
			continue;
		}
	return 0;
}	/* iifOpt */


int iifuse ()
{
	return eRr(ERR_INVAL, "[iifimp|iifexp]:\n"
		"-F<file>: path/filename of iif (default = malete db.iif)\n"
#ifndef WIN32
		"\t-F- for stdin/out\n"
#endif
		"-Nomarc: do not assume MARC structure 22/450# (imp)\n"
		"-P[<ii><c>]: prepend indicators ii and, where needed, subfield c (exp)\n"
		"-Rid: use control number as rid (imp)\n"
		"use tools like yaz-marcdump for character set recoding\n"
	);
}	/* iifuse */


static int iifOpen (file *iif, IifOpt *opt, Db *db)
{
	int filmode = opt->write ? FIL_RDWR|FIL_CREAT|FIL_TRUNC : FIL_RD;
	int ret;
	char *path;
	if ( opt->file.len ) {
#ifndef WIN32
		if ( 1 == opt->file.len && '-' == *opt->file.val ) {
			*iif = opt->write ? 1 : 0;
			return 0;
		}
#endif
		path = VDUPZ(&opt->file);
	} else {
		int l = strlen(db->pat);
		path = mAlloc(l+5);
		memcpy(path, db->pat, l);
		memcpy(path+l, ".iif", 5);
	}
	ret = fOpen(iif, path, filmode) ?
		eRr(LOG_ERROR, "no access to IIF file '%s'", path) : 0;
	mFree(path);
	return ret;
}	/* iifOpen */


/*
tags may contain alpha; uppercase or lowercase ("but not both")
UNIMARC and others specify plain numeric tags
MARC specs first two to be digits; we map:
3 digits to 0-999 (000 actually forbidden)
2 digits: 3rd 'a' to 1000-1099, 'b' to 1100-1199 up to 3599
1 digit: 1st + 360*2ndbase36 + 10*3rdbase36 up to 12959
others are base 36 (>=12960 = 'a00') < 36^3 = 46656
*/
static int rdtag (char *t)
{
	int tag;
	switch ( a2il(t, 3, &tag) ) {
	case 3: return tag;
	case 2: return tag + 100*b36val[(unsigned char)t[2]];
	case 1: return tag
		+ 360*b36val[(unsigned char)t[1]]
		+ 10*b36val[(unsigned char)t[2]];
	}
	return 1296*b36val[(unsigned char)t[0]]
		+ 36*b36val[(unsigned char)t[1]]
		+ b36val[(unsigned char)t[2]];
}

static OPT_INLINE void pr (char *t, unsigned i, unsigned l)
{
	for ( ; l--; i/=10) t[l] = b36dig[i%10];
}

static void prtag (char *t, unsigned tag)
{
	if (1000 > tag)
		pr(t, tag, 3);
	else if (3600 > tag) {
		pr(t, tag%100, 2);
		t[2] = b36dig[tag/100];
	} else if (12960 > tag) {
		t[0] = b36dig[tag%10]; tag /= 10;
		t[2] = b36dig[tag%36]; tag /= 36;
		t[1] = b36dig[tag];
	} else {
		t[2] = b36dig[tag%36]; tag /= 36;
		t[1] = b36dig[tag%36]; tag /= 36;
		t[0] = b36dig[tag%36];
	}
}



int iifimp (Db *db, Fld *args)
{
	char buf[99999];
	IifOpt opt;
	file iif;
	Leader ldr;
	List rec;
	int ret = 0;

	if ( iifOpt(&opt, args) )
		return iifuse();
	if ( iifOpen(&iif, &opt, db) )
		return eRr(ERR_INVAL, "bad IIF base");
#define CLEANUP( args ) do { ret = eRr args; goto cleanup; } while (0)
	lInit(&rec, "");
	for (;;) {
		char head[8];
		char *ent = buf, *base;
		int len, off, eln, dln, lof = 0, scp = 0;
		int got = fRead(&iif, &ldr, sizeof(ldr));
		int rid = 0;
		if ( !got )
			break;
		if ( (int)sizeof(ldr) != got )
			CLEANUP((ERR_TRASH, "reading leader got %d", got));
		LOG_DBG(LOG_TRACE, "leader '%.*s'", 24, &ldr);
		len = a2i(ldr.len, 5) - 24;
		off = a2i(ldr.off, 5);
		if ( opt.nomarc ) {
			ldr.ind -= '0';
			ldr.idl -= '0';
			ldr.lof -= '0';
			ldr.scp -= '0';
			ldr.idp -= '0';
		} else {
			ldr.ind = 2;
			ldr.idl = 2;
			ldr.lof = 4;
			ldr.scp = 5;
			ldr.idp = 0;
		}
		eln = 3+ldr.lof+ldr.scp+ldr.idp; /* entry length */
		dln = (off-25)/eln; /* dict len: off - leader - FT */
		eRr(LOG_VERBOSE,
			"len %d sts %c typ %c ind %d idl %d off %d map 3+%d+%d+%d=%d %d fields",
			len, ldr.sts, ldr.typ, ldr.ind, ldr.idl, off,
			ldr.lof, ldr.scp, ldr.idp, eln, dln);
		if ( off != 25+eln*dln )
			eRr(ERR_TRASH, "bad off %d != 25+%d*%d", off, eln, dln);
		got = fRead(&iif, buf, len);
		if ( len != got )
			CLEANUP((ERR_TRASH, "reading body %d got %d", len, got));
		base = buf+off-24;
		if ( FT != base[-1] )
			eRr(ERR_TRASH, "no FT after dict");
		if ( RT != buf[len-1] )
			eRr(ERR_TRASH, "no RT");
		memcpy(head, &ldr.sts, 5); /* sts,typ,id0 */
		memcpy(head+5, ldr.id1, 3);
		lClr(&rec);
		LADD(&rec, -1, head, 8);
		for ( ; dln--; ent+=eln ) {
			int tag = rdtag(ent);
			char *v, *e, *p;
			if ( ldr.scp )
				scp = a2i(ent+3+ldr.lof, ldr.scp);
			else
				scp += lof;
			v = base+scp;
			if ( !ldr.lof ) {
				if ( !(e = memchr(v, FT, len-off+24)) )
					CLEANUP((ERR_TRASH, "no FT after field"));
				lof = e-v+1;
			} else {
				if ( !(lof = a2i(ent+3, ldr.lof)) ) { /* long field */
					int max=1, dig=ldr.lof, add;
					while (dig--) max *= 10;
					lof = max;
					do {
						ent += eln;
						if ( !dln-- ) CLEANUP((ERR_TRASH, "eod on long field"));
						if ( tag != a2i(ent, 3) ) eRr(ERR_TRASH, "bad tag");
						add = a2i(ent+3, ldr.lof);
						lof += add ? add : max;
					} while ( !add );
				}
				if ( FT != *(e = v+lof-1) )
					eRr(ERR_TRASH, "no FT after field");
			}
			for ( p=v; p<e; p++ ) if ( DL == *p ) *p = TAB;
			/* eOut(tag, "%.*s", lof-1, v); */
			LADD(&rec, tag, v, (unsigned)lof-1);
		}
		if ( opt.rid && 1<LLEN(&rec) && 1 == rec.fld[1].tag ) {
			int num;
			if ( (int)rec.fld[1].len == a2il(rec.fld[1].val, rec.fld[1].len, &num) )
				rid = num;
			else
				eRr(LOG_WARN, "bad control number '%.*s'",
					rec.fld[1].len, rec.fld[1].val);
		}
		dWrite(db, rec.fld, rid);
	}
cleanup:
#undef CLEANUP
	fClose(&iif);
	return ret;
}	/* iifimp */


int iifexp (Db *db, Fld *args)
{
	char buf[99999];
	IifOpt opt;
	file iif;
	Leader ldr;
	List rec;
	int i, ret = 0;

	if ( iifOpt(&opt, args) )
		return iifuse();
	opt.write = 1;
	if ( iifOpen(&iif, &opt, db) )
		return eRr(ERR_INVAL, "opening IIF");
	ldr.ind = ldr.idl = '2';
	ldr.lof = '4';
	ldr.scp = '5';
	ldr.idp = '0';
	ldr.res = ' ';
	lInit(&rec, "");
	for (i=1; i<=db->rdx.mid; i++) { /* loop records */
		char head[8], rid[12];
		char *p;
		int l, n, o, ll, ridl;
		Entry *e;
		Fld *f;
		if ( 0 >= dRead(lClr(&rec), db, i) ) /* no rec */
			continue;
		ll = LLEN(&rec)-1;
		/* fill head */
		for (p = rec.fld->val, l = rec.fld->len; l-- && TAB != *p++; ) ;
		if ( 8 <= l )
			memcpy(head, p, 8);
		else {
			memset(head, ' ', 8);
			if ( 0 < l )
				memcpy(head, p, l);
		}
		memcpy(&ldr.sts, head, 5);
		memcpy(ldr.id1, head+5, 3);
		e = (Entry*)(buf + 24); /* after leader */
		l = 26; /* leader + dict FT + RT */
		o = 0; /* offset to base */
		n = 0; /* # fields used */
		/* rid -> control number */
		if ( ll && 1 == rec.fld[1].tag )
			ridl = 0;
		else {
			ridl = i2a(rid, i);
			prtag(e->tag, 1);
			pr(e->lof, o = ridl+1, 4);
			pr(e->scp, 0, 5);
			l += o + 12;
			e++;
		}
		for ( f = rec.fld+1; l < 99900 && n < ll; n++,e++,f++ ) {
			int fln = 1 /* FT */, vln = f->len;
			l += 12; /* Entry */
			if ( f->tag > 9 ) { /* non control */
				if ( opt.ind[0] ) fln += 2;
				if ( opt.sub && vln && TAB != *f->val ) fln += 2;
			}
			if ( vln > 9994 ) vln = 9994; /* max field len */
			if ( vln > 99994-l ) vln = 99994-l; /* max rec len */
			l += fln += vln;
			prtag(e->tag, f->tag);
			pr(e->lof, fln, 4);
			pr(e->scp, o, 5);
			o += fln;
		}
		p = (char*)e;
		*p++ = FT;
		pr(ldr.len, l, 5);
		pr(ldr.off, p - buf, 5);
		memcpy(buf, &ldr, 24);
		/* now add the field data */
		if ( ridl ) {
			memcpy(p, rid, ridl);
			p += ridl;
			*p++ = FT;
		}
		for ( f = rec.fld+1; n--; f++ ) {
			int vln = f->len;
			char *v = f->val;
			if ( f->tag > 9 ) { /* non control */
				if ( opt.ind[0] ) {
					*p++ = opt.ind[0];
					*p++ = opt.ind[1];
				}
				if ( opt.sub && vln && TAB != *f->val ) {
					*p++ = DL;
					*p++ = opt.sub;
				}
			}
			if ( vln > 9994 ) vln = 9994;
			if ( vln > 99994-(p-buf) ) vln = 99994-(p-buf);
			while ( vln-- ) if ( TAB == (*p++ = *v++) ) p[-1] = DL;
			*p++ = FT;
		}
		*p++ = RT;
		if ( l != p-buf ) {
			eRr(ERR_IDIOT, "miscalculated len %d got %d rec %d", l, p-buf, i);
			break;
		}
		fWrite(&iif, buf, l);
	}
	fClose(&iif);
	return ret;
}	/* iifexp */
