/*
	The Malete project - the Z39.2/Z39.50 database framework of OpenIsis.
	Version 0.9.x (patchlevel see file Version)
	Copyright (C) 2001-2003 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	see README for more information
EOH */

/*
	$Id: qdx.c,v 1.2 2005/04/25 15:49:25 kripke Exp $
	the btree.
*/

#include <stdlib.h>
#include <string.h>

#include "core.h"



/* ************************************************************
	private types
*/


/**
	Dictionary layout for forks (inner nodes of tree):
*/
typedef struct Dict { /* dictionary of block entries */
	unsigned short pos; /* position of entry, native byte order */
	unsigned char  vln; /* number of values */
	unsigned char  kln; /* length of key */
} Dict;
/*
	For leaves, the max block size is 8K,
	which need 13 bit for pos and 11 bit for vln (max 8K/min vsz 4 values).
	The 1st 3 bytes are:
	1. lower 8 bit of pos
	2. higher 5 bit of pos | higher 3 bit of vlen << 5
	3. lower 8 bit of vln
*/
#ifdef CPU_BIG_ENDIAN
#define LEAFPOS( d ) (*((unsigned char*)(d)) \
	| (((unsigned char*)(d))[1]&0x1f)<<8)
#define LEAFVLN( d ) ((d)->vln \
	| (((unsigned char*)(d))[1]&0xe0)<<3)
#define LEAFPOSADD( d, a ) do { \
		unsigned char *__u = (unsigned char*)(d); \
		unsigned short __p = (a) + __u[0] + (__u[1]<<8); \
		__u[0] = (unsigned char)__p;  \
		__u[1] = (unsigned char)((__u[1]&0xe0)|((__p>>8)&0x1f)); \
	} while (0)
#define LEAFVLNADD( d, a ) do { \
		unsigned char *__u = (unsigned char*)(d); \
		unsigned short __v = (a) + __u[2] + ((__u[1]&0xe0)<<3); \
		__u[2] = (unsigned char)__v; \
		__u[1] = (unsigned char)((__u[1]&0x1f)|((__v>>3)&0xe0)); \
	} while (0)
#define LEAFSETPV( d, p, v ) do { \
		unsigned char *__u = (unsigned char*)(d); \
		__u[0] = (unsigned char)(p); \
		__u[1] = (unsigned char)((((p)>>8)&0x1f)|(((v)>>3)&0xe0)); \
		__u[2] = (unsigned char)(v); \
	} while (0)
#else
#define LEAFPOS( d ) (0x1fff & (d)->pos)
#define LEAFVLN( d ) ((d)->vln | (~0x1fff & (d)->pos)>>5)
#define LEAFPOSADD( d, a ) do { \
		(d)->pos = (~0x1fff & (d)->pos) | (0x1fff & ((d)->pos + (a))); \
	} while (0)
#define LEAFVLNADD( d, a ) do { \
		unsigned _vln = LEAFVLN( d ) + (a); \
		(d)->pos = (0x1fff & (d)->pos) \
			| (~0x1fff & ((_vln)<<5)); \
		(d)->vln = (unsigned char)(_vln); \
	} while (0)
#define LEAFSETPV( d, p, v ) do { \
		(d)->pos = (0x1fff & (p)) | (~0x1fff & ((v)<<5)); \
		(d)->vln = (unsigned char)(v); \
	} while (0)
#endif

#define LEAFSETPVK( d, p, v, k ) do { \
		LEAFSETPV( d, p, v ); \
		(d)->kln = (unsigned char)(k); \
	} while (0)


/*
	the first 8 bytes are actually redundant,
	however are here for convenience and as a check
*/
typedef struct Block {
	unsigned       num; /* number of this block, for check and convenience */
	unsigned char  typ; /* block type & endianess */
	unsigned char  ksz; /* max key length */
	unsigned char  ptr; /* ptr type; for forks+plain leaves: value size */
	unsigned char  lev; /* level over bottom, 0 for leaves */
	unsigned       nxt; /* right sibling, free list for root */
	unsigned short ent; /* # entries */
	unsigned short stk; /* offset of keyspace stack */
	Dict           dct[1];
}	Block;
#define DICTOFF ((char *)&((Block*)0)->dct - (char*)((Block*)0))
#define GETENT(b, i) (((unsigned char *)(b)) \
	+ ((b)->lev ? (b)->dct[i].pos : LEAFPOS((b)->dct+(i))))

/*
	LAYOUT:
	dct is an array of dictionary entries in increasing key order
	stk is growing downwards from block end towards dict
	position of entry i is b->stk <= LEAFPOS(b->dct+i) < dx->t/lsz
	decreasing on i (i.e. stack space is in reverse ordering)

	at LEAFPOS(b->dct+i), there are b->dct[i].kln bytes key
	for a leaf entry, the key is followed by
	LEAFVLN(b->dct+i) (between 0..2047 incl.) values
	of dx->vsz (between 8..23 incl) each
	for a fork entry, the key is followed by
	a single 4 bytes block number (vln = 0)
	or vln pairs of qualifying values (vsz each) and block numbers
	(without the key compression flag to be implemented, only vln=1 is supported)

	SIZES:
	block size >= 512
	block base overhead DICTOFF is 16 bytes
	avail size >= 496
	Dict entry is 4 bytes
	max. used stack space per entry is:
	- fork: kln+4+dx->vsz <= 255+4+255 = 514
	- new leaf entry (one value): kln+dx->vsz <= 510
	max total space for a new entry incl. 4 bytes dict is 518.
	For leaves of minimal block size 512,
	the sum of ksz+vsz must not exceed 492=512-16-4,
	so that each block can hold at least one entry.
	Forks are at least 4K and thus can hold min 7 entries.
	Even in pathological cases with max kln and vsz and all forks
	split to the worst, a level's fork factor does not drop below 4.
*/

enum {
	QMAXLEAF = 0x2000, /* max leaf size -- 8K fixed !!! */
	QMAXFORK = 1<<CPU_PAGE_SHIFT,
#if 0x2000 < 1<<CPU_PAGE_SHIFT
	QMAXBLCK = 1<<CPU_PAGE_SHIFT,
#else
	QMAXBLCK = 0x2000,
#endif
	/* With 2Gig files there are at most 2^31/2^9=2^22 leaves,
	needing 11 fork levels. */
	QLEVELS  = 12 /* max depth incl. leaves */
};
typedef union { Block b; unsigned char base[QMAXLEAF]; } Leaf;
typedef union { Block b; unsigned char base[QMAXFORK]; } Fork;
typedef union { Block b; unsigned char base[QMAXBLCK]; } Bloke; /* any dude */


enum {
	QENTSIZE = 8, /* minimum cost of entry: 4 byte dict, 4 byte val */
	/* the number of entries is limited by (QMAXFORK-16)/QENTSIZE < 8K */
	QENTBITS = 13, /* bits for # enries */
	QENTMASK = 0x1fff,
	QENTMAX  = QENTMASK,
	QMATCH   = 0x8000 /* entry matched flag in found entry numbers */
};

#if 0
#	define SORT_Q
/* glibc's qsort is up to 20% faster than our basic untuned WHS. However:
-	dietlibc's is slower
-	it still has poor worst case performance
-	many implementations are known to be buggy
*/
#endif
/* memory for internal sorting */
/*
loaded 5MB index in 5 seconds with as little as
#define SORTMEM  0x30000 = 192K (+96K Keys) internal
#define SORTMEM  0x400000 = 4MB (+2MB Keys) internal
*/
#define SORTMEM  0x400000
#define SORTKEYS (SORTMEM/8) /* + another 50%: #entries (assume > 8 Bytes) */

/* blocksize for files */
/*
#define MERGEBLK 0x20000 = 128K buffer, 48-way merge
#define MERGEBLK 0x3000  = 12K buffer, 512-way merge

12K blocks perform quite well and allow for 512-way merge,
so we never have a second pass for 31-bit addressable files
*/
#define MERGEBLK 0x3000
#define MERGECHK (SORTMEM/MERGEBLK) /* initial chunk size */

/*
	By organizing files in MERGEBLKs, so that entries don't span blocks,
	we may be wasting up to one full entry (256 bytes) per block:
	if we have only exactly the space for one such longest entry,
	use one terminating 0 byte instead.
*/
#define SORTUSE (MERGECHK*(MERGEBLK-256))

typedef struct {
	unsigned char *keyb[SORTKEYS];
	unsigned char memb[SORTMEM];
} SortBuf;
typedef unsigned char MergeBlk[MERGEBLK];

/*	n-way merge
#define MERGEWAY (sizeof (SortBuf)/sizeof (MergeBlk))
*/
#ifndef MERGEWAY /* define possible max */
#	define MERGEWAY ((4*SORTKEYS+SORTMEM)/MERGEBLK) /* as pre-pro constant */
#endif
#if 0x80000000UL > (MERGEWAY*SORTMEM) 
#	define MERGEMULTIPASS
/* else (512-way merge on 4MB) we can do 2GB in one pass */
#endif

typedef struct QLoad {
	unsigned got; /* current for this key */
	unsigned tot; /* total for this key */
	unsigned let;
	/* stats */
	unsigned keys;
	unsigned keyl;
	unsigned maxl;
	unsigned vals;
	unsigned maxv;
	unsigned span;
	unsigned free;
	/* load buffer */
	Leaf     cur;
	Key      key;
	Val      val[QENTMAX];
	/* sort buffer */
	file     fil;
	unsigned blks;
	unsigned char **nkey;
	unsigned char *nmem;
#if 0 /* hard to measure ... */
#	define ALIGN4K __attribute__ ((aligned (0x1000)))
#else
#	define ALIGN4K
#endif
	union {
		SortBuf s;
		MergeBlk m[MERGEWAY]; /* mergebuf */
	} buf ALIGN4K;
	MergeBlk mout; /* only used temporary, but don't want to kill stack */
}	QLoad;



typedef struct {	/* structure filled by findkey */
	Qdx           *dx;	/* the tree */
	Key           *key; /* the key searched */
	unsigned       n[QLEVELS]; /* block numbers of path */
	unsigned short e[QLEVELS]; /* entries in path with match flag */
}	FindKey;


typedef struct {
	unsigned key;
	unsigned cmp;
	unsigned fget;
	unsigned lget;
}	Stat;

/* buffer to print a value to */
typedef char PrVal[64];


/* ************************************************************
	private data
*/

static Stat stat;


/* ************************************************************
	private functions
*/

/* bail out */
#define DONE( args ) do { ret = eRr args; goto done; } while (0)

#ifdef CPU_NEED_ALIGN
static unsigned GETINT ( void *m ) /* bus error safe version of *(int*) */
{
	unsigned l;
	memcpy( &l, m, 4 );
	return l;
}
static void SETINT ( void *m, unsigned i )
{
	memcpy( m, &i, 4 );
}
#else
#define GETINT( m ) (*(unsigned*)(m))
#define SETINT( m, i ) (*(unsigned*)(m) = (i))
#endif

void qMkVal ( Val *v, Ptr *p, unsigned char typ )
{
	unsigned char *c = v->byt;
	/* TODO: may macros to address the bytes in the ints would be faster */
	if ( !(QDX_FULTXT & typ) ) /* leading tag */
		switch ( QDX_TAGMSK&typ ) {
		case QDX_TAG2: *c++ = (unsigned char) (p->tag >> 8);
		case QDX_TAG1: *c++ = (unsigned char)  p->tag;
		}
	switch ( QDX_RIDMSK&typ ) {
	case QDX_RID6: *c++ = (unsigned char) (p->ext >> 8);
	case QDX_RID5: *c++ = (unsigned char)  p->ext;
	case QDX_RID4: *c++ = (unsigned char) (p->rid >> 24);
	case QDX_RID3:
		*c++ = (unsigned char) (p->rid >> 16);
		*c++ = (unsigned char) (p->rid >> 8);
		*c++ = (unsigned char)  p->rid;
	}
	if ( QDX_FULTXT & typ )
		switch ( QDX_TAGMSK&typ ) {
		case QDX_TAG2: *c++ = (unsigned char) (p->tag >> 8);
		case QDX_TAG1: *c++ = (unsigned char)  p->tag;
		}
	switch ( QDX_POSMSK&typ ) {
	case 4: *c++ = (unsigned char) (p->pos >> 24);
	case 3: *c++ = (unsigned char) (p->pos >> 16);
	case 2: *c++ = (unsigned char) (p->pos >> 8);
	case 1: *c++ = (unsigned char)  p->pos;
	}
	v->len = c - v->byt;
}	/* qMkVal */


int qRdKey ( Qdx *qdx, char *plain, int l, Key *key )
{
	if (!qdx->cdx) {
		if ( l > key->len ) l = key->len;
		memcpy(plain, key->byt, l);
	} else
		l = cDec(qdx->cdx, (unsigned char*)plain, l, key);
	return l;
}	/* qRdKey */


void qMkKey ( Qdx *qdx, Key *key, char *b, int l )
{
	if (!qdx->cdx) {
		if ( l > qdx->ksz ) l = qdx->ksz;
		memcpy(key->byt, b, key->len = l);
	} else {
		key->len = qdx->ksz;
		cEnc(qdx->cdx, key, (unsigned char*)b, l, 0);
	}
}	/* qMkKey */


int qMkKeyVal ( Qdx *qdx, Key *key, char *b, int l )
{
	Ptr p = {0,0,0,0};
	char *t = memchr(b, '\t', l);
	if ( !t )
		return -1;
	qMkKey(qdx, key, b, t - b);
	if ( 0 >= (l -= t-b+1) || !(t = memchr( b=t+1, '\t', l )) )
		return -1;
	p.rid = (unsigned) a2i( b, t-b );
	if ( 0 >= (l -= t-b+1) || !(t = memchr( b=t+1, '\t', l )) )
		return -1;
	p.tag = (unsigned short) a2i( b, t-b );
	if ( 0 >= (l -= t-b+1) || !(t = memchr( b=t+1, '\t', l )) )
		return -1;
	p.pos = (unsigned) a2i( b, t-b );
	if ( 0 < (l -= t-b+1) )
		p.pos = p.pos << 16 | (unsigned short) a2i( t+1, l );
	eRr(LOG_TRACE, "%'%b %d %d 0x%4x",
		key->len, key->byt, p.rid, p.tag, p.pos );
	qMkVal( &key->val, &p, qdx->ptr );
	return 0;
}	/* qMkKeyVal */


void qRdVal ( Ptr *p, const unsigned char *c, unsigned char typ )
{
	p->tag = p->ext = p->rid = p->pos = 0;
	if ( !(QDX_FULTXT & typ) ) /* leading tag */
		switch ( QDX_TAGMSK&typ ) {
		case QDX_TAG2: p->tag  = (unsigned short)*c++ << 8;
		case QDX_TAG1: p->tag |=                 *c++;
		}
	switch ( QDX_RIDMSK&typ ) {
	case QDX_RID6: p->ext  = (unsigned short)*c++ << 8;
	case QDX_RID5: p->ext |=                 *c++;
	case QDX_RID4: p->rid  = (unsigned   int)*c++ << 24;
	case QDX_RID3:
		p->rid |= (unsigned   int)*c++ << 16;
		p->rid |= (unsigned   int)*c++ << 8;
		p->rid |=                 *c++;
	}
	if ( QDX_FULTXT & typ )
		switch ( QDX_TAGMSK&typ ) {
		case QDX_TAG2: p->tag  = (unsigned short)*c++ << 8;
		case QDX_TAG1: p->tag |=                 *c++;
		}
	switch ( QDX_POSMSK&typ ) {
	case 4: p->pos  = (unsigned   int)*c++ << 24;
	case 3: p->pos |= (unsigned   int)*c++ << 16;
	case 2: p->pos |= (unsigned short)*c++ << 8;
	case 1: p->pos |=                 *c++;
	}
}	/* qRdVal */


static char *prval ( PrVal buf, unsigned char *c, unsigned typ )
{
	char *d = buf;
	Ptr p;
	if ( ! typ ) {
		buf[0] = '#';
		buf[1] = 0;
		return buf;
	}
	qRdVal( &p, c, typ );
	if ( p.ext ) { d += u2a( d, p.ext ); *d++ = '.'; }
	d += u2a( d, p.rid ); *d++ = '.';
	d += u2a( d, p.tag ); *d++ = '.';
	d += u2a( d, p.pos ); *d++ = '.';
	*d = 0;
	return buf;
}	/* prval */

#define PRVAL( buf, v ) prval( buf, v.byt, dx->ptr )

#define QLOCK(n,l) (void)FLOCK(dx->mqd, n, l)
#define QLOCKSH(n) (void)FLOCKSH(dx->mqd, n)
#define QLOCKEX(n) (void)FLOCKEX(dx->mqd, n)
#define QLOCKUN(n) (void)FLOCKUN(dx->mqd, n)
#ifdef BUILD_SHMODE
enum {
	LOCK_READ  = FIL_BLOCK,
	LOCK_WRITE = FIL_BLOCK|FIL_WR
};
#endif


/* ************************************************************
	block I/O
*/


static int chkblk ( Block *b, Qdx *dx )
{
	int i;
	unsigned short eod = DICTOFF + b->ent*sizeof(Dict), siz, pos, vln, key;
	unsigned char *base = (unsigned char*)b;
	unsigned len = b->lev ? dx->fln : dx->lln;
	unsigned bsz = b->lev ? env.psz : dx->lsz;
	unsigned end = bsz, tot = 0;
	if ( b->num > len || (b->num && b->nxt > len) )
		return eRr(ERR_TRASH, "bad num %d / nxt %d on lev %d",
			b->num, b->nxt, b->lev );
	if ( b->typ != (b->lev ? dx->ftp : dx->typ)
		|| b->ksz != dx->ksz || b->ptr != (b->lev?dx->vsz:dx->ptr)
	)
		return eRr(ERR_TRASH, "bad sig 0x%02x,%d,%d in block %d want 0x%02x,%d,%d",
			b->typ, b->ksz, b->ptr, b->num,
			b->lev ? dx->ftp : dx->typ, dx->ksz, (b->lev?dx->vsz:dx->ptr) );
	if ( b->ent > bsz / QENTSIZE || b->stk > bsz || b->stk < eod )
		return eRr(ERR_TRASH, "bad ent/stk %d,%d in block %d eod %d sz %d",
			b->ent, b->stk, b->num, eod, bsz );
	for ( i=0; i<b->ent; i++ ) {
		Dict *d = &b->dct[i];
		if ( b->lev ) {
			pos = d->pos; vln = d->vln;
			siz = vln ? vln*(dx->vsz+4) : 4;
		} else {
			pos = LEAFPOS(d); vln = LEAFVLN(d);
			siz = vln*dx->vsz;
		}
		siz += key = d->kln;
		if ( pos < b->stk || end != (unsigned)pos + siz )
			return eRr(ERR_TRASH,
				"bad pos/vln/key %d/%d/%d in entry %d of block %d"
				" stack %d sz %d end %d expected %d",
				pos, vln, key, i, b->num, b->stk, siz, pos+siz, end );
		if ( b->lev ) {
			unsigned lnk = GETINT( base + pos + key + vln*dx->vsz );
			if ( lnk >= (1==b->lev ? dx->lln : dx->fln) )
				return eRr(ERR_TRASH,
					"bad link %u in entry %d of block %d",
					lnk, i, b->num );
		}
#ifndef NDEBUG
		if ( i ) { /* not on first */
			unsigned short op = b->lev ? d[-1].pos : LEAFPOS(d-1), okl = d[-1].kln;
			int mc = memcmp( base+pos, base+op, key<okl ? key : okl ), j;
			if ( 0 < mc || (!mc && 0 < (mc = (int)key - (int)okl)) )
				goto ok;
			j = 0;
			if ( !mc && b->lev && vln ) { /* compare values */
				unsigned ovln = d[-1].vln;
				if ( ! ovln || 0 < (j = memcmp( base+pos+key, base+op+okl, dx->vsz )) )
					goto ok;
			}
			return eRr(ERR_TRASH,
				"bad key order entry %d of block %d lev %d"
				" mc %d j %d key %d okl %d vln %d",
				i, b->num, b->lev, mc, j, key, okl, vln
			);
		}
	ok:
#endif
		tot += siz;
		end = pos;
	} /* for i (entries) */
	if ( tot > bsz - b->stk )
		return eRr(ERR_TRASH,
			"bad total entry size %d in block %d stack %d",
			tot, b->num, b->stk );
#ifndef NDEBUG
	if ( LOG_TRACE >= env.log ) {
		eRr(LOG_TRACE, 
			"block %d lev %d sig 0x%02x,%d ent %d eod %d stk %d",
			b->num, b->lev, b->typ, b->ksz, b->ent, eod, b->stk
		);
		for ( i=0; i<2; i++ ) {
			int j = i ? b->ent-1 : 0;
			Dict *d = &b->dct[j];
			if ( i >= b->ent ) /* if b->ent is 0 or 1, log no or one key */
				break;
			if ( b->lev ) {
				pos = d->pos; vln = d->vln;
			} else {
				pos = LEAFPOS(d); vln = LEAFVLN(d);
			}
			key = d->kln;
			/* start of value */
			if ( b->lev )
				eRr(LOG_TRACE, "key %d '%b'/%d -> %d",
					j, key, base+pos, vln, GETINT(base+pos+key+(vln?4:0)) );
			else {
				Ptr p;
				qRdVal( &p, base+pos+key, dx->ptr );
				eRr(LOG_TRACE, "key %d '%b'[%d] = %d[%d,%d]",
					j, key, base+pos, vln, p.rid, p.tag, p.pos );
			}
		}
	}
#endif
	return 0;
}	/* chkblk */


/*
	init a new leaf in the local buffer b (usually a stack based Leaf).
	increment dx's lln
*/
static void newleaf ( Block *b, Qdx *dx )
{
	memset(b, 0, dx->lsz);
	b->num = dx->lln++;
	b->typ = dx->typ;
	b->ksz = dx->ksz;
	b->ptr = dx->ptr;
	b->lev = 0;
	b->stk = dx->lsz;
}	/* newleaf */


/*
	init a new fork in the local buffer b (usually a stack based Fork).
	increment dx's fln
	TODO: for qSet, use variant checking for free list at root->nxt.
*/
static Block *newfork ( Block *b, Qdx *dx, unsigned char lev )
{
	memset(b, 0, env.psz);
	b->num = dx->fln++;
	b->typ = dx->ftp;
	b->ksz = dx->ksz;
	b->ptr = dx->vsz;
	b->lev = lev;
	b->stk = env.psz;
#if 16 == CPU_PAGE_SHIFT /* possible short overflow on ia64 */
	if ( !b->stk ) b->stk--;
#endif
	return b;
}	/* newfork */


#ifdef CPU_BIG_ENDIAN
static unsigned rvi ( unsigned i ) {
	unsigned r;
	((char*)&r)[0] = ((char*)&i)[3];
	((char*)&r)[1] = ((char*)&i)[2];
	((char*)&r)[2] = ((char*)&i)[1];
	((char*)&r)[3] = ((char*)&i)[0];
	return r;
}
static unsigned short rvs ( unsigned short i ) {
	unsigned short r;
	((char*)&r)[0] = ((char*)&i)[1];
	((char*)&r)[1] = ((char*)&i)[0];
	return r;
}
#define SWI( i ) b->i = rvi( b->i )
#define SWS( s ) b->s = rvs( b->s )
static void swapleaf ( Block *b ) /* leaves are little endian */
{
	SWI(num);
	SWI(nxt);
	SWS(ent);
	SWS(stk);
}
#else
#	define swapleaf(b)
#endif

/*
	writing a leaf will render the leaf scrambled on big endians
*/
static int wrtleaf ( Block *b, Qdx *dx )
{
	unsigned num = b->num;
	int got;
	if ( chkblk( b, dx ) ) /* don't write no shize */
		return eRr(ERR_TRASH, "\tbefore writing dx block %d", b->num );
	swapleaf(b); /* a leaf written is a leaf lost */
	got = fPwrite(&dx->mqd, b, dx->lsz, num*dx->lsz );
	eRr(LOG_TRACE, "writing leaf %u got %u", num, got );
	if ( dx->lsz != (unsigned)got )
		return eRr(LOG_SYSERR, "\twhen writing dx block %d", b->num );
	return 0;
} /* wrtleaf */


/*
	writing a fork may invalidate all mapped forks due to remap
*/
static int wrtfork ( Block *b, Qdx *dx )
{
	int got;
	if ( chkblk( b, dx ) ) /* don't write no shize */
		return eRr(ERR_TRASH, "\tbefore writing dx block %d", b->num );
	LOG_DBG( LOG_DEBUG, "writing fork %d/%d from %8x/%8x", b->num, dx->mqx.npg,
		b, dx->mqx.map);
	if ( dx->mqx.npg ) { /* have mmap -- not during qload */
		if ( b->num < dx->mqx.npg ) { /* should be in map */
			if ( (char*)b != dx->mqx.map + (b->num<<env.psh) )
				return eRr(ERR_TRASH, "fork %d should be mapped", b->num);
			if ( ENV_MSYNC & env.flg ) fMSync(&dx->mqx, b->num);
			return 0;
		}
		if ( b->num < dx->mqx.lim /* try to extend mapping */
			&& 0 < (got = fMap(&dx->mqx, b->num+1))
			&& b->num < (unsigned)got
		) {
			eRr(LOG_VERBOSE, "remapped to %d at %8x", dx->mqx.npg, dx->mqx.map);
			memcpy(dx->mqx.map + (b->num<<env.psh), b, env.psz);
			fMSync(&dx->mqx, b->num);
			return 0;
		}
		eRr(LOG_WARN, "remap to %d failed", b->num+1);
		/* else write to file */
	}
	got = fPwrite(&dx->mqx.fil, b, env.psz, b->num*env.psz );
	eRr(LOG_TRACE, "writing fork %u got %u", b->num, got );
	if ( env.psz != (unsigned)got )
		return eRr(LOG_SYSERR, "\twhen writing dx block %d", b->num );
	return 0;
} /* wrtfork */



static int chkread ( Block *b, Qdx *dx,
	int got, unsigned bsz, unsigned num, unsigned char lev )
{
	if ( 0 > got || bsz != (unsigned)got )
		return eRr(LOG_SYSERR, "\twhen reading dx block %d", num );
	if ( b->num != num )
		return eRr(ERR_TRASH, "bad num %d in block %d", b->num, num );
	if ( lev && b->lev != lev )
		return eRr(ERR_TRASH, "bad lev %d in block %d want %d",
			b->lev, b->num, lev );
	return chkblk( b, dx );
}	/* chkread */


/* read a leaf into the provided local buffer */
static Block *getleaf ( Block *b, Qdx *dx, unsigned num
#ifdef BUILD_SHMODE
	, int lck
#	define GETLEAF(leaf, dx, num, lck) getleaf(leaf, dx, num, lck)
#else
#	define GETLEAF(leaf, dx, num, lck) getleaf(leaf, dx, num)
#endif
)
{
	int got;
#ifdef BUILD_SHMODE
	if ( lck ) QLOCK(num<<1, lck);
#endif
	got = fPread(&dx->mqd, b, dx->lsz, num*dx->lsz);
	stat.lget++;
#ifdef CPU_BIG_ENDIAN
	if ( 0 < got && dx->lsz == (unsigned)got ) swapleaf(b);
#endif
	if ( chkread(b, dx, got, dx->lsz, num, 0) )
		b = 0;
#ifdef BUILD_SHMODE
	if ( LOCK_READ == lck ) QLOCKUN(num<<1);
#endif
	return b;
}	/* getleaf */


/* pick a fork from cache or read into the provided local buffer */
static Block *getfork ( Block *b, Qdx *dx, unsigned num, unsigned char lev )
{
	int got = env.psz;
	stat.fget++;
	if ( num < dx->mqx.npg )
		b = (Block*)(dx->mqx.map + num*env.psz);
	else
		got = fPread(&dx->mqx.fil, b, env.psz, num*env.psz);
	return chkread(b, dx, got, env.psz, num, lev) ? 0 : b;
}	/* getfork */



/** binsearch position for key in a leaf.
	position is either the index of the matching entry (0..b->ent-1)
	or the index (0..b->ent) where key should be inserted.
	@return position for key (0..b->ent), |yfied with QMATCH, if found
*/
static unsigned short keyleaf ( Block *b, Key *key )
{
	unsigned char *base = (unsigned char*)b;
	unsigned char *kbt = key->byt;
	unsigned char  kln = key->len;
	unsigned short l = 0, r = b->ent;
	Dict          *d = b->dct;
	/* binsearch invariant:
		key is < all entries i >= r (we do know relation at r)
		key is > all entries i < l (don't know relation at l)
	*/
	while ( l < r ) { /* invariant: l <= keypos < r */
		unsigned short i = (l + r) >> 1; /* l <= i < r */
		unsigned short pos = LEAFPOS(&d[i]);
		unsigned char cln = d[i].kln;
		int cmp = memcmp( kbt, base+pos, kln<cln ? kln : cln );
		stat.cmp++; /* dirty */
		if ( ! cmp && !(cmp = (int)kln - cln) )
			return QMATCH | i;
		if ( cmp > 0 )
			l = i+1;
		else
			r = i;
	}
	return l; /* == r */
}	/* keyleaf */

/* same for forks.
	somewhat more complex since we might have duplicate keys
	with different values.
	Must not use the _POS/_VLN macros.
*/
static unsigned short keyfork ( Block *b, Key *key, unsigned char vsz )
{
	unsigned char *base = (unsigned char*)b;
	unsigned char *kbt = key->byt;
	unsigned char  kln = key->len;
	unsigned short l = 0, r = b->ent;
	Dict          *d = b->dct;
	/* binsearch invariant:
		key is < all entries i >= r (we do know relation at r)
		key is > all entries i < l (don't know relation at l)
	*/
	while ( l < r ) { /* invariant: l <= keypos < r */
		unsigned short i = (l + r) >> 1; /* l <= i < r */
		unsigned short pos = d[i].pos;
		unsigned char cln = d[i].kln;
		int cmp = memcmp( kbt, base+pos, kln<cln ? kln : cln );
		stat.cmp++; /* dirty */
		if ( ! cmp && !(cmp = (int)kln - cln) ) { /* key match -- cmp value */
			if ( ! d[i].vln ) { /* common case */
				if ( !key->val.len ) /* no value */
					return QMATCH | i;
				cmp = 1;
			} else if ( ! key->val.len )
				cmp = -1;
			else if ( !(cmp = memcmp( key->val.byt, base+pos+cln,
					key->val.len < vsz ? key->val.len : vsz))
				&& !(cmp = (int) key->val.len - vsz)
			)
				return QMATCH | i;
		}
		if ( cmp > 0 )
			l = i+1;
		else
			r = i;
	}
	return l; /* == r */
}	/* keyfork */


/** find position for (leaf) val.
	search v in array b of n values of size sz
	@return position for value (0..n), |yfied with QMATCH, if found
*/
static unsigned short valpos ( unsigned char *b,
	unsigned short n, unsigned char sz, unsigned char *v )
{
	unsigned short l = 0, r = n;
	/* binsearch invariant:
		v is < all values i >= r (we do know relation at r)
		v is > all values i < l (don't know relation at l)
	*/
	while ( l < r ) { /* invariant: l <= valpos < r */
		unsigned short i = (l + r) >> 1; /* l <= i < r */
		int cmp = memcmp( v, b+i*sz, sz );
		if ( ! cmp )
			return QMATCH | i;
		if ( cmp > 0 )
			l = i+1;
		else
			r = i;
	}
	return l; /* == r */
}	/* valpos */


#define TRACEKEY( l, f ) (void)(f)


/*
	find and fetch key's leaf into the provided buffer.
*/
static int findkey ( Block *leaf, FindKey *f
#ifdef BUILD_SHMODE
	, int lck
#	define FINDKEY(leaf, f, lck) findkey(leaf, f, lck)
#else
#	define FINDKEY(leaf, f, lck) findkey(leaf, f)
#endif
)
{
#ifdef BUILD_SHMODE
	Qdx *dx = f->dx; /* for the lock macros */
#endif
	Fork fork[2]; /* buffer */
	Leaf lbuf;
	int i, ret=-1;
	unsigned char *pkey = 0, pkln = 0, pvln = 0;
	Block *b;

	QLOCKSH(1);
	b = getfork(&fork[0].b, f->dx, 0, 0);
	if ( !b )
		DONE((ERR_TRASH, "no root"));
	if ( b->lev >= QLEVELS )
		DONE((ERR_TRASH, "tree is too deep: %u", b->lev));
	stat.key++;
	for ( i=QLEVELS; i--; ) /* paranoia */
		f->n[i] = f->e[i] = 0;
	for ( i=b->lev;; ) {
		unsigned short e;
		unsigned  n;
		Dict *d;
		int tog = 0;

		f->n[i] = b->num;
		e = i ? keyfork(b, f->key, f->dx->vsz) : keyleaf(b, f->key);
		LOG_DBG( LOG_DEBUG, "keypos %4x", e );
		/*
			here goes the Lehmann/Yao race detection:
			If we ran on right boundary, follow right link
		*/
		while ( !(QMATCH & e) && b->ent == e && b->nxt && i < f->dx->dpt ) {
			Block *s;
			unsigned se;
			/* peek next block, toggling buffers */
			if ( i ) {
				/* TODO: actually we should never see this in forks
				as long as we lock the complete fork file */
				s = getfork(&fork[(++tog)&1].b, f->dx, b->nxt, i);
				se = keyfork(s, f->key, f->dx->vsz);
			} else {
				s = GETLEAF((++tog)&1?&lbuf.b:leaf, f->dx, b->nxt, lck);
				se = keyleaf(s, f->key);
#ifdef BUILD_SHMODE
				if ( LOCK_WRITE == lck ) QLOCKUN(se ? b->num : s->num);
#endif
			}
			if ( !se ) /* lower than first */
				break;
			LOG_DBG( LOG_DEBUG, "LYao %d -> %d lev %d for '%b' got 0x%4x",
				b->num, s->num, i, f->key->len, f->key->byt, se );
			/* trade b for s */
			f->n[i] = (b = s)->num;
			e = se;
		}
		f->e[i] = e;
		if ( ! i ) {
			if ( b != leaf )
				memcpy(leaf, b, f->dx->lsz);
			return 0; /* keep the lock */
		}
		if ( !(QMATCH & e) ) {
			if ( e )
				f->e[i] = --e;
			else {
				TRACEKEY( ERR_TRASH, f );
				return eRr(ERR_IDIOT,
					"ran on left bound of block %u ent %u",
					b->num, b->ent );
			}
		}
		d = b->dct + (QENTMASK & e);
		/* direct access ok in fork */
		pkey = ((unsigned char*)b)+d->pos;
		pkln = d->kln;
		pvln = d->vln ? f->dx->vsz : 0;
		n = GETINT( pkey+pkln+pvln );
		if ( --i )
			b = getfork(&fork[0].b, f->dx, n, i);
		else {
			QLOCKUN(1);
			b = GETLEAF(leaf, f->dx, n, lck); /* read 1st leaf */
			/* continue only to do the LY */
		}
	}
done:
	QLOCKUN(1);
	return ret; /* go away */
}	/* findkey */





/* ************************************************************
	public functions
*/

int qRemake ( Qdx *dx )
{
	Key last;
	Bloke child;
	Fork fork;
	unsigned char lev = 0; /* the current child level to be indexed */
	unsigned left = 0; /* block on child level */
	unsigned iblk = 0;
	unsigned ikey = 0;
	unsigned ikeyl = 0;
	unsigned ipfx = 0;
	unsigned ispn = 0;

	fTrunc(&dx->mqx.fil, 0);
	fMap(&dx->mqx, 0);
	dx->fln = 1; /* pretend we had a root */
	/* build tree */
	for (dx->fln = 1;;lev++) { /* levels */
		unsigned  len = 0; /* current is n'th block on level */
		Block    *b = 0;
		unsigned char *base = 0; /* == b */
		unsigned  have = 0; /* room in current block b */
		Dict     *d = 0; /* of current block b */
		unsigned  next = left; /* child */

		left = dx->fln;
		/* get child c and create child pointer in block b */
		do {
			unsigned char  vln = 0; /* actual byte length: dx->vsz or 0 */
			unsigned char *key;
			unsigned char  kln;
			unsigned  need;
			Block *c = lev
				? getfork(&child.b, dx, next, lev)
				: GETLEAF(&child.b, dx, next, 0);/* must have full ex lock */
			if ( ! c )
				return eRr(ERR_NOMEM, "no child block" );
			next = c->nxt;
			if ( ! c->ent ) /* OOPS empty node */
				continue;
			key = GETENT(c, 0);
			kln = c->dct[0].kln;
			if ( ! b ) /* the leftmost child on this level */
				kln = 0; /* always 0 key on left */
			else if ( lev ) /* the child is a fork -- have to carry vals ... */
				vln = c->dct[0].vln ? dx->vsz : 0; /* ... if the child starts with */
			else if ( last.len == kln /* compare key to last of prev. ... */
				&& !memcmp( last.byt, key, kln ) /* ... which was saved in last */
			) { /* key spans blocks */
				LOG_DBG( LOG_DEBUG, "found span key '%b' child %u",
					kln, key, c->num );
				vln = dx->vsz; /* qualify with value */
			} else {
				unsigned char x = 0;
				/* find prefix key */
				if ( kln > last.len )
					kln = last.len;
				while ( x < kln && key[x] == last.byt[x] )
					x++;
				if ( x < kln ) /* they differ at [x] */
					kln = x+1;
				else /* equal on the shorter length */
					kln++;
				if ( kln > c->dct[0].kln )
					return eRr(ERR_TRASH,
						"bad key '%b' on level %u matching previous of length %u",
						c->dct[0].kln, key, lev, last.len );
				LOG_DBG( LOG_DEBUG, "prefix on key '%b' child %u save %u",
					kln, key, c->num, c->dct[0].kln - kln );
				ipfx += c->dct[0].kln - kln;
			}
			if ( vln )
				ispn++;
			ikeyl += kln;
			need = sizeof(Dict) + kln + vln + 4;
			if ( need > have ) {
				if ( b ) {
					eRr(LOG_TRACE, "level %u block %u is full %u ent stack %u",
						lev, b->num, b->ent, b->stk );
					b->nxt = dx->fln; /* link */
					if ( wrtfork(b, dx) )
						goto nowrite;
					len++;
				}
				b = newfork(&fork.b, dx, lev+1);
				base = (unsigned char*)b;
				have = env.psz - DICTOFF; /* - ql->let; TODO: should we ??? */
				d = b->dct;
				len++;
				iblk++;
			}
			SETINT( base+(b->stk -= 4), c->num );
			d->kln = kln;
			d->vln = vln ? 1 : 0;
			if ( kln += vln ) /* key + consecutive qualifying value */
				memcpy( base+(b->stk -= kln), key, kln );
			d->pos = b->stk;
			d++;
			b->ent++;
			ikey++;
			have -= need;
			/* save the last dance for me */
			memcpy(last.byt, GETENT(c, c->ent-1), last.len = c->dct[c->ent-1].kln);
		} while (next);

		if ( !len )
			return eRr(ERR_TRASH, "level %u was empty", lev );
		/* close level */
		eRr(LOG_INFO, "closing level %u with %u blocks %u..%u",
			lev, len, left, b->num );
		if ( 1 == len )
			b->num = 0; /* make it the root */
		if ( wrtfork(b, dx) )
			goto nowrite;
		if ( 1 == len ) {
			dx->dpt = b->lev;
			break;
		}
	}
#define PM(a,b) (unsigned)(a*LULU(1000)/b) /* per mille */
	if (ikey && ikeyl) /* divisors */
	eRr(LOG_INFO,
		"\tused %u fork blocks %u keys %u bytes avg %d\n"
		"\t%u #spans (%d/oo) %u pfx (- %d/oo)",
		iblk, ikey, ikeyl, ikeyl/ikey,
		ispn, PM(ispn*dx->vsz,ikeyl), ipfx, PM(ipfx,(ipfx+ikeyl)) );
	dx->fln--; /* counted one over */
	fMap(&dx->mqx, dx->fln);

	return 0;
nowrite:
	dx->fln = 0;
	return eRr(LOG_SYSERR, "\twhen writing tree block" );
}	/* qRemake */




static int load ( Qdx *dx, unsigned char *key )
{
	int cmp = 0, keyl = 0, l;
	QLoad *ql = dx->qld;
	if (key) {
		keyl = *key - dx->vsz;
		l = keyl < ql->key.len ? keyl : ql->key.len; /* min */
		if ( !(cmp = memcmp(key+1, ql->key.byt, l)) )
			cmp = keyl - ql->key.len;
		if (0 > cmp)
			return eRr(LOG_ERROR, "qload key '%b' < last '%b'",
				keyl, key+1, ql->key.len, ql->key.byt );
	}
	/* flush ? */
	while ( !key /* end */
		|| QENTMAX == ql->got /* full */
		|| cmp /* next key */
	) {
		Block *b = &ql->cur.b;
		unsigned have = b->stk - DICTOFF - b->ent*sizeof(Dict) - ql->let;
		unsigned need = sizeof(Dict) + ql->key.len;
		unsigned want = need + ql->got * dx->vsz;
		unsigned take = ql->got;
		if ( have < want ) { /* you can't always get ... */
			if ( have <= need )
				take = 0;
			else {
				take = (have - need) / dx->vsz;
				if ( 256 > take * dx->vsz )
					take = 0;
			}
		}
		LOG_DBG( LOG_DEBUG, "qload '%b' got %u need %u want %u have %u take %u",
			ql->key.len, ql->key.byt, ql->got, need, want, have, take );
		/* we have enough room in b for take entries */
		if ( take ) {
			unsigned char *base = (unsigned char*)b;
			unsigned char *s = base + b->stk;
			int i = take;
			Dict *d = &b->dct[ b->ent++ ];
			while ( i-- )
				memcpy( s -= dx->vsz, ql->val[i].byt, dx->vsz );
			memcpy( s -= ql->key.len, ql->key.byt, ql->key.len );
			b->stk = (unsigned short) (s - base);
			LEAFSETPV( d, b->stk, take );
			d->kln = ql->key.len;
			if ( ql->got > take )
				memcpy( ql->val, ql->val+take, sizeof(ql->val[0])*(ql->got - take) );
			ql->got -= take;
		}
		if ( ! ql->got )
			break;
		if ( take ) /* key spans blocks */
			ql->span++;
		ql->free += b->stk - DICTOFF - b->ent*sizeof(Dict); /* wasted */
		b->nxt = dx->lln; /* link */
		LOG_DBG( LOG_DEBUG, "qload: %u entries left, getting block %u",
			ql->got, dx->lln );
		if ( wrtleaf(b, dx) )
			return eRr(LOG_SYSERR, "\twhen writing qload block" );
		/* so couldn't flush all, thus we need another block */
		eRr(LOG_TRACE, "new qload leaf %d", dx->lln);
		newleaf(b = &ql->cur.b, dx);
		/* will continue on end or keychange */
		LOG_DBG( LOG_TRACE, "new block %u has %u ent %u stk, will keep %u",
			b->num, b->ent, b->stk, ql->let );
	}
	if (cmp || !key) /* done with key: stats */
		if (ql->maxv < ql->tot)
			ql->maxv = ql->tot;
	if (key) {
		if (cmp) {	/* set new key */
			if (ql->got)
				return eRr(ERR_IDIOT, "qload not empty on new key");
			ql->tot = 0;
			memcpy(ql->key.byt, key+1, ql->key.len = keyl);
			ql->keys++;
			ql->keyl += keyl;
			if (ql->maxl < (unsigned)keyl)
				ql->maxl = keyl;
		}
		/* add new val */
		/* LOG_HEX( key->val.byt, dx->vsz ); */
		memcpy(ql->val[ ql->got++ ].byt, key+1+keyl, dx->vsz);
		ql->vals++;
		ql->tot++;
	} else {
		/* stats */
		unsigned size = dx->lln * dx->lsz;
		unsigned sval = ql->vals * dx->vsz;
		unsigned sove = dx->lln * DICTOFF + sizeof(Dict)*(ql->keys+ql->span);

		if (size && ql->keys) /* divisors */
		eRr(LOG_INFO,
			"end qload:\n"
			"\t#blks %u total len %u free %u (%d/oo) overhead %u (%d/oo)\n"
			"\t#keys %u total len %u (%d/oo) avg %d max %u\n"
			"\t#vals %u total len %u (%d/oo) avg %d max %u spans %u",
			dx->lln, size, ql->free, PM(ql->free,size), sove, PM(sove,size),
			ql->keys, ql->keyl, PM(ql->keyl,size), ql->keyl/ql->keys, ql->maxl,
			ql->vals, sval, PM(sval,size), ql->vals/ql->keys, ql->maxv, ql->span
		);
		wrtleaf(&ql->cur.b, dx);
	}
	
	return 0;
}	/* load */


static int ncomp;
#ifdef SORT_Q
static int compar (const void *pa, const void *pb)
{
	unsigned char *a = *(unsigned char **)pa;
	unsigned char *b = *(unsigned char **)pb;
	int l = *a, cmp;
	if (l > *b) l = *b;
	ncomp++;
	return (cmp = memcmp(a+1, b+1, l)) ? cmp : (int)*a - (int)*b;
}
#endif
static OPT_INLINE int keyGt (unsigned char *a, unsigned char *b)
{
	int l = *a, cmp;
	if (l > *b) l = *b;
	ncomp++;
	return 0 < (cmp = memcmp(a+1, b+1, l)) || (!cmp && *a > *b);
}


static int sort (Qdx *dx, unsigned char **key, int len)
{
	unsigned char rev[SORTKEYS];
	unsigned char *tmp;
	int i, ret = 0;
	lolo tm;

	tUpd(&tm);
	if (2 > len)
		goto cp;
#ifdef SORT_Q
	qsort(key, len, sizeof key[0], compar);
#else
#	define SORT_WH
	/* basic weak-heap-sort (Dutton 93) just like rSort; c.f. uti.c
	``On the Performance of WEAK-HEAPSORT
	http://www.informatik.uni-freiburg.de/~edelkamp/publications/Weakheap.pdf
	for tuned versions see
	http://www.jea.acm.org/2002/EdelkampHeapsort/
	and ``Pushing the Limits in Sequential Sorting
	http://www.informatik.uni-freiburg.de/~edelkamp/publications/Improvements.pdf
	*/
	memset(rev, 0, len);
	/* weak-heap-forestify
		every node roots a weak heap AND is <= root
		(especially the leftist independent roots are <= root)
	*/
	for ( i=len; --i; ) {
		int g = i; /* granma is 1st left i or root, if i is totally leftist ;) */
		while (!(1&g)) /* we're left child */
			g >>= 1; /* shift granma right :( */
		g >>= 1;
		if (!keyGt(key[i], key[g])) /* merge: swap if g < i */
			continue;
		tmp = key[i]; key[i] = key[g]; key[g] = tmp; /* swap */
		rev[i] = 1;
	}

	/* re-weak-heap-forestify */
	for ( i=len; --i>1; ) {
		int j=1, left;
		/* rev[i] = 0; i's REVBIT no longer needed */
		/* follow path of left childs (independent roots) */
		while (i > (left = (j<<1) + rev[j]))
			j = left;
		/* put max of them and i (new root) at i */
		do {
			if (!keyGt(key[j], key[i])) /* merge: swap if i < j */
				continue;
			tmp = key[i]; key[i] = key[j]; key[j] = tmp; /* swap */
			rev[j] ^= 1; /* toggle revbit */
		} while (j >>= 1);
	}
	/* rev[1] = 0; 2... have been cleared */
	key[len] = key[0]; /* initial root still untouched biggest */
#endif
#ifndef SORT_WH
	key--;
#endif
	eRr(LOG_VERBOSE, "%d sorted %d keys with %d comparisions -- loading ...",
		tUpd(&tm), len, ncomp);
cp:
	/* finally cp to dest */
	if (FIL_NONE == dx->qld->fil) { /* internal -- load */
		unsigned char **e = key+len+1;
		while (++key != e)
			if ((ret = load(dx, *key)))
				goto err;
		eRr(LOG_VERBOSE, "%d loaded", tUpd(&tm));
	} else { /* external -- write MERGECHK blks */
		unsigned char **e = key+len+1;
		unsigned char *out = dx->qld->mout, *end = out+MERGEBLK, *dst = out;
		unsigned blks = 0;

		while (++key != e) {
			unsigned l = 1+**key;
			if (end <= dst+l) {
				*dst = 0;
				if (0 > (ret = fWrite(&dx->qld->fil, out, MERGEBLK)))
					goto err;
				dst = out;
				if (MERGECHK == ++blks)
					return eRr(ERR_IDIOT, "CHUNK!");
			}
			memcpy(dst, *key, l);
			dst += l;
		}
		do {
			*dst = 0;
			fWrite(&dx->qld->fil, out, MERGEBLK);
			dst = out;
		} while (MERGECHK != ++blks);
		eRr(LOG_VERBOSE, "%d wrote %d blocks ...", tUpd(&tm), blks);
		dx->qld->blks += MERGECHK;
		ret = 0;
	}
err:
	return ret;
}	/* sort */


static int merge (Qdx *dx, QLoad *ql)
{
#ifndef MERGEMULTIPASS
#	define chunk MERGECHK
#else
	unsigned char *const out = ql->mout, *const end = out+MERGEBLK;
	unsigned chunk = MERGECHK, pass = 1;
	file fil2 = FIL_NONE;
#endif
	lolo tm;
	int ret = 0;

	eRr(LOG_VERBOSE, "%d-way merging %d blocks in %d chunks of %d ...",
		MERGEWAY, ql->blks, (ql->blks+MERGECHK-1)/MERGECHK, MERGECHK);
	tUpd(&tm);
	ncomp = 0;
#ifndef MERGEMULTIPASS
	{ /* single pass */
#else
	if (ql->blks > MERGECHK*MERGEWAY) /* multipass */
		fOpen(&fil2, 0, 0);
	for (;; pass++) {
#endif
		unsigned blkno = 0;
#ifndef MERGEMULTIPASS
#	define inf &ql->fil
		{ /* single run in single pass */
#else
		unsigned nchunk = chunk*MERGEWAY; /* lastrun, if >= ql->blks */
		int lastpass = nchunk >= ql->blks, run = 1;
		file *inf, *outf;

		if (1 & pass) {
			inf = &ql->fil; outf = &fil2;
		} else {
			inf = &fil2; outf = &ql->fil;
		}
		if (!lastpass)
			fSeek(outf, 0);
		while (blkno < ql->blks) {
			/* merge MERGEWAY chunks of length chunk to n chunks of size nchunk */
			/* only one run in last pass */
			unsigned char *dst = out;
			unsigned blks = 0;
#endif
#ifdef MERGEDEBUG
			unsigned char lastkey[256];
#endif
			struct {
				unsigned char *blk; /* merge blk */
				unsigned char *key; /* current key */
				unsigned nxt; /* next block no */
				unsigned end; /* end */
			} in[MERGEWAY], *nxt[MERGEWAY], *tmp;
			unsigned ways = 0, i;

			for (ways=0; ways<MERGEWAY; ways++) { /* prefill */
				if (blkno >= ql->blks)
					break;
				in[ways].key = in[ways].blk = ql->buf.m[ways];
				if (MERGEBLK != (ret =
					fPread(inf, in[ways].blk, MERGEBLK, blkno*MERGEBLK)))
					goto done;
				in[ways].nxt = blkno+1;
				if (ql->blks < (blkno += chunk)) blkno = ql->blks;
				in[ways].end = blkno;
				/* insert */
				for (i=0; i<ways && !keyGt(nxt[i]->key, in[ways].key); ) i++;
				if (i < ways) memmove(nxt+i+1, nxt+i, (ways-i)*sizeof nxt[0]);
				nxt[i] = in+ways;
			}
#ifndef MERGEMULTIPASS
			eRr(LOG_VERBOSE, "%d prepared %d blocks %d ways", tUpd(&tm), blkno, ways);
#else
			eRr(LOG_VERBOSE, "%d prepared %d/%d blocks %d ways run %d of pass %d%s",
				tUpd(&tm), blkno, ql->blks, ways, run, pass, lastpass?" (last)":"");
#endif
#ifdef MERGEDEBUG
			lastkey[0] = 0;
			for (i=0; i<ways; i++)
				eRr(LOG_VERBOSE, "%d %b", i, *nxt[i]->key, nxt[i]->key+1);
#endif
			for (;;) { /* merge */
				int len = *nxt[0]->key + 1;
				/* stuff away 1st key */
#ifdef MERGEDEBUG
				if (keyGt(lastkey, nxt[0]->key)) {
					ret = eRr(ERR_IDIOT, "'%b' > '%b' way %d",
						*lastkey, lastkey+1, *nxt[0]->key, nxt[0]->key+1, nxt[0]-in);
					goto done;
				}
				memcpy(lastkey, nxt[0]->key, len);
#endif
#ifdef MERGEMULTIPASS
				if (lastpass) {
#endif
					if ((ret = load(dx, nxt[0]->key))) goto done;
#ifdef MERGEMULTIPASS
				} else {
					if (len >= end - dst) { /* flush out block */
#	ifdef MERGEDEBUG
						eRr(LOG_VERBOSE, "out blk %d len %d %b", blks, dst-out, *out, out+1);
#	endif
						*dst = 0; /* terminate */
						if (MERGEBLK != (ret = fWrite(outf, dst = out, MERGEBLK)))
							goto done;
						blks++;
					}
					memcpy(dst, nxt[0]->key, len);
					dst += len;
				}
#endif
				/* reload nxt[0] */
				nxt[0]->key += len;
				if (!*nxt[0]->key) { /* eob */
					if (nxt[0]->nxt != nxt[0]->end) {
						if (MERGEBLK != (ret = fPread(inf, nxt[0]->key = nxt[0]->blk,
							MERGEBLK, nxt[0]->nxt++ * MERGEBLK)))
							goto done;
					}
					if (!*nxt[0]->key) { /* done with this stream */
						if (!--ways)
							break;
						memmove(nxt, nxt+1, ways*sizeof nxt[0]);
						continue;
					}
				}
#ifdef MERGEDEBUG
				if (keyGt(lastkey, nxt[0]->key)) {
					ret = eRr(ERR_IDIOT, "'%b' > '%b' in blk %d off %d way %d",
						*lastkey, lastkey+1, *nxt[0]->key, nxt[0]->key+1,
						nxt[0]->nxt-1, nxt[0]->key-nxt[0]->blk, nxt[0]-in);
					goto done;
				}
#endif
				if (1 == ways) continue;
				/*
				this may look like pretty brute force.
				However, in practice we rarely see reinsertions at all
				and get by with about 1.1 comparisions per key.
				With sorted input we have even less than 1 per key,
				due to the continue (last exit oneway) above.
				*/
				for (i=1; i<ways && keyGt(nxt[0]->key, nxt[i]->key); ) i++;
				/* nxt[0]->key is not gt nxt[i]->key: move to --i */
				if (--i) {
					tmp = nxt[0];
					memmove(nxt, nxt+1, i*sizeof nxt[0]);
					nxt[i] = tmp;
				}
			}	/* merge */
			ret = 0;
#ifndef MERGEMULTIPASS
			eRr(LOG_VERBOSE, "%d merged %d comparisions", tUpd(&tm), ncomp);
		}	/* single run */
#else
			eRr(LOG_VERBOSE, "%d merged %d comparisions %d/%d blocks pass %d",
				tUpd(&tm), ncomp, blks, nchunk, pass);
			if (lastpass)
				goto done;
			if (nchunk <= blks)
				eRr(ERR_IDIOT, "wrote %d >= %d blks pass %d", blks, nchunk, pass);
			do {
				*dst = 0;
				fWrite(outf, out, MERGEBLK);
				dst = out;
			} while (nchunk != ++blks);
			run++;
		}	/* for new chunks */
		chunk = nchunk;
#endif
	}	/* pass */
done:
#ifdef MERGEMULTIPASS
	if (FIL_NONE != fil2) fClose(&fil2);
#endif
	return ret;
}	/* merge */


int qLoad ( Qdx *dx, Key *key )
{
	QLoad *ql = dx->qld;
	int ret, l = key->len+dx->vsz;

	if ( !ql ) {
		if (! key->val.len) return 0; /* double close ? */
		ncomp = 0;
		if (ENV_EXCL!=env.wri || !(DX_WRITE & dx->flg))
			return eRr(ERR_TRASH, "need excl write access to load index");
		ql = dx->qld = (QLoad*)mAlloc(sizeof (QLoad));
		if ( dx->let > 50 )
			dx->let = 50;
		ql->let = dx->let * dx->lsz / 100;
		fTrunc(&dx->mqd, 0);
		dx->lln = 0;
		newleaf(&ql->cur.b, dx);
		ql->fil = FIL_NONE;
		ql->nkey = ql->buf.s.keyb;
		ql->nmem = ql->buf.s.memb;
		eRr(LOG_INFO,
			"loading %d/%dK mem, %d keys, %dK blocks in %d-chunks, %d-way merge"
#ifdef MERGEMULTIPASS
			" (multipass)"
#endif
			, SORTUSE>>10, SORTMEM>>10, SORTKEYS, MERGEBLK>>10, MERGECHK, MERGEWAY
		);
	}
	if ( !key->val.len /* done */
		|| SORTKEYS-2 <= ql->nkey-ql->buf.s.keyb /* reserve 1 for WHS */
		|| SORTUSE <= ql->nmem-ql->buf.s.memb
	) {
		if (key->val.len && FIL_NONE == ql->fil)
			fOpen(&ql->fil, 0, 0);
		if ((ret = sort(dx, ql->buf.s.keyb, ql->nkey - ql->buf.s.keyb))
			|| !key->val.len
		)
			goto done;
		ql->nkey = ql->buf.s.keyb;
		ql->nmem = ql->buf.s.memb;
	}
	/* add key to sort buffer */
	if (key->val.len == dx->vsz && l < 256) {
		*ql->nkey++ = ql->nmem;
		*ql->nmem = (unsigned char)l;
		memcpy(++ ql->nmem, key->byt, key->len);
		memcpy(ql->nmem += key->len, key->val.byt, dx->vsz);
		ql->nmem += dx->vsz;
		return 0;
	}
	ret = eRr(ERR_INVAL, "bad klen %d vlen %d", key->len, key->val.len);
done:
	if (FIL_NONE != ql->fil) {
		if (!ret) ret = merge(dx, ql);
		fClose(&ql->fil);
	}
	if (!ret) {
		load(dx, 0); /* close last leaf */
		qRemake(dx); /* build tree */
	}
	mFree(dx->qld);
	dx->qld = 0;
	return ret;
}	/* qLoad */


int qLoadf ( Qdx *dx, file *f )
{
	Key key;
	lolo tm;

	FIL_DEFBUF(*f);
	memset( &key, 0, sizeof(key) );
	key.val.len = dx->vsz;
	tUpd(&tm);
	while ( fGets(&fb) && fb.l ) {
		if ( qMkKeyVal(dx, &key, fb.p, fb.l) ) {
			eRr(LOG_VERBOSE, "bad key at line %d offset %d", fb.n, fb.o);
			continue;
		}
		if ( qLoad( dx, &key ) )
			break;
		if ( !(fb.n & 0x3ff) )
			eRr(LOG_VERBOSE, "%dK lines", fb.n >> 10 );
	}
	eRr(LOG_INFO, "%d did %d lines %d bytes (%s)", tUpd(&tm), fb.n, fb.o,
		FIL_NONE == fb.f ? "eof" : "blank" );
	*f = fb.f;
	key.val.len = 0; /* close it */
	return qLoad(dx, &key);
}	/* qLoadf */


/* TODO: this function is an awful mess and needs complete rewrite, KR
*/
int qSet ( Qdx *dx, Key *key )
{
	Bloke blk[2];
	Key up;
	Block * b = &blk[0].b; /* mainblock */
	unsigned char * base = blk[0].base;
	int ret = 0;
	unsigned short e, i = 0;
	Block *s = 0; /* sibling */
	Dict *d;
	unsigned char *sbase, *src, *v, *stk;
	FindKey f;
	int hadkey, split = 0, insibling = 0;
	unsigned need;
	unsigned sibnum = 0;
	unsigned char lev = 0;
	PrVal pv;
#ifdef BUILD_SHMODE /* locks we hold on block, sibling, tree */
	int lb=-1, ls=-1, lt=0; 
#endif

	f.dx  = dx;
	f.key = key;
	if ( dx->ksz && key->len > dx->ksz )
		key->len = dx->ksz;
	if ( FINDKEY(b, &f, LOCK_WRITE) ) {
		b = 0;
		DONE((ERR_TRASH, "could not find key"));
	}
#ifdef BUILD_SHMODE
	lb = b->num;
#endif
	e = (QENTMASK & f.e[0]);
	d = b->dct + e;
	if ( ! (hadkey = QMATCH & f.e[0]) ) /* new key */
		need = key->len + sizeof(Dict) + dx->vsz;
	else {
		unsigned vln = LEAFVLN(d);
		v = base+LEAFPOS(d)+d->kln;
		if ( 1 == vln ) { /* check stopword -- value entirely zeroed */
			int l = dx->vsz;
			unsigned char *C = v;
			while ( l-- && !*C++ ) /* huh huh, he said "C++" */
				;
			if ( -1 == l ) {
				LOG_DBG(LOG_VERBOSE, "key '%b' is a stopword", key->len, key->byt);
				return QST_STPWRD;
			}
			/* TODO: check QST_UNIKEY */
		}
		i = valpos( v, vln, dx->vsz, key->val.byt );
		if ( QMATCH & i ) {
			LOG_DBG( LOG_DEBUG,
				"adding key '%b'/%s: already there pos %u",
				key->len, key->byt, PRVAL( pv, key->val ), i & ~QMATCH );
			return QST_FOUND; /* value already there */
		}
		need = dx->vsz;
	}
	LOG_DBG( LOG_DEBUG,
		"adding key '%b'/%s in block %d had %d ent %d stk %d need %d",
		key->len, key->byt, PRVAL( pv, key->val ),
		b->num, hadkey, b->ent, b->stk, need );

	if ( need > b->stk - DICTOFF - b->ent*sizeof(Dict) ) {
		/* now that's bad: not enough space
			create new block s, move last entries there
			starting from the end, entries (including the new one) are collected,
			until their combined size reaches more than half a block.
			If the combined size then is more than 2/3 a block,
			and the last entry alone takes more than 1/3 a block, it is split.
			TODO: check existing sibling for free space
		*/
		unsigned half  = (dx->lsz - DICTOFF) >> 1;
		unsigned tot   = 0;
		unsigned mv = b->ent;

		LOG_DBG( LOG_DEBUG, "splitting block %d need %d have %d (stk %d ent %d)",
			b->num, need, b->stk - DICTOFF - b->ent*sizeof(Dict), b->stk, b->ent );
		if ( need > (dx->lsz - DICTOFF) / 3 )
			DONE((ERR_IDIOT, "OOPS! excessive need for %d bytes", need));
		/* figure out split position */
		if ( !hadkey && b->ent == e ) {
			tot = need;
			insibling = !0;
		}
		d = b->dct+mv;
		while ( d--, mv-- ) {
			unsigned sz = sizeof(Dict)+d->kln+LEAFVLN(d)*dx->vsz;
			if ( half > (tot += sz) ) { /* sz fits */
				if ( e == mv && !hadkey ) {
					/*
						check size of new entry.
						In case of hadkey, we don't bother for the dx->vsz bytes.
					*/
					if ( half < (tot + need) ) {
						if ( (half - tot) > need/2 ) { /* more balanced if we move */
							tot += need;
							insibling = !0;
						}
						break;
					}
					tot += need;
					insibling = !0;
				}
				if ( mv )
					continue; /* else croak on ! mv below */
			}
			/* exceeded half -- maybe we move entry mv anyway */
			if ( tot > (dx->lsz - DICTOFF)*2/3 ) {
				if ( tot-sz < (dx->lsz - DICTOFF)/3 )
					split = !0;
				else { /* undo -- don't move this entry */
					tot -= sz;
					d++; mv++;
				}
			} else if ( ! mv )
				DONE((ERR_IDIOT, "OOPS! didn't find end tot %d bytes", tot));
			break;
		}	/* while mv-- */
		if ( mv < e || (mv == e && hadkey && !split) )
			insibling = !0;
		/* now mv is the first entry to move */
		/* create new leaf block dx->lln */
#ifdef BUILD_SHMODE
		for (;;) {
			int size;
			QLOCKEX(dx->lln<<1);
			size = fSize( dx->mqd );
			if ( size == (int)(dx->lln * dx->lsz) ) {
				ls = dx->lln;
				break;
			}
			QLOCKUN(dx->lln<<1);
			if ( size % dx->lsz || size < (int)(dx->lln * dx->lsz) )
				DONE((ERR_TRASH, "bad mqd size 0x%08x", size));
			dx->lln = size / dx->lsz;
		}
		/* have lock on new leaf block dx->lln */
#endif
		newleaf(s = &blk[1].b, dx);
		sbase = (unsigned char*)s; /* == blk[1].base */
		LOG_DBG( LOG_DEBUG,
			"split: move %d of %d to new blk %d e %d tot %d insibling %d split %d",
			mv, b->ent, s->num, e, tot, insibling, split );
		s->nxt = b->nxt;
		b->nxt = s->num;
		if ( split ) {
			/* now total includes full size of entry mv.
				if we included a new entry, it might even exceed lsz.
				calculate sizes for both blocks w/o mv's values.
			*/
			unsigned esz = sizeof(Dict)+d->kln; /* entry's base cost */
			unsigned vln = LEAFVLN(d);
			unsigned keep, diff, mvval, kpval, mvsiz, pos = LEAFPOS(d);
			/*
				from currently used space, substract tot (including all
				values and possibly new key "need"), but add entry base cost.
				that should be the size remaining if we where moving
				all of the values but keep an entry and key.
				maybe slightly biased if need is dx->vsz (the hadkey case).
			*/
			keep = dx->lsz - b->stk + b->ent*sizeof(Dict) + need - tot + esz;
			tot -= vln * dx->vsz; /* undo entries */
			diff = (keep > tot ? keep - tot : tot - keep) / dx->vsz;
			if ( diff > vln )
				DONE((ERR_IDIOT, "OOPS! got diff %d vln %d entry %d tot %d keep %d",
					diff, vln, mv, tot, keep));
			mvval = (vln - diff)/2;
			if ( keep > tot )
				mvval += diff;
			kpval = vln - mvval;
			if ( ! mvval || ! kpval ) {
				LOG_DBG( LOG_WARN, "bad mv/kpval %d/%d keep %d tot %d vln %d diff %d",
					mvval, kpval, keep, tot, vln, diff );
				if ( ! mvval )
					kpval = vln - (mvval = 1);
				else
					mvval = vln - (kpval = 1);
			}
			tot += mvsiz = mvval*dx->vsz;
			keep += kpval*dx->vsz;
			src = base+pos;
			LOG_DBG( LOG_DEBUG,
				"split: entry %d '%b' mv %d of %d values, i %d keep %d tot %d",
				mv, d->kln, src, mvval, vln, i, keep, tot );
			if ( keep > (dx->lsz-DICTOFF) || tot > (dx->lsz-DICTOFF) )
				DONE((ERR_IDIOT,
					"OOPS! got diff %d vln %d mvval %d entry %d tot %d keep %d",
					diff, vln, mvval, mv, tot, keep));
			/* SIGH -- do it */
			memcpy( sbase + (s->stk -= mvsiz),
				src+d->kln+kpval*dx->vsz, mvsiz );
			if ( e == mv && hadkey && i > kpval ) {
				insibling = !0;
				s->stk -= dx->vsz;
				i -= kpval;
				memmove( sbase+s->stk, sbase+s->stk+dx->vsz, i*dx->vsz );
				memcpy( sbase+s->stk+i*dx->vsz, key->val.byt, dx->vsz );
				mvval++;
			}
			memcpy( sbase + (s->stk -= d->kln), src, d->kln );
			LEAFSETPVK( s->dct, s->stk, mvval, d->kln );
			s->ent = 1;
			memmove( src + mvsiz, src, d->kln+kpval*dx->vsz );
			pos += mvsiz;
			LEAFSETPV( d, pos, kpval );
			d++; mv++;
		}
		/* now entries mv..b->ent are moved, mv > 0 */
		b->stk = LEAFPOS(b->dct+mv-1); /* pre-adjust stack */
		if ( mv < e ) { /* move entries mv..e-1 */
			/* used stack space is from pos of entry e-1 to end of entry mv
				(might differ from pos of mv-1 due to split)
			*/
			unsigned from = LEAFPOS(b->dct+e-1), nent = e-mv,
				sz = LEAFPOS(b->dct+mv)+b->dct[mv].kln
					+LEAFVLN(b->dct+mv)*dx->vsz - from;
			int adj = (s->stk -= sz) - from; /* entry position adjustment */
			memcpy( sbase + s->stk, base+from, sz );
			memcpy( s->dct+s->ent, b->dct+mv, nent*sizeof(Dict) );
			d = s->dct + s->ent;
			s->ent += nent;
			for ( ; nent--; d++ )
				LEAFPOSADD( d, adj );
		}
		if ( insibling && mv <= e ) { /* now mv or insert e, unless done on split */
			d = b->dct + e;
			if ( hadkey ) { /* mv existing entry e >= mv, add new val */
				/* i is valpos */
				unsigned vln = LEAFVLN(d);
				src = base+LEAFPOS(d)+d->kln;
				if ( i < vln ) {
					unsigned sz = (vln - i)*dx->vsz;
					memcpy( sbase + (s->stk -= sz), src + i*dx->vsz, sz );
				}
				memcpy( sbase + (s->stk -= dx->vsz), key->val.byt, dx->vsz );
				if ( i ) {
					unsigned sz = i*dx->vsz;
					memcpy( sbase + (s->stk -= sz), src, sz );
				}
				memcpy( sbase + (s->stk -= d->kln), base+LEAFPOS(d), d->kln );
				LEAFSETPVK( s->dct + s->ent, s->stk, vln+1, d->kln );
				e++; /* e is moved */
			} else {
				memcpy( sbase + (s->stk -= dx->vsz), key->val.byt, dx->vsz );
				memcpy( sbase + (s->stk -= key->len), key->byt, key->len );
				LEAFSETPVK( s->dct + s->ent, s->stk, 1, key->len );
			}
			s->ent++;
		} /* mv or insert e */
		if ( e < b->ent && mv < b->ent ) { /* move entries max(e,mv) to end */
			unsigned fst = mv < e ? e : mv;
			/* used stack space is from pos of entry b->ent-1 (the original b->stk)
				to end of entry fst (might differ from pos of fst-1 due to split)
			*/
			unsigned from = LEAFPOS(b->dct+b->ent-1), nent = b->ent-fst,
				sz = LEAFPOS(b->dct+fst)+b->dct[fst].kln
					+LEAFVLN(b->dct+fst)*dx->vsz - from;
			int adj = (s->stk -= sz) - from; /* entry position adjustment */
			LOG_DBG( LOG_DEBUG, "split: entries %d..%d sz %d from %d to %d",
				fst, b->ent-1, sz, from, s->stk );
			memcpy( sbase + s->stk, base+from, sz );
			memcpy( s->dct+s->ent, b->dct+fst, nent*sizeof(Dict) );
			d = s->dct + s->ent;
			s->ent += nent;
			for ( ; nent--; d++ ) {
#if !defined(NDEBUG)
				unsigned p = LEAFPOS(d);
#endif
				LEAFPOSADD( d, adj );
#if !defined(NDEBUG)
				LOG_DBG( LOG_DEBUG, "new ent %d p/v/k %d/%d/%d o %d",
					d - s->dct, LEAFPOS(d), LEAFVLN(d), d->kln, p );
#endif
			}
		}
		b->ent = mv; /* kill entries */
		LOG_DBG( LOG_DEBUG,
			"split: blks %d/%d with %d/%d entries, s starts '%b'",
			b->num, s->num, b->ent, s->ent, s->dct[0].kln, sbase+LEAFPOS(s->dct) );
	}	/* new sibling */

	if ( ! insibling ) {
		/* now we do have enough space */
		d = b->dct + e;
		stk = base + b->stk; /* bottom of data stack */
		if ( hadkey ) { /* insert value at i */
			unsigned char *val = base + LEAFPOS(d)+d->kln+i*dx->vsz;
			need = dx->vsz;
#if 0
		LOG_DBG( LOG_DEBUG,
			"in entry %d '%b' val %d: memmove( %p %p %d) val %p need %d",
			e, d->kln, base+LEAFPOS(d), i,
			stk-need, stk, (val-stk)+need, val, need );
#endif
			memmove( stk - need, stk, (val - stk)+need );
			memcpy( val - need, key->val.byt, dx->vsz );
			LEAFVLNADD( d, 1 );
		} else { /* insert key, value and dict entry */
			/* take dict and data space from entry e */
			unsigned char *nu;
			need = key->len + dx->vsz;
			if ( e == b->ent )
				nu = base + b->stk - need;
			else {
				/* at end of next entry, i.e. end of block or beg of previous */
				nu = base + LEAFPOS(d) + d->kln + LEAFVLN(d)*dx->vsz;
				memmove( stk - need, stk, nu - stk );
				nu -= need;
				memmove( d+1, d, (b->ent - e)*sizeof(Dict) );
			}
			memcpy( nu, key->byt, key->len );
			memcpy( nu + key->len, key->val.byt, dx->vsz );
			LEAFSETPV( d, nu-base, 1 );
			d->kln = key->len;
			d++;
			e++;
			b->ent++;
		}
		/*
			need is set to the needed (i.e. additionally used) data space,
			e is the first dict entry to redirect need bytes lower
		*/
		b->stk -= need;
		for ( ; e < b->ent; e++, d++ )
			LEAFPOSADD( d, -need );
	}	/* ! insibling */
	if ( s ) { /* save upkey before writing, which may kill s */
		unsigned char *e0 = ((unsigned char*)s) + LEAFPOS(s->dct);
		memcpy( up.byt, e0, up.len = s->dct[0].kln );
		sibnum = s->num;
		if ( split )
			memcpy( up.val.byt, e0+up.len, up.val.len = dx->vsz );
		else
			up.val.len = 0;
		LOG_DBG( LOG_DEBUG, "split: had %d/%d key '%b' vln %d %s",
			b->num, s->num, up.len, up.byt, up.val.len, PRVAL( pv, up.val ) );
		wrtleaf(s, dx);
	}
	wrtleaf(b, dx);

	/*
	WHOA -- now all that was just to care for the leaf level!
	Now walk up inserting pointers in parent forks while we split.
	*/

#ifdef BUILD_SHMODE /* lock tree, unlock leaves */
	QLOCKUN(lb<<1);
	lb = -1;
	if ( sibnum ) {
		QLOCKEX(1); /* get tree lock */
		lt = 1;
		QLOCKUN(ls<<1);
		ls = -1;
	}
#endif
	while ( sibnum ) { /* created sibling */
		unsigned char *e0;
		int tog = 0;
		Block *ins;

		/* get parent */
		lev++;
		b = getfork(&blk[0].b, dx, f.n[lev], lev);
		LOG_DBG( LOG_DEBUG,
			"split: got parent %d lev %d", b->num, b->lev );
		e = keyfork(b, &up, dx->vsz);
		/* again L/Y link hunt */
		while ( !(QMATCH & e) && b->ent == e && b->nxt && b->num ) {
			unsigned se;
			/* peek next block, toggling buffers */
			s = getfork(&blk[(++tog)&1].b, dx, b->nxt, lev);
			se = keyfork(s, &up, dx->vsz);
			if ( !se ) /* lower than first */
				break;
			LOG_DBG( LOG_DEBUG, "LYao %d -> %d lev %d for '%b' got 0x%4x",
				b->num, s->num, i, up.len, up.byt, se );
			/* trade b for s */
			b = s;
			e = se;
		}
		base = (unsigned char*)b;
		ins = b;
		s = 0;
		LOG_DBG( LOG_DEBUG, "split: got parent pos %4x", e );
		if ( QMATCH & e ) /* !??? */
			DONE((ERR_TRASH, "OOPS! found key in parent"));
		if ( ! e ) {
			TRACEKEY( ERR_TRASH, &f );
			DONE((ERR_TRASH, "OOPS! parent insert at 0 lev %d", lev));
		}
		need = sizeof(Dict) + up.len + 4 + up.val.len;
		if ( need > b->stk - DICTOFF - b->ent*sizeof(Dict) ) {
			unsigned half = ((env.psz - b->stk)+b->ent*sizeof(Dict)+need) / 2;
			unsigned mv = b->ent, tot = 0, nent, sz;
			int adj;
			if ( e == b->ent ) {
				tot = need;
				ins = 0;
			}
			for ( d = b->dct+mv; d--, mv--; ) {
				sz = sizeof(Dict) + d->kln + 4 + (d->vln ? dx->vsz : 0);
				if ( half > (tot += sz) ) {
					if ( e == mv ) {
						if ( half > tot+need || tot+need-half < half-tot ) {
							tot += need;
							ins = 0; /* new link goes to sibling */
						} else
							break;
						if ( half <= tot )
							break;
					}
					if ( mv )
						continue;
				}
				if ( tot - half > sz/2 ) { /* don't move this */
					tot -= sz;
					d++; mv++;
				} else if ( ! mv ) /* !??? */
					mv = 1;
				break;
			}
			/* split here a little bit simpler */
			s = newfork( b==&blk[0].b ? &blk[1].b : &blk[0].b, dx, lev );
			s->nxt = b->nxt;
			b->nxt = s->num;
			/* straight move entries mv..b->ent-1 to sibling */
			sbase = (unsigned char*)s;
			d = b->dct + mv;
			sz = b->dct[mv-1].pos - b->stk;
			nent = b->ent - mv;
			LOG_DBG( LOG_DEBUG,
				"split: parent lev %d split %d to new %d mv %d"
				" '%b'/%s nent %d sz %d",
				lev, b->num, s->num, mv, d->kln, base+d->pos,
				prval( pv, base+d->pos+d->kln, d->vln?dx->ptr:0 ), nent, sz );
			memcpy( sbase + (s->stk -= sz), base + b->stk, sz );
			adj = s->stk - b->stk;
			b->stk += sz;
			memcpy( s->dct, d, nent*sizeof(Dict) );
			b->ent = mv;
			s->ent = nent;
			for ( d = s->dct; nent--; d++ )
				d->pos += adj;
			if ( ! ins ) {
				ins = s;
				e -= mv;
			}
			if ( (ret = chkblk( s, dx )) )
				break;
		}
		/* now insert link in block ins at pos e */ {
			unsigned isz = up.len + 4 + up.val.len, nent = ins->ent - e;
			unsigned char *ibase = (unsigned char*)ins;
			unsigned end = e ? ins->dct[e-1].pos : env.psz;
			LOG_DBG( LOG_DEBUG,
				"split: ins parent %d lev %d pos %d key '%b' -> %d vln %d %s",
				ins->num, ins->lev, e, up.len, up.byt, sibnum, up.val.len,
				PRVAL( pv, up.val ) );
			memmove( ibase + ins->stk - isz, ibase + ins->stk, end - ins->stk );
			ins->stk -= isz;
			SETINT(ibase + (end -= 4), sibnum);
			if ( up.val.len )
				memcpy( ibase + (end -= up.val.len), up.val.byt, up.val.len );
			memcpy( ibase + (end -= up.len), up.byt, up.len );
			for ( d = ins->dct+ins->ent; nent--; d-- ) {
				*d = d[-1];
				d->pos -= isz;
			}
			d->pos = end;
			d->vln = up.val.len ? 1 : 0;
			d->kln = up.len;
			ins->ent++;
			if ( (ret = chkblk( ins, dx )) )
				break;
		}
		wrtfork(b, dx);
		if ( ! s ) /* not split - we're done */
			break;
		/* save sibling's key */
		sibnum = s->num;
		e0 = ((unsigned char*)s) + s->dct[0].pos;
		memcpy( up.byt, e0, up.len = s->dct[0].kln );
		if ( s->dct[0].vln )
			memcpy( up.val.byt, e0+up.len, up.val.len = dx->vsz );
		else
			up.val.len = 0;
		LOG_DBG( LOG_DEBUG, "split: had %d/%d on lev %d key '%b' vln %d %s",
			b->num, sibnum, s->lev, up.len, up.byt, up.val.len, PRVAL( pv, up.val ) );
		wrtfork(s, dx);
		if ( lev < dx->dpt )
			continue;
		LOG_DBG( LOG_DEBUG, "root split for dx %8x!", dx);
		if ( b != &blk[0].b && b != &blk[1].b ) /* was in map -- could be gone */
			b = getfork(&blk[0].b, dx, 0, lev); /* refetch */
		LOG_DBG( LOG_DEBUG, "refetched root %8x!", b);
		/* move old root to new block, use sibling buffer (if any) */
		s = newfork(b==&blk[0].b ? &blk[1].b : &blk[0].b, dx, lev);
		/* cp all but the num */
		memcpy( ((unsigned char*)s)+4, ((unsigned char*)b)+4, env.psz-4 );
		s->nxt = sibnum; /* and link */
		wrtfork(s, dx);
		b->stk = env.psz; /* empty the root */
		/* add two childs linking to old root and it's sibling */
		base = (unsigned char*)b;
		SETINT( base + (b->stk -= 4), s->num ); /* link the old root */
		b->dct[0].pos = b->stk; b->dct[0].vln = 0; b->dct[0].kln = 0;
		SETINT( base + (b->stk -= 4), sibnum ); /* link the sibling */
		if ( ! up.val.len )
			b->dct[1].vln = 0;
		else {
			memcpy( base + (b->stk -= dx->vsz), up.val.byt, dx->vsz );
			b->dct[1].vln = 1;
		}
		memcpy( base + (b->stk -= up.len), up.byt, b->dct[1].kln = up.len );
		b->dct[1].pos = b->stk;
		b->ent = 2;
		b->lev++; /* increase tree depth */
		dx->dpt = b->lev;
		ret = wrtfork(b, dx);
		break;
	}
done:
#ifdef BUILD_SHMODE
	if ( -1 != lb ) QLOCKUN(lb<<1);
	if ( -1 != ls ) QLOCKUN(ls<<1);
	if ( lt ) QLOCKUN(1);
#endif
	if ( 0 <= ret && b->lev && FINDKEY(&blk[0].b, &f, LOCK_READ) )
		ret = eRr(ERR_TRASH, "could not find key after add" );
	return ret;
}	/* qSet */


int qDel ( Qdx *dx, Key *key )
{
	Leaf leaf;
	Block * const b = &leaf.b;
	unsigned char * const base = leaf.base;
	int ret = 0;
	unsigned short e, n, i, mv;
	Dict *d;
	unsigned char *v;
	FindKey f;
	PrVal pv;

	f.dx  = dx;
	f.key = key;
	LOG_DBG( LOG_DEBUG, "deleting key '%b'/%s",
		key->len, key->byt, PRVAL( pv, key->val ) );
	if ( FINDKEY(b, &f, LOCK_WRITE) )
		return eRr(ERR_TRASH, "could not find key");
	if ( !(QMATCH & f.e[0]) ) /* key not found - ok */
		goto done;
	LOG_DBG( LOG_DEBUG, "key found" );
	e = (QENTMASK & f.e[0]);
	d = b->dct + e;
	v = base+LEAFPOS(d)+d->kln; /* start of values */
	n = LEAFVLN(d);
	if ( ! n )
		goto done;
	i = valpos( v, n, dx->vsz, key->val.byt );
	LOG_DBG( LOG_DEBUG, "val %sfound pos %u of %u",
		(QMATCH & i) ? "" : "not ", i & ~QMATCH, n );
	if ( !(QMATCH & i) ) /* val not found - ok */
		goto done;
	i &= ~QMATCH;
	LOG_DBG( LOG_DEBUG,
		"del blk %d e %d i/n %d/%d", b->num, e, i, n );
	if ( 1 < n ) {
		v += i*dx->vsz; /* end of stack area to move: start of this value */
		mv = dx->vsz; /* amount to move: one value size */
		LEAFVLNADD( d, -1 );
	} else { /* sole value -- rm whole entry */
		v = base+LEAFPOS(d); /* start of key */
		mv = d->kln+dx->vsz;
		for ( i=b->ent - e - 1; i--; d++ ) /* move the i entries after d down */
			*d = d[1];
		d = b->dct + e; /* reset */
		b->ent--;
	}
	memmove( base+b->stk+mv, base+b->stk, v-(base+b->stk) );
	b->stk += mv;
	for ( ; e++ < b->ent; d++ )
		LEAFPOSADD( d, mv );
	wrtleaf(b, dx);
done:
	QLOCKUN(b->num<<1);
	return ret;
}	/* qDel */


static int setKeyVal (QSet *qst, char *val, int len, int split)
{
	Qdx *qdx = qst->qdx;
	char *e = val+len;
	Key key;
	int nope, done = 0;

	if (split && qst->pfx.len)
		split = 1+qst->pfx.len; /* don't split on 1st pfx.len bytes */
	do {
		char buf[512], *b;
		int l;

		if (!split || qdx->cdx)
			l = e-val;
		else {
			char *s;
			while (val<e && (CT_WHITE|CT_SPECL)&lat1ct[(unsigned char)*val]) val++;
			if (e == (s = val))
				break;
			while (s<e && !((CT_WHITE|CT_SPECL)&lat1ct[(unsigned char)*s])) s++;
			l = s-val;
		}
		b = val;
		if (qst->pfx.len) {
			memcpy(b = buf, qst->pfx.byt, qst->pfx.len);
			if (l > (int)sizeof buf-qst->pfx.len)
				l = sizeof buf-qst->pfx.len;
			memcpy(buf+qst->pfx.len, val, l);
			l += qst->pfx.len;
		}
		/* qMkKey(qst->qdx, &key, val, len): */
		if (!qdx->cdx)
			memcpy(key.byt, b, key.len = l<qdx->ksz ? l : qdx->ksz);
		else {
			key.len = qdx->ksz;
			l = cEnc(qdx->cdx, &key, (unsigned char*)b, l, split);
			if ( split && (!key.len /* wordless */
				|| l <= qst->pfx.len) /* houston, we have a problem */
			)
				break;
		}
		qMkVal(&key.val, &qst->ptr, qdx->ptr);
		qst->ptr.pos++;
		eRr(LOG_VERBOSE, "%s '%.*s'->'%b' %u %u %u %u",
			qst->del ? "del" : "set", l, b, key.len, key.byt,
			qst->ptr.rid, qst->ptr.tag, qst->ptr.pos>>16, qst->ptr.pos&0xffff);
		nope = qst->del ? qDel(qdx, &key) : qSet(qdx, &key);
		if (!split)
			return nope;
		val += l - qst->pfx.len;
		if (!nope)
			done++;
	} while (val < e);
	return done;
}	/* qSetKeyVal */

int qSetKeyVal (QSet *qst, char *val, int len)
{
	return setKeyVal(qst, val, len, 0);
}

int qSetKeyVals (QSet *qst, char *val, int len)
{
	return setKeyVal(qst, val, len, 1);
}	/* qSetKeyVals */



int qLoop ( QLoop *self )
{
	Leaf leaf;
	Qdx *dx = self->qdx;
	int ret = 0;
	unsigned i = 0;
	int stop = QSTOP & self->flg;
	int skip = QSKIP & self->flg;
	Block *b = &leaf.b;
	Key *cmp = &self->key;

	if ( ! cmp->len ) { /* start at 1st leaf */
		if ( b != GETLEAF(b, dx, 0, LOCK_READ) )
			DONE((ERR_TRASH, "could not read leaf 0"));
		self->cur.len = 1; /* avoid QSAME match */
	} else { /* locate key */
		FindKey f;
		f.dx  = dx;
		f.key = cmp;
		if ( FINDKEY(b, &f, LOCK_READ) )
			DONE((ERR_TRASH, "could not find key"));
		i = (unsigned)(QENTMASK & f.e[0]);
		if ( !(QMATCH & f.e[0]) && QEQ == stop )
			goto done;
		self->cur.len = 0; /* avoid QSAME match */
	}
	if ( QUPTO <= stop )
		cmp = &self->to;
	for ( ;; i=0 ) { /* blocks */
		unsigned char *base = (unsigned char*)b;
		Dict *d;
		if ( ! b || b->lev )
			return ERR_TRASH;
		d = b->dct + i;
		for ( ; i<b->ent; i++, d++ ) {
			unsigned short pos = LEAFPOS( d );
			unsigned short vln = LEAFVLN( d );
			unsigned short kln = d->kln;
			unsigned char *v = base+pos+kln;
			Ptr p;

			if ( skip ) {
				if ( kln == self->key.len
					&& !memcmp(self->key.byt, base+pos, kln)
				)
					continue;
				skip = 0;
			}
			if ( stop ) {
				int diff = memcmp( cmp->byt, base+pos,
					kln <= cmp->len ? kln : cmp->len );
				switch ( stop ) {
				case QPF:
					if ( diff || kln < cmp->len )
						goto moan;
					break;
				case QEQ:
					if ( diff || kln != cmp->len )
						goto done;
					break;
				case QUPTO:
					if ( 0 >= diff )
						goto done;
					break;
				case QINCL:
					if ( 0 > diff || (!diff && kln > cmp->len) )
						goto done;
					break;
				default: /* unused */
					moan:
					eRr(LOG_INFO,
						"loop ends on key '%b'(%d) %d %d",
						kln, base+pos, kln, diff, cmp->len );
					goto done;
				}
			}	/* if stop */

			if ( !i && kln == self->cur.len && !memcmp(self->cur.byt, base+pos, kln) )
				self->flg |= QSAME;
			else {
				self->flg &= ~QSAME;
				memcpy( self->cur.byt, base+pos, self->cur.len = kln );
			}
			if ( self->qcb ) {
				self->vals = v;
				self->nvals = vln;
				if ( (ret = self->qcb(self)) )
					goto done;
			} else { /* internal check */
				Leaf ll;
				FindKey f;
				f.dx  = dx;
				f.key = &self->cur;
				memcpy( self->cur.val.byt, v, self->cur.val.len = dx->vsz ); 
				if ( FINDKEY(&ll.b, &f, LOCK_READ) )
					DONE((ERR_TRASH, "could not find key"));
				if ( !(QMATCH & f.e[0])
					|| b->num != ll.b.num
					|| i != (unsigned)(QENTMASK & f.e[0])
				) {
					eRr(ERR_TRASH,
						"OOPS! want %u.%u got %u.%u (match %4x)",
						b->num, i, ll.b.num, QENTMASK & f.e[0], f.e[0] );
					qRdVal( &p, v, dx->vsz );
					eRr(ERR_TRASH,
						"\twhen looking for '%b'/%u.%u.%u",
						self->cur.len, self->cur.byt, p.rid, p.tag, p.pos );
					if ( ll.b.ent <= (QENTMASK & f.e[0]) )
						eRr(ERR_TRASH, "\tI found nothing" );
					else {
						Dict *dd = ll.b.dct + (QENTMASK & f.e[0]);
						unsigned char *bb = ((unsigned char*)&ll.b)+LEAFPOS(dd);
						qRdVal( &p, bb+dd->kln, dx->vsz );
						eRr(ERR_TRASH,
							"\tI found '%b'/%u.%u.%u",
							dd->kln, bb, p.rid, p.tag, p.pos );
					}
				} else
					eRr(LOG_TRACE, "key '%b' ok", self->cur.len, self->cur.byt );
			} /* internal check */
		}
		if ( ! b->nxt || !GETLEAF(b, dx, b->nxt, LOCK_READ) )
			break;
		/* TODO: 
		b now is the block that was our sibling when reading the old b.
		As long as we don't shift entries,
		there is no need to advance to next key that is actually greater !?
		*/
	}
done:
	if ( ! self->qcb )
		eRr(LOG_INFO,
			"checked %d keys in %u blks of %dK depth %u cmp/fget/lget %u/%u/%u",
			stat.key, dx->lln, dx->lsz>>10, dx->dpt, stat.cmp, stat.fget, stat.lget );
	return ret;
}	/* qLoop */



int qInit ( Qdx *dx )
{
	Bloke blk;
	int size;

	/* defaults */
	if ( !dx->ksz ) dx->ksz = 255;

	/* check leaves file */
	size = fSize( dx->mqd );
	if ( 0x3ff & size ) /* not on 1K boundary */
		return eRr(LOG_ERROR, "mqd has bad size %dK + %d",
			size >> 10, (int)size & 0x3ff );
	if ( size ) { /* read 1st leaf header */
		Block b;
		int got = fPread( &dx->mqd, (unsigned char*)&b, sizeof(b), 0 );
		if ( sizeof(b) != (unsigned)got )
			return eRr(LOG_SYSERR, "could not get leaf head" );
		if ( b.num )
			return eRr(ERR_TRASH, "leaf has num %d", b.num );
		if ( dx->typ && dx->typ != b.typ )
			eRr(LOG_WARN, "typ want 0x02%2x got 0x02%2x", dx->typ, b.typ );
		if ( QDX_COMPRS & dx->typ )
			eRr(LOG_WARN, "sorry this version does not support compressed keys");
		dx->typ = b.typ;
		dx->ksz = b.ksz;
		dx->ptr = b.ptr;
	}
	if ( QDX_LEAFPV & dx->typ )
		dx->vsz = dx->ptr;
	else
		dx->vsz = (3&(dx->ptr>>6)) + 3+(3&(dx->ptr>>4)) + (7&dx->ptr);
	if (!size && dx->ksz > 255 - dx->vsz)
		dx->ksz = 255 - dx->vsz;
	eRr(LOG_VERBOSE, "dx ptr %2x vsz %d=%d+%d+%d", dx->ptr, dx->vsz,
		(3&(dx->ptr>>6)), 3+(3&(dx->ptr>>4)), (7&dx->ptr));
	dx->lsz = 1<<QDX_LEAFSH<<(QDX_SIZMSK&dx->typ);
	if ( size % dx->lsz )
		return eRr(ERR_TRASH, "size 0x%08x %d blksize", size, dx->lsz );
	dx->lln = size / dx->lsz;
	if ( ! dx->lln
		&& (!(DX_WRITE & dx->flg) || (newleaf(&blk.b, dx), wrtleaf(&blk.b, dx)))
	)
		return eRr(LOG_SYSERR, "could not init new leaf file" );

#ifdef CPU_BIG_ENDIAN
#	define QDX_FORK (QDX_FORKBE | (env.psh-QDX_FORKSH))
#else
#	define QDX_FORK (QDX_FORKLE | (env.psh-QDX_FORKSH))
#endif
	size = fSize(dx->mqx.fil);
	if ( size ) {
		if ( size % env.psz ) /* not on env.psz boundary */
			return eRr(LOG_ERROR, "mqx has bad size %dK + %d",
				size >> 10, (int)size & 0x3ff );
		dx->fln = size >> env.psh;
		if ( (int)dx->fln != fMap(&dx->mqx, dx->fln) )
			return eRr(LOG_ERROR, "could not map mqx");
		dx->ftp = ((Block*)dx->mqx.map)->typ;
		if ( dx->ftp != QDX_FORK ) /* bad pagesize, endianess or flags */
			return eRr(LOG_ERROR, "root has type %2x != %2x",
				dx->ftp, QDX_FORK );
		dx->dpt = ((Block*)dx->mqx.map)->lev;
	}
	dx->fln = size / env.psz;
	if ( ! dx->fln ) {
		dx->ftp = QDX_FORK;
		eRr(LOG_INFO, "creating index type 0x%2x/0x%2x lsz %u psz %u ksz %u vsz %u",
			dx->typ, dx->ftp, dx->lsz, env.psz, dx->ksz, dx->vsz );
		if ( 1 < dx->lln ) {
			if (ENV_EXCL!=env.wri || !(DX_WRITE & dx->flg))
				return eRr(ERR_TRASH, "need excl write access to rebuild tree");
			if ( qRemake(dx) )
				return eRr(LOG_ERROR, "could not rebuild tree");
		} else {
			/* init: link root to leaf */
			Block *b = newfork(&blk.b, dx, 1);
			SETINT( ((unsigned char*)b)+(b->stk -= 4), 0 );
			b->dct[0].pos = b->stk; /* 0, 0 */
			b->ent = 1;
			if ( wrtfork(b, dx) )
				return eRr(LOG_SYSERR, "\twhen writing root" );
			fMap(&dx->mqx, 1);
			dx->dpt = 1;
		}
	}
	eRr(LOG_INFO, "psz %d mapped %d", env.psz, dx->mqx.npg );

	return 0;
}	/* qInit */


void qFini ( Qdx *dx )
{
	fMap(&dx->mqx, 0);
}	/* qFini */
