/*
	The Malete project - the Z39.2/Z39.50 database framework of OpenIsis.
	Version 0.9.x (patchlevel see file Version)
	Copyright (C) 2001-2003 by Erik Grziwotz, erik@openisis.org

	This library is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
	See the GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	see README for more information
EOH */

/*
	$Id: rdx.c,v 1.1 2005/01/24 18:03:28 kripke Exp $
	implementation of general db access functions.
*/


#include "core.h"



/*
	type of rx
	posbytes in [4..8], lenbytes in [3..6], fldbytes in [0..2]
*/
#define RX_TYPE(posbytes,lenbytes,fldbytes) \
	((posbytes-4)<<4 | (lenbytes-3)<<2 | fldbytes)
enum { /* by now, only the two 8 byte types are implemented */
	RX_431 = RX_TYPE(4,3,1),
	RX_440 = RX_TYPE(4,4,0)
};

/* ************************************************************
	private types
*/

typedef union {
	lolo bar;
	char r[16];
} Xrf;

/** header of first record index block */
typedef struct Rxh {
	char     mag[3]; /* magic MRX or mrx */
	char     typ;
	unsigned mid; /* max record id == total number of used pointers */
	/* for type of more than 8 byte: upper mid */
} Rxh;


/* ************************************************************
	data
*/

/* the magic we need depending on our endianess */
static char MAG_MRX[] =
#ifdef CPU_BIG_ENDIAN
	"MRX"
#else
	"mrx"
#endif
;


/* ************************************************************
	private functions
*/
#define RLOCKSH(n) (void)FLOCKSH(dx->mrd, n)
#define RLOCKEX(n) (void)FLOCKEX(dx->mrd, n)
#define RLOCKUN(n) (void)FLOCKUN(dx->mrd, n)

/*
	create a pointer from the least significant bytes of pos, len, fld
	buf must have dx->ptl bytes (up to 16 = 8+4+4)
	and the most strict alignment (i.e. 4 or 8) possible for dx->ptl
	returns buf
	supports only RX_431 and RX_440
*/
static char *mkxrf ( char *buf, Rdx *dx,
	unsigned pos, unsigned len, unsigned fld )
{
	((unsigned*)buf)[0] = pos;
	if ( RX_431 == dx->typ ) {
		if ( ~0xff & fld ) fld = 0; /* too big */
#ifdef CPU_BIG_ENDIAN /* the first = high order 3 bytes are len */
		len = len<<8 | fld;
#else /* the first = low order 3 bytes are len */
		len = (0xffffff&len) | fld<<24;
#endif
	}
	((unsigned*)buf)[1] = len;
	return buf;
}


/*
	read pointer, return len
	if 0x0f00 & dx->typ, fld must not be 0, else *fld is untouched
	supports only RX_431 and RX_440
*/
static unsigned
	rdxrf ( unsigned *pos, unsigned *fld, Rdx *dx, char *buf )
{
	unsigned len = ((unsigned*)buf)[1];
	*pos = *(unsigned*)buf;
	if ( RX_431 == dx->typ ) {
		*fld = ((unsigned char *)buf)[7];
#ifdef CPU_BIG_ENDIAN
		len >>= 8;
#else
		len &= 0xffffff;
#endif
	}
	return len;
}


static void setXrf ( Rdx *dx, unsigned rid,
	unsigned pos, unsigned len, unsigned fld )
{
	Xrf xrf;
	unsigned off = rid*dx->ptl;
	eRr(LOG_INFO,"setXrf %d %d %d %d",rid,pos,len,fld);
	RLOCKEX(rid);
	if ( dx->mrx.npg ) {
		if ( (off+dx->ptl) > (dx->mrx.npg<<env.psh) ) { /* remap ? */
			unsigned npg = 1 + ((off+dx->ptl-1)>>env.psh); /* num pages we need */
			if ( npg > dx->mrx.lim ) { /* at limit, drop the mapping completely */
				eRr(LOG_WARN, "dropping mrx mapping at limit %u", dx->mrx.lim);
				npg = dx->mrx.lim = 0;
			}
			fMap(&dx->mrx, npg);
		}
		if ( (off+dx->ptl) <= (dx->mrx.npg<<env.psh) ) {
			mkxrf(dx->mrx.map + off, dx, pos, len, fld);
			if ( ENV_MSYNC & env.flg ) fMSync(&dx->mrx, off>>env.psh);
			goto done;
		}
	}
	fPwrite(&dx->mrx.fil, mkxrf(xrf.r, dx, pos, len, fld), dx->ptl, off);
done:
	if ( rid > (unsigned)dx->mid ) {
		if ( dx->mrx.npg )
			((Rxh*)dx->mrx.map)->mid = rid;
		else
			fPwrite(&dx->mrx.fil, &rid, 4, 4);
		dx->mid = rid;
	}
	RLOCKUN(rid);
}	/* setXrf */


static unsigned getXrf ( unsigned *pos, unsigned *fld, Rdx *dx, int rid )
{
	unsigned ret;
	Xrf xrf;
	unsigned off = (unsigned)rid*dx->ptl;
	RLOCKSH(rid);
	if ( (off+dx->ptl) <= (dx->mrx.npg<<env.psh) )
		ret = rdxrf(pos, fld, dx, dx->mrx.map + off);
	else {
#ifdef BUILD_SHMODE
		if ( ENV_SHARED==env.wri ) { /* might have been extended */
			unsigned pgl;
			/* TODO: maybe we should guard this by the record lock ? */
			if ( dx->mrx.npg )
				dx->mid = ((Rxh*)dx->mrx.map)->mid;
			else
				fPread(&dx->mrx.fil, &dx->mid, 4, 4);
			if ( rid > dx->mid ) {
				ret = 0;
				goto out;
			}
			pgl = fSize(dx->mrx.fil) >> env.psh;
			if ( pgl != dx->mrx.npg )
				fMap(&dx->mrx, pgl);
			if ( (off+dx->ptl) <= (dx->mrx.npg<<env.psh) ) {
				ret = rdxrf(pos, fld, dx, dx->mrx.map + off);
				goto out;
			}
		}
#endif
		ret = rid > dx->mid || dx->ptl != fPread(&dx->mrx.fil, xrf.r, dx->ptl, off)
			? 0 : rdxrf(pos, fld, dx, xrf.r);
	}
#ifdef BUILD_SHMODE
out:
	RLOCKUN(rid);
#endif
	return ret;
}	/* getXrf */


static int rEmake ( Rdx *dx )
{
	char buf[65536+1]; /* need 64K buf for copying DO NOT SHRINK !!! */
	int sz;
	char *p, *q;
	Xrf xrf;
	unsigned base = 0; /* of current block */
	unsigned pos = 0; /* of last record */
	/* record parsing starts at the first char of rid 1, so fld is initially 1 */
	unsigned fld = 1; /* of last record */
	unsigned nrid = 1; /* next rid = maxrid+1 */
	unsigned xrid = 0; /* explicitly given */
	char op = 0;
	int more; /* buf not empty flag */
	char *last; /* of current block */

	/* TODO: base on FBuf? */
	if ( fTrunc(&dx->mrx.fil, 0) )
		return eRr( LOG_ERROR, "could not trunc mrx" );
	/* write signature */
	memcpy( xrf.r, MAG_MRX, 3 );
	xrf.r[3] = dx->typ;
	memset( xrf.r+4, 0, dx->ptl-4 );
	fPwrite( &dx->mrx.fil, xrf.r, dx->ptl, 0 );
	/* loop the masterfile */
	fSeek( &dx->mrd, 0 );
	last = (p = buf) + fRead( &dx->mrd, buf, 8192 ) - 1;
	more = last > buf; /* one byte is no byte ;) */
	for (;;) { /* records */
		unsigned len, rid;
		for (;;) { /* lines and stuff to end of record */
			if ( p < last ) { /* have one lookahead */
				if ( LF != *p++ )
					continue; /* the tight loop ... or use memchr ? */
				if ( LF != *p ) { /* now p <= last */
					fld++; /* count field, unless we really recognize a opline */
					if ( 'W' != *p )
						continue;
					sz = last - p; /* avail after p */
					if ( sz > 127 ) /* longer -> no opline */
						sz = 127;
					if ( ! sz || ! (q = memchr(p+1, LF, sz)) ) {
						if ( sz >= 127 || ! more )
							continue; /* too long or undelimited last */
						p--; /* back to \n, so we come here again */
						goto gimmemore;
					}
					if ( q < p+3 || p[2] < '0' || '9' < p[2] )
						continue;
					/* TODO:
					take a closer look at whether the whole line makes sense
					*/
					if ( op ) { /* yeah, two metas in sequence! weird stuff! */
						p--; /* step back to newline */
						pos = base+(p-buf); /* fake pos as if we had no line at all */
						break; /* go handle the PREVIOUS opline */
					}
					op = *p;
					xrid = a2i( p+2, q-p-2 );
					continue;
				}
				break;
			}
		gimmemore:
			LOG_DBG( LOG_DEBUG, "MORE %d at pos %d base %d p +%d last +%d",
				more, pos, base, p-buf, last-buf );
			if ( !more )
				goto schicht; /* german: done */
			base += p - buf; /* shift out bytes before p */
			len = last-p; /* bytes to keep after p; < 128 */
			if ( len ) /* we're probing for more lookahead */
				memmove( buf, p, 1+last-p );
			else /* typically */
				*buf = *p; /* but save the last dance */
			p = buf;
			last = buf + len;
			/* reload */
			if ( 0 < (sz = fRead( &dx->mrd, buf+1+len, 8192 )) ) {
				last += sz;
				continue;
			}
			more = 0; /* but yet, finish this up */
			/* since *buf = *last was the files last character,
				we'd expect a newline
			*/
			if ( last == p )
				p = buf+(LF==*buf ? 1 : 2); /* pretend buf started \n */
			if ( ! len )
				break;
			/* else try again opline */
		} /* lines and stuff */
		/* now p is on a delimiting blank lines \n -- or such ... */
		len = base + (p-buf) - pos; /* >= 0 */
		rid = xrid ? xrid : nrid;
		eRr( LOG_INFO, "mrx %c %d(%d/%d) pos %d len %d",
			op?op:'>', rid, xrid, nrid, pos, len );
		if ( base + (p-buf) < pos ) /* FOO !!! */
			len = 0;
		if ( rid && (len || op) )
			fPwrite( &dx->mrx.fil,
				mkxrf( xrf.r, dx, pos, len, fld), dx->ptl, rid*dx->ptl );
		else if (!more)
			break;
		pos = base + (p-buf) + 1; /* next starts after p */
		if (!op)
			nrid++; /* continue after this */
		else if (nrid <= xrid)
			nrid = xrid+1;
		xrid = fld = op = 0;
	}
schicht: ;
	dx->mid = nrid-1;
	fPwrite(&dx->mrx.fil, &dx->mid, 4, 4);
	fTrunc(&dx->mrx.fil,
		((1+dx->mid)*dx->ptl+env.psz-1)>>env.psh<<env.psh);
	return 0;
}	/* rEmake */



/* ************************************************************
	public functions
*/

/**
	get text
	the original text is read contigously at base.
*/
int rRead ( List *l, Rdx *dx, int rid, unsigned mpos )
{
	unsigned pos, len, n=0;
	int head; /* offset of header */
	Fld *f;
	char *rd, *p, *q, *e;
	int got, first = 1, cont = 1;

	len = getXrf( &pos, &n, dx, rid ); /* this maybe locked */
	head = LLEN(l); /* not necessarily 0 */
	LOG_DBG( LOG_TRACE, "dRead %d pos %d len %d fld %d at head %d",
		rid, pos, len, n, head );
	if ( !len )
		return 0;
	/*
	reading the current version is easy, since we know the byte length
	and usually also number of fields.
	When backtracking to older versions, we have to figure out everything
	from the position alone.
	*/
	for (; cont; first=0) { /* extend or backtrack */
		if ( !n ) {
			n = len / 36; /* assume one (costing 12 bytes) per 36 bytes data */
			if ( n < 8 ) /* small record is likely to have some short fields */
				n = 8;
		}
		/* allow 22 bytes to print a nice rid@pos in header,
			plus one for a sentinel LF */
		if ( l->fav < n || LAVL(l) < len+23 )
			lExtend(l, len+23, n); /* the call alone is about 2% ! */
		if ( 0 >= (got = fPread(&dx->mrd, rd = l->buf+22, len, pos)) )
			return got;
		if ( len != (unsigned)got ) {
			if ( first )
				return eRr(ERR_TRASH, "wanted %d bytes got %d at %d", len, got, pos);
			len = got;
			cont = 0;
		}
		e = (p = rd) + len;
		/* check header for rid@pos, mv p to leader */
		if ( 2 < len && 'W'==*p && TAB==p[1] ) { /* else plain record */
			int val;
			p += 2;
			if ( (got = a2il(p,e-p,&val)) ) {
				if ( val != rid )
					return eRr(ERR_TRASH, "wanted rid %d got %d at %d",
						rid, val, pos);
				p += got;
				if ( p < e && '@'==*p ) {
					p++;
					if ( (got = a2il(p,e-p,&val)) ) {
						if ( mpos && pos >= mpos ) {
							if ( val < 0 || (unsigned)val >= pos )
								return eRr(ERR_TRASH, "bad hpos %d at %d", val, pos);
							pos = (unsigned)val;
							if ( first )
								len += len>>1; /* give some safety margin */
							continue;
						}
						p += got;
					}
				}
			}
		}
		if ( mpos && pos >= mpos )
			return 0;
		if ( !first ) { /* check for complete record */
			for ( q=p; q<e && (q = memchr(q,LF,e-q)); )
				if ( ++q<e && LF==*q )
					goto ok;
			len += 32 + (len>>1); /* try using more space */
			continue;
		ok:
			len = q+1 - rd;
		}
		LOG_DBG( LOG_TRACE, "'%.*s'", len, rd );
		/* go for it */
		f = l->fld + head;
		f->len = u2a(l->buf, rid); /* print header */
		l->buf[f->len++] = '@';
		f->len += u2a(l->buf+f->len, pos); /* now f->len <= 21 */
		if ( p == rd ) { /* got no header -- rec data starts at rd */
			if ( l->buf < (f->val = rd - f->len - 1) )
				memmove(f->val, l->buf, f->len); /* just to keep it contiguous */
			f->val[f->len] = LF;
		} else { /* TAB-leader-LF at p */
			memmove(f->val = p - f->len, l->buf, f->len);
			q = memchr(p,LF,e-p);
			if ( !q ) return eRr(ERR_TRASH, "missing LF");
			f->len = q - f->val;
		}
		p = f->val + f->len + 1;
		f->tag = -1;
		if ( head ) l->fld->tag--; /* count in total length */
		l->fav--;
		f++;
		*e = LF; /* sentinel for memchr */
		for ( n=0; p<e && p!=(q = memchr(p, LF, e-p+1)); p = q+1 ) {
			int min, t=0, d;
			if ( !l->fav ) {
				lExtend( l, 0, 16 );
				f = LEND(l);
			}
			/* p += a2il( p, q-p, &f->tag ); */
			if ( (min = ('-' == *p)) )
				p++;
			for ( ; 10>(d = b36val[(unsigned char)*p]); p++ )
				t = 10*t + d;
			f->tag = min ? -t : t;
			if ( TAB == *p )
				p++;
			f->len = q - (f->val = p);
			l->fld->tag--;
			l->fav--;
			f++;
			n++;
		}
		if ( head )
			l->fld[head].tag = -1-n;
		l->buf = e;
		return n+1;
	}
	return 0;
}	/* rRead */


int rWrite ( Rdx *dx, const Fld *r, int rid, int opos, unsigned siz )
{
	char buf[128 + 65536];
	LBlk *blk;
	unsigned pos = 0, rlen;
	char *p, *b;
	int intro = 1, ret;

	if ( !(dx->flg & DX_WRITE) )
		return eRr(ERR_BADF, "dx is not writable");
	rlen = RLEN(r);
	if ( rid && dx->mid >= rid ) {
		unsigned fld;
		getXrf( &pos, &fld, dx, rid );
		if ( 0<=opos && pos != (unsigned)opos )
			return eRr(ERR_AGAIN, "pos %d != opos %d", pos, opos);
	}
	RLOCKEX(0);
	if ( ! rid || rid == dx->mid+1 ) {
		rid = dx->mid+1; /* assign new rid; mid updated in setXrf */
		if ( !rlen || !r->len ) /* no header */
			intro = 0;
	}
	if ( !siz )
		siz = rSiz( r );
	siz += 13*rlen + 128; /* field, header, trailing LF */
	if ( sizeof(buf) >= siz ) {
		blk = 0;
		b = buf;
	} else {
		blk = mBlkAlloc( siz ); /* hmm ... maybe would be better ... */
		b = blk->byt; /* ... to serialize and write several 64K chunks */
	}
	p = b;
	if ( intro ) {
		*p++ = 'W';
		*p++ = TAB;
		p += u2a( p, rid );
		if ( pos ) {
			*p++ = '@';
			p += u2a( p, pos );
		}
		if ( rlen && r->len ) {
			*p++ = TAB;
			memcpy(p, r->val, r->len);
			p += r->len;
		}
		*p++ = LF;
	}
	p += rSer( p, r );
	dx->flg |= DX_MODIF;
	pos = dx->rdl;
	dx->rdl += p-b;
	ret = fPwrite(&dx->mrd, b, p-b, pos);
	if ( ret == p - b ) {
		setXrf(dx, (unsigned)rid, pos, p-b, rlen);
		ret = rid;
	} else if (0 <= ret) /* should not happen */
		ret = ERR_IDIOT;
	if ( blk )
		mBlkFree( blk );
	RLOCKUN(0);
	return ret;
}	/* rWrite */


int rInit ( Rdx *dx )
{
	int remake = 1;

	dx->rdl = fSize( dx->mrd );

	{ /* reasons why we should remake */
		Rxh head;
		int mxlen = fSize( dx->mrx.fil );

		if ( 8 != fRead(&dx->mrx.fil,&head,8) )
			eRr(LOG_WARN, "mrx too short");
		else if ( memcmp(MAG_MRX,head.mag,3) ) /* FOO! */
			eRr(LOG_WARN, "mrx has black magic '%.*s'", 3, head.mag);
		else if ( 0x80 & head.typ )
			eRr(LOG_WARN, "mrx has bad type 0x%04x", head.typ);
		else if ( (dx->typ && dx->typ != head.typ) ) /* other type configured */
			eRr(LOG_WARN, "mrx type 0x%02x != cfg 0x%02x", head.typ, dx->typ);
		else if ( mxlen < (int)(1+head.mid)*dx->ptl )
			eRr(LOG_WARN, "mrx too short for len %d", head.mid);
		else if ( mxlen % env.psz )
			eRr(LOG_WARN, "mrx len mismatch pagesize %d", env.psz);
		else if ( (DX_WRITE&dx->mrx.flg) && fTime(dx->mrx.fil) < fTime(dx->mrd) )
			eRr(LOG_WARN, "mrx older than data"); /* ignore with RO */
		else {
			dx->typ = head.typ; /* but further checked below */
			dx->mid = head.mid;
			remake = 0;
		}
	}

	switch (dx->typ) {
	default:
		eRr(LOG_WARN, "mrx type 0x%02x unsupported", dx->typ);
	case 0:
		dx->typ = RX_431;
		remake = 1;
	case RX_431:
	case RX_440:
		break;
	}
	dx->ptl = 8;

	if ( remake ) {
		if (ENV_EXCL!=env.wri || !(DX_WRITE & dx->flg))
			return eRr(ERR_TRASH, "need excl write access to rebuild .mrx");
		rEmake(dx);
	}

	fMap(&dx->mrx, fSize(dx->mrx.fil)>>env.psh);
	eRr(LOG_INFO, "mapped %d*%d = %d", 1+dx->mid, dx->ptl, dx->mrx.npg);
	return 0;
}	/* rInit */


void rFini ( Rdx *dx )
{
	if ( DX_MODIF & dx->flg ) {
		if ( dx->mrx.npg ) {
			((Rxh*)dx->mrx.map)->mid = dx->mid;
			fMap(&dx->mrx, 0);
		} else
			fPwrite(&dx->mrx.fil, &dx->mid, 4, 4);
	}
}	/* rFini */
